--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		12/12/2021
--------------------------------------------------------------------------------
-- IP name:		hybrid_compr_img_tail
--
-- Description: It computes the 'compression image tail' to integrate at the
--				very end of the Hybrid Entropy compression.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.hybrid_flush_tables.all;
use work.utils_encoder.all;

entity hybrid_compr_img_tail is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	  : std_logic_vector(1 downto 0);
		-- Maximum number of bits (size) of every encoded data sample
		MAX_NUM_BITS_G	  : integer
	);
	port (
		clock_i			  : in  std_logic;
		reset_i			  : in  std_logic;
		
		enable_i		  : in  std_logic;
		enable_o		  : out std_logic;
		img_coord_i		  : in  img_coord_t;
		img_coord_o		  : out img_coord_t;
		
		high_res_accu_i	  : in  std_logic_vector(D_C*2-1 downto 0);				-- "Σ~z(t)" (high resolution accumulator)
		curr_prefix_arr_i : in  array_string_t(0 to NUM_HYBRID_TABLES_G-1)(1 to LONGEST_STRING_SIZE_G);	-- Current prefix for every Hybrid table		
		enc_num_bits_i	  : in  std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		rvd_codeword_i	  : in  std_logic_vector(Umax_C+D_C-1 downto 0);		-- "R'k(j)" (reversed codeword)

		out_num_bits_o	  : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		out_codeword_o	  : out std_logic_vector(Umax_C+D_C-1 downto 0)
	);
end entity hybrid_compr_img_tail;

architecture behavioural of hybrid_compr_img_tail is
	-- Signals declaration
	signal enable_s				: std_logic   := '0';
	signal img_coord_s			: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal high_res_accu_arr_s  : array_slv_t(0 to NZ_C-1)(D_C*2-1 downto 0)  := (others => (others => '0'));	-- To save the last "Σ~z(t)" of every spectral band
	signal flush_prefix_arr_s	: array_integer_t(0 to NUM_HYBRID_TABLES_G-1) := (others => 0);					-- No negative numbers are accepted here!
	
	signal last_bit_flag_s		: std_logic   := '0';
	signal tail_start_s			: std_logic   := '0';
	signal tail_stop_s			: std_logic   := '0';
	signal tail_cnt_s			: integer range 0 to NUM_HYBRID_TABLES_G+NZ_C+2 := 0;
	
	signal out_num_bits_orig_s	: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal out_num_bits_tail_s	: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal out_codeword_orig_s	: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	signal out_codeword_tail_s	: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_hyb_img_tail_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s	<= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_hyb_img_tail_delay;
	
	-- Data forwarding while the image has not been fully procesed
	p_forwarding : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				tail_start_s		<= '0';
				high_res_accu_arr_s <= (others => (others => '0'));
				flush_prefix_arr_s	<= (others => 0);
				out_num_bits_orig_s	<= (others => '0');
				out_codeword_orig_s	<= (others => '0');
			else
				if (enable_i = '1') then
					-- At the end of each spectral band
					if (img_coord_i.t >= NX_C*NY_C-1) then
						-- For every spectral band, the last value of "high_res_accu_i" is stored into an array
						high_res_accu_arr_s(img_coord_i.z) <= high_res_accu_i;
						
						-- In case this is the last spectral band (the end of an image)
						if (img_coord_i = end_img_coord) then
							-- For every "Hybrid Flush Table", find out a match (and there MUST be one) of the "current prefix" with the "active prefix" in the corresponding "hybrid flush table"
							for i in 0 to (NUM_HYBRID_TABLES_G-1) loop
								flush_prefix_arr_s(i) <= match_flush_prefix(HYBRID_FLUSH_TABLES_ARR_C(i).active_prefix, HYBRID_FLUSH_TABLES_ARR_C(i).table_size, curr_prefix_arr_i(i));
							end loop;
							
							-- A flag is asserted to enable the image tail generation
							tail_start_s <= '1';
						end if;
					end if;
					
					-- Input data is forwarded outside
					out_num_bits_orig_s <= enc_num_bits_i;
					out_codeword_orig_s <= rvd_codeword_i;
				else
					out_num_bits_orig_s	<= (others => '0');
					out_codeword_orig_s	<= (others => '0');
					
					-- Start flag restarted once the image tail has already been sent out
					if (tail_stop_s = '1') then
						tail_start_s	<= '0';
					end if;
				end if;
			end if;
		end if;
	end process p_forwarding;
	
	p_gen_tail : process(clock_i) is
		variable flush_word_v : integer := 0;
		variable flush_size_v : integer := 0;
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				flush_word_v		:= 0;
				flush_size_v		:= 0;
				last_bit_flag_s		<= '0';
				tail_stop_s			<= '0';
				tail_cnt_s			<= 0;
				out_num_bits_tail_s	<= (others => '0');
				out_codeword_tail_s	<= (others => '0');
			else
				if (tail_start_s = '1' and tail_stop_s = '0') then
					if (tail_cnt_s <= NUM_HYBRID_TABLES_G-1) then
						-- Flush codeword and its width signals are extracted into variables for the sake of readability
						flush_word_v		:= HYBRID_FLUSH_TABLES_ARR_C(tail_cnt_s).flush_word(flush_prefix_arr_s(tail_cnt_s));
						flush_size_v		:= HYBRID_FLUSH_TABLES_ARR_C(tail_cnt_s).flush_size(flush_prefix_arr_s(tail_cnt_s));
						-- The 16 flush codewords are outputted
						out_num_bits_tail_s <= std_logic_vector(to_unsigned(flush_size_v, MAX_NUM_BITS_G));
						out_codeword_tail_s <= std_logic_vector(to_unsigned(flush_word_v, Umax_C+D_C));
					elsif (tail_cnt_s <= NUM_HYBRID_TABLES_G+NZ_C-1) then
						out_num_bits_tail_s <= std_logic_vector(to_unsigned(2 + D_C + Y_C, MAX_NUM_BITS_G));
						out_codeword_tail_s <= std_logic_vector(resize(unsigned(high_res_accu_arr_s(tail_cnt_s-NUM_HYBRID_TABLES_G)), Umax_C+D_C));
					elsif (last_bit_flag_s = '0') then
						out_num_bits_tail_s <= std_logic_vector(to_unsigned(1, MAX_NUM_BITS_G));
						out_codeword_tail_s <= std_logic_vector(to_unsigned(1, Umax_C+D_C));
						last_bit_flag_s		<= '1';
					else
						out_num_bits_tail_s	<= (others => '0');
						out_codeword_tail_s	<= (others => '0');
						tail_stop_s			<= '1';
					end if;
					
					tail_cnt_s <= tail_cnt_s + 1;
				else
					tail_cnt_s <= 0;
				end if;
			end if;
		end if;
	end process p_gen_tail;
	
	-- Outputs
	enable_o		<= enable_s				when tail_start_s /= '1' else '1';
	img_coord_o		<= img_coord_s			when tail_start_s /= '1' else end_img_coord;
	out_num_bits_o	<= out_num_bits_orig_s	when tail_start_s /= '1' else out_num_bits_tail_s;
	out_codeword_o	<= out_codeword_orig_s	when tail_start_s /= '1' else out_codeword_tail_s;
end architecture behavioural;