--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		10/04/2021
--------------------------------------------------------------------------------
-- IP name:		hybrid_high_entr_code
--
-- Description: The high-entropy (reversed length-limited GPO2) codeword to
--				encode with the Hybrid Entropy Coder.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.utils_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;

entity hybrid_high_entr_code is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- Maximum number of bits (size) of every encoded data sample
		MAX_NUM_BITS_G		: integer;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic
	);
	port (
		clock_i			 : in  std_logic;
		reset_i			 : in  std_logic;

		enable_i		 : in  std_logic;
		enable_o		 : out std_logic;
		img_coord_i		 : in  img_coord_t;
		img_coord_o		 : out img_coord_t;
		
		err_lim_upd_en_i : in  std_logic;
		err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
		
		high_res_accu_i	 : in  std_logic_vector(D_C*2-1 downto 0);	-- "Σ~z(t)" (high resolution accumulator)
		high_res_accu_o	 : out std_logic_vector(D_C*2-1 downto 0);	-- "Σ~z(t)" (high resolution accumulator)
		counter_i		 : in  std_logic_vector(D_C-1 downto 0);	-- "Γ(t)"   (counter)
		reconst_bit_i	 : in  std_logic;							-- Lets decoder reconstruct "Σ~z(t)" values
		
		entropy_select_i : in  std_logic;
		entropy_select_o : out std_logic;

		enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		rvd_codeword_o	 : out std_logic_vector(Umax_C+D_C-1 downto 0) -- "R'k(j)" (reversed codeword)
	);
end entity hybrid_high_entr_code;

architecture behavioural of hybrid_high_entr_code is
	constant CONST_49_C		 : unsigned(D_C-1 downto 0)						:= to_unsigned(49, D_C);
	signal numerator_s		 : std_logic_vector(D_C*2-1 downto 0)			:= (others => '0');
	
	signal enable1_s		 : std_logic	:= '0';
	signal enable2_s		 : std_logic	:= '0';
	signal img_coord1_s		 : img_coord_t	:= reset_img_coord(SMPL_ORDER_G);
	signal img_coord2_s		 : img_coord_t	:= reset_img_coord(SMPL_ORDER_G);
	
	signal high_res_accu1_s	 : std_logic_vector(D_C*2-1 downto 0)			:= (others => '0');
	signal high_res_accu2_s	 : std_logic_vector(D_C*2-1 downto 0)			:= (others => '0');
	signal entropy_select1_s : std_logic	:= '1';
	signal entropy_select2_s : std_logic	:= '1';
	signal reconst_bit_s	 : std_logic	:= '0';
	
	signal err_lim_cnt_s	 : integer range 0 to 2*NZ_C					:= 0;
	signal err_lim_upd_en_s  : std_logic := '0';
	signal err_lim_val_s	 : std_logic_vector(D_C-1 downto 0)				:= (others => '0');
	signal data_mp_quan_s	 : std_logic_vector(D_C-1 downto 0)				:= (others => '0');
	signal counter_s		 : std_logic_vector(D_C-1 downto 0)				:= (others => '0');
	
	signal kz_s				 : integer range 0 to D_C						:= 0;
	signal enc_num_bits_s	 : std_logic_vector(MAX_NUM_BITS_G-1 downto 0)	:= (others => '0');
	signal rvd_codeword_s	 : std_logic_vector(Umax_C+D_C-1 downto 0)		:= (others => '0');
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_hyb_high_entr_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable1_s		  <= '0';
				enable2_s		  <= '0';
				img_coord1_s	  <= reset_img_coord(SMPL_ORDER_G);
				img_coord2_s	  <= reset_img_coord(SMPL_ORDER_G);
				high_res_accu1_s  <= (others => '0');
				high_res_accu2_s  <= (others => '0');
				entropy_select1_s <= '1';
				entropy_select2_s <= '1';
				reconst_bit_s	  <= '0';
				err_lim_upd_en_s  <= '0';
				data_mp_quan_s	  <= (others => '0');
				counter_s		  <= (others => '0');
				err_lim_val_s	  <= (others => '0');
			else
				enable1_s		  <= enable_i;
				enable2_s		  <= enable1_s;
				img_coord1_s	  <= img_coord_i;
				img_coord2_s	  <= img_coord1_s;
				high_res_accu1_s  <= high_res_accu_i;
				high_res_accu2_s  <= high_res_accu1_s;
				entropy_select1_s <= entropy_select_i;
				entropy_select2_s <= entropy_select1_s;
				reconst_bit_s	  <= reconst_bit_i;
				err_lim_upd_en_s  <= err_lim_upd_en_i;
				data_mp_quan_s	  <= data_mp_quan_i;
				counter_s		  <= counter_i;
				err_lim_val_s	  <= err_lim_val_i;
			end if;
		end if;
	end process p_hyb_high_entr_delay;
	
	-- Calculation of "kz(t)" (Variable length code parameter)
	p_calc_kz : process(clock_i) is
		variable comp1_v : std_logic_vector(D_C*2-1 downto 0) := (others => '0');
		variable total_v : std_logic_vector(D_C*2-1 downto 0) := (others => '0');
		variable log2_v	 : integer range -2 to D_C			  := 0;
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_v		:= (others => '0');
				total_v		:= (others => '0');
				log2_v		:= 0;
				numerator_s <= (others => '0');
				kz_s		<= 0;
			else
				if (enable_i = '1' and err_lim_upd_en_i = '0') then
					-- NOTE: The "t>0" check must NOT be added here, as it is already integrated when computing the operands of these operations...
					comp1_v		:= std_logic_vector(resize(shift_usg(resize(CONST_49_C * unsigned(counter_i), D_C), -5), D_C*2));
					-- As an additional delay is required, this is not a variable anymore, but a signal instead
					numerator_s <= std_logic_vector(resize(unsigned(high_res_accu_i) + unsigned(comp1_v), D_C*2));
					
					total_v		:= std_logic_vector(resize(unsigned(numerator_s) / resize(unsigned(counter_s), D_C*2), D_C*2));
					-- To ensure no negative values, this function rounds up automatically
					log2_v		:= log2(to_integer(unsigned(total_v)) + 1) - 2;		-- '+1' to deal with the case of exactly matching 2-multiple value...
					
					-- The previous value must be constrained within [0, max(D-2, 2)]
					kz_s		<= iif(log2_v <= 0, 0, iif(log2_v > max(D_C-2, 2), max(D_C-2, 2), log2_v-1));	-- '-1' to remove the round up from previous 'log2' function
					-- NOTE: For high-entropy samples, it can be seen that kz(t) ≥ 2
				end if;
			end if;
		end if;
	end process p_calc_kz;

	-- Calculation of "R'k(j)" (Reversed Length-Limited GPO2 Codeword of Mapped Quantizer Index)
	p_calc_high_codeword : process(clock_i) is
		variable var_length_v	: integer range 0 to 2**D_C := 0;		-- No need to be a "std_logic_vector"
		variable err_lim_type_v	: std_logic	:= '0';
		variable rvd_codeword_v	: std_logic_vector(Umax_C+D_C-1 downto 0) := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				var_length_v   := 0;
				err_lim_type_v := '0';
				rvd_codeword_v := (others => '0');
				err_lim_cnt_s  <= 0;
				enc_num_bits_s <= (others => '0');
				rvd_codeword_s <= (others => '0');
			else
				-- To be sure the packer does not produce problems after this IP, output data is restarted when fully processed...
				if (img_coord2_s = end_img_coord) then
					enc_num_bits_s <= (others => '0');
					rvd_codeword_v := (others => '0');
				end if;
				
				if (enable1_s = '1') then
					rvd_codeword_v := (others => '0');
					
					if (err_lim_upd_en_s = '0') then		-- Mapped quantizer indexes ("?z(t)") are encoded here!
						err_lim_cnt_s						<= 0;
						if (img_coord1_s.t = 0) then		-- Sample for t=0 must not be encoded!
							enc_num_bits_s					<= std_logic_vector(to_unsigned(D_C, MAX_NUM_BITS_G));
							rvd_codeword_v					:= data_mp_quan_s & (Umax_C-1 downto 0 => '0');
						else	-- img_coord1_s.t > 0
							var_length_v					:= to_integer(shift_usg(unsigned(data_mp_quan_s), -kz_s));
							if (var_length_v < Umax_C) then
								-- 5.4.3.3.5.1.2: When code selection statistics rescaled, LSb from "high_res_accu(t-1) -> here high_res_accu1_s" is encoded in bitstream before anything else
								if (reconst_bit_s = '1') then
									enc_num_bits_s			<= std_logic_vector(to_unsigned(var_length_v + 1 + kz_s + 1, MAX_NUM_BITS_G));
									rvd_codeword_v(kz_s+2-1 downto 0) := data_mp_quan_s(kz_s-1 downto 0) & '1' & high_res_accu1_s(0);
								else	-- reconst_bit_s = '0'
									enc_num_bits_s			<= std_logic_vector(to_unsigned(var_length_v + 1 + kz_s, MAX_NUM_BITS_G));
									rvd_codeword_v(kz_s+1-1 downto 0) := data_mp_quan_s(kz_s-1 downto 0) & '1';
								end if;
								rvd_codeword_v				:= rvd_codeword_v sll var_length_v;
							else	-- var_length_v >= Umax_C
								enc_num_bits_s				<= std_logic_vector(to_unsigned(Umax_C + D_C, MAX_NUM_BITS_G));
								rvd_codeword_v(Umax_C+D_C-1 downto Umax_C) := data_mp_quan_s;
								-- 5.4.3.3.5.1.2: When code selection statistics rescaled, LSb from "high_res_accu(t-1) -> here high_res_accu1_s" is encoded in bitstream before anything else
								if (reconst_bit_s = '1') then
									rvd_codeword_v(Umax_C-1) := high_res_accu1_s(0);
								end if;
							end if;
						end if;
					
					else	-- err_lim_upd_en_s = '1'		-- Error limit values are encoded here!
						if (FIDEL_CTRL_TYPE_G = ABS_ERR_C) then
							enc_num_bits_s					<= std_logic_vector(to_unsigned(DA_C, MAX_NUM_BITS_G));
							rvd_codeword_v(DA_C-1 downto 0)	:= err_lim_val_s(DA_C-1 downto 0);
						elsif (FIDEL_CTRL_TYPE_G = REL_ERR_C) then
							enc_num_bits_s					<= std_logic_vector(to_unsigned(DR_C, MAX_NUM_BITS_G));
							rvd_codeword_v(DR_C-1 downto 0)	:= err_lim_val_s(DR_C-1 downto 0);
						elsif (FIDEL_CTRL_TYPE_G = ABS_REL_C) then
							if ((ABS_ERR_BAND_TYPE_G = '1' and err_lim_cnt_s = 0) or (ABS_ERR_BAND_TYPE_G = '0' and err_lim_cnt_s < NZ_C)) then
								err_lim_type_v				:= '1';			-- Absolute error limit values (always the first one!)
							else
								err_lim_type_v				:= '0';			-- Relative error limit values (always the second one!)
							end if;
							err_lim_cnt_s					<= err_lim_cnt_s + 1;
							
							if (err_lim_type_v = '1') then	-- Time to output "absolute error limit values"
								enc_num_bits_s				<= std_logic_vector(to_unsigned(DA_C, MAX_NUM_BITS_G));
								rvd_codeword_v(DA_C-1 downto 0)	:= err_lim_val_s(DA_C-1 downto 0);
							else	-- err_lim_type_v = '0'	-- Time to output "relative error limit values"
								enc_num_bits_s				<= std_logic_vector(to_unsigned(DR_C, MAX_NUM_BITS_G));
								rvd_codeword_v(DR_C-1 downto 0)	:= err_lim_val_s(DR_C-1 downto 0);
							end if;
						else	-- FIDEL_CTRL_TYPE_G = LOSSLESS_C		-- This case can never happen (as "err_lim_upd_en_s" would be always '0')
							enc_num_bits_s					<= (others => '0');
							rvd_codeword_v					:= (others => '0');
						end if;
					end if;
					
					rvd_codeword_s <= rvd_codeword_v;
				end if;
			end if;
		end if;
	end process p_calc_high_codeword;

	-- Outputs
	enable_o		  <= enable2_s;
	img_coord_o		  <= img_coord2_s;
	high_res_accu_o	  <= high_res_accu2_s;
	entropy_select_o  <= entropy_select2_s;
	enc_num_bits_o	  <= enc_num_bits_s;
	rvd_codeword_o	  <= rvd_codeword_s;
end architecture behavioural;