--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		18/04/2021
--------------------------------------------------------------------------------
-- IP name:		hybrid_wrapper
--
-- Description: Top entity for the Hybrid Entropy Coder
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;
use work.comp_encoder.all;

entity hybrid_wrapper is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- Maximum number of bits (size) of every encoded data sample
		MAX_NUM_BITS_G		: integer;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic
	);
	port (
		clock_i			 : in  std_logic;
		reset_i			 : in  std_logic;
		
		enable_i		 : in  std_logic;
		enable_o		 : out std_logic;
		img_coord_i		 : in  img_coord_t;
		img_coord_o		 : out img_coord_t;
		
		err_lim_upd_en_i : in  std_logic;
		err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
		
		enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		rvd_codeword_o	 : out std_logic_vector(Umax_C+D_C-1 downto 0)		-- "R'k(j)" (reversed codeword)
	);
end entity hybrid_wrapper;

architecture behavioural of hybrid_wrapper is
	signal enable1_s			: std_logic   := '0';
	signal enable3_high_s		: std_logic   := '0';
	signal enable3_low_s		: std_logic   := '0';
	signal enable3_s			: std_logic   := '0';
	signal enable4_s			: std_logic   := '0';
	signal img_coord1_s			: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord3_high_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord3_low_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord3_s			: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord4_s			: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal err_lim_upd_en_s		: std_logic	  := '0';
	signal err_lim_val_s		: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	signal data_mp_quan_s		: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	
	signal high_res_accu1_s		: std_logic_vector(D_C*2-1 downto 0)		  := (others => '0');
	signal high_res_accu3_s		: std_logic_vector(D_C*2-1 downto 0)		  := (others => '0');
	signal counter_s			: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	signal reconst_bit_s		: std_logic	  := '0';
	signal curr_prefix_arr_s	: array_string_t(0 to NUM_HYBRID_TABLES_G-1)(1 to LONGEST_STRING_SIZE_G) := (others => (others => ' '));
	
	signal entropy_select1_s	: std_logic	  := '1';
	signal entropy_select3_s	: std_logic	  := '1';

	signal enc_num_bits_high_s	: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal enc_num_bits_low_s	: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal enc_num_bits_mux_s	: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal enc_num_bits_s		: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal rvd_codeword_high_s	: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	signal rvd_codeword_low_s	: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	signal rvd_codeword_mux_s	: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	signal rvd_codeword_s		: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');

begin
	i_hybrid_statistic : hybrid_statistic
	generic map(
		SMPL_ORDER_G	  => SMPL_ORDER_G
	)
	port map(
		clock_i			  => clock_i,
		reset_i			  => reset_i,
		
		enable_i		  => enable_i,
		enable_o		  => enable1_s,
		img_coord_i		  => img_coord_i,
		img_coord_o		  => img_coord1_s,
		
		err_lim_upd_en_i  => err_lim_upd_en_i,
		err_lim_upd_en_o  => err_lim_upd_en_s,
		err_lim_val_i	  => err_lim_val_i,
		err_lim_val_o	  => err_lim_val_s,
		data_mp_quan_i	  => data_mp_quan_i,
		data_mp_quan_o	  => data_mp_quan_s,
		
		high_res_accu_o	  => high_res_accu1_s,
		counter_o		  => counter_s,
		reconst_bit_o	  => reconst_bit_s,
		
		entropy_select_o  => entropy_select1_s
	);
	
	i_hybrid_high_entr : hybrid_high_entr_code
	generic map(
		SMPL_ORDER_G		=> SMPL_ORDER_G,
		MAX_NUM_BITS_G		=> MAX_NUM_BITS_G,
		FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G
	)
	port map(
		clock_i			  => clock_i,
		reset_i			  => reset_i,

		enable_i		  => enable1_s,
		enable_o		  => enable3_high_s,
		img_coord_i		  => img_coord1_s,
		img_coord_o		  => img_coord3_high_s,
		
		err_lim_upd_en_i  => err_lim_upd_en_s,
		err_lim_val_i	  => err_lim_val_s,
		data_mp_quan_i	  => data_mp_quan_s,
		
		high_res_accu_i	  => high_res_accu1_s,
		high_res_accu_o	  => high_res_accu3_s,
		counter_i		  => counter_s,
		reconst_bit_i	  => reconst_bit_s,
		
		entropy_select_i  => entropy_select1_s,
		entropy_select_o  => entropy_select3_s,
		
		enc_num_bits_o	  => enc_num_bits_high_s,
		rvd_codeword_o	  => rvd_codeword_high_s
	);

	i_hybrid_low_entr : hybrid_low_entr_code
	generic map(
		SMPL_ORDER_G		=> SMPL_ORDER_G,
		MAX_NUM_BITS_G		=> MAX_NUM_BITS_G,
		FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G
	)
	port map(
		clock_i			  => clock_i,
		reset_i			  => reset_i,

		enable_i		  => enable1_s,
		enable_o		  => enable3_low_s,
		img_coord_i		  => img_coord1_s,
		img_coord_o		  => img_coord3_low_s,
		
		err_lim_upd_en_i  => err_lim_upd_en_s,
		err_lim_val_i	  => err_lim_val_s,
		data_mp_quan_i	  => data_mp_quan_s,
		
		high_res_accu_i	  => high_res_accu1_s,
		counter_i		  => counter_s,
		reconst_bit_i	  => reconst_bit_s,
		
		curr_prefix_arr_o => curr_prefix_arr_s,
		enc_num_bits_o	  => enc_num_bits_low_s,
		rvd_codeword_o	  => rvd_codeword_low_s
	);
	
	-- Multiplexer to select output signals from 'high/low entropy code IP's and feed them into 'image tail IP'
	enable3_s			  <= enable3_high_s			when (entropy_select3_s = '1') else enable3_low_s;
	img_coord3_s		  <= img_coord3_high_s		when (entropy_select3_s = '1') else img_coord3_low_s;
	enc_num_bits_mux_s	  <= enc_num_bits_high_s	when (entropy_select3_s = '1') else enc_num_bits_low_s;
	rvd_codeword_mux_s	  <= rvd_codeword_high_s	when (entropy_select3_s = '1') else rvd_codeword_low_s;
	
	i_hybrid_img_tail : hybrid_compr_img_tail
	generic map(
		SMPL_ORDER_G	  => SMPL_ORDER_G,
		MAX_NUM_BITS_G	  => MAX_NUM_BITS_G
	)
	port map(
		clock_i			  => clock_i,
		reset_i			  => reset_i,
		
		enable_i		  => enable3_s,
		enable_o		  => enable4_s,
		img_coord_i		  => img_coord3_s,
		img_coord_o		  => img_coord4_s,
		
		high_res_accu_i	  => high_res_accu3_s,
		curr_prefix_arr_i => curr_prefix_arr_s,
		
		enc_num_bits_i	  => enc_num_bits_mux_s,
		rvd_codeword_i	  => rvd_codeword_mux_s,

		out_num_bits_o	  => enc_num_bits_s,
		out_codeword_o	  => rvd_codeword_s
	);

	-- Outputs
	enable_o		<= enable4_s;
	img_coord_o		<= img_coord4_s;
	enc_num_bits_o	<= enc_num_bits_s;
	rvd_codeword_o	<= rvd_codeword_s;
end architecture behavioural;