--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		07/08/2021
--------------------------------------------------------------------------------
-- IP name:		tb_hybrid_wrapper
--
-- Description: Testbench for the "Hybrid Encoder" module
--
--
--------------------------------------------------------------------------------

-- Library instantiated, then declaration of files/functions to use
library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

-- A reference to VUnit (and related) set of libraries and packages
library vunit_lib;
context vunit_lib.vunit_context;
context vunit_lib.com_context;
context vunit_lib.data_types_context;

-- Components and packages related to the Image IP
use vunit_lib.param_image.all;
use vunit_lib.types_image.all;
use vunit_lib.utils_image.all;
use vunit_lib.comp_image.all;

-- Components and packages related to the Encoder IP
use vunit_lib.param_encoder.all;
use vunit_lib.types_encoder.all;
use vunit_lib.utils_encoder.all;
use vunit_lib.comp_encoder.all;

entity tb_hybrid_wrapper is
    generic (
        encoded_tb_cfg	: string;
        runner_cfg		: string
    );
end entity tb_hybrid_wrapper;

architecture behavioural of tb_hybrid_wrapper is
	-- Record type to pack all signals (from Python script) together
	type tb_cfg_t is record
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic;
	end record tb_cfg_t;

	-- Function to decode the Python signals and connect them into the VHDL testbench
	impure function decode(encoded_tb_cfg : string) return tb_cfg_t is
	begin
	return (
		SMPL_ORDER_G		=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "SMPL_ORDER_PY")), 2)),
		FIDEL_CTRL_TYPE_G	=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "FIDEL_CTRL_TYPE_PY")), 2)),
		ABS_ERR_BAND_TYPE_G => std_logic'value(get(encoded_tb_cfg, "ABS_ERR_BAND_TYPE_PY")),
		REL_ERR_BAND_TYPE_G => std_logic'value(get(encoded_tb_cfg, "REL_ERR_BAND_TYPE_PY"))
	);
	end function decode;

	-- Constant of type "tb_cfg_t" and initialized by "decode()" function
	constant tb_cfg			: tb_cfg_t	:= decode(encoded_tb_cfg);
	-- Maximum number of bits (size) of every encoded data sample
	constant MAX_NUM_BITS_C	: integer	:= log2(Umax_C+D_C);
	
	signal clock_s			: std_logic := '0';
	signal reset_s			: std_logic := '1';
	
	signal enable_in_s		: std_logic := '0';
	signal enable_out_s		: std_logic := '0';
	signal img_coord_in_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_out_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	
	signal err_lim_upd_en_s	: std_logic	  := '0';
	signal err_lim_val_s	: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	signal data_mp_quan_s	: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	
	signal enc_num_bits_s	: std_logic_vector(MAX_NUM_BITS_C-1 downto 0) := (others => '0');
	signal rvd_codeword_s	: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0'); 	-- "R'k(j)" (reversed codeword)

	signal flag_start, flag_stop : std_logic := '0';

begin
	reset_s <= '0' after 20 ns;
	clock_s <= not clock_s after 6.4 ns;	 -- Main clock frequency: 78125000 Hz
	
	-- Simulation MAIN process (if this one finishes, the whole simulation too)
	test_runner : process is
	begin
		-- Process starts by setting up VUnit using test_runner_setup procedure
		test_runner_setup(runner, runner_cfg);

		while test_suite loop	-- Pay very attention to the testcase name, arrows or accents prevent the system to work...
			if run("Hybrid Encoder Wrapper") then
				info("Running test case = " & to_string(running_test_case));

				flag_start <= '1';
			end if;
			
			wait until (flag_stop = '1');
		end loop;

		-- The process ends with the test_runner_cleanup procedure which will force the simulation to stop
		test_runner_cleanup(runner);
	end process;

	-- Set Watchdog with limited time (this number must be higher as any "wait for" statement used in the testbench)
	test_runner_watchdog(runner, 1 ms);

	-- Input signals stimulus calculation
	p_mp_quan_update : process(clock_s) is
	begin
		if (rising_edge(clock_s)) then
			if (reset_s = '1') then
				err_lim_upd_en_s <= '0';
				err_lim_val_s	 <= (others => '0');
				data_mp_quan_s	 <= (others => '0');
			else
				if (flag_start = '1') then
					if ((img_coord_out_s.x < NX_C-1) or (img_coord_out_s.y < NY_C-1) or (img_coord_out_s.z < NZ_C-1)) then
						if (data_mp_quan_s = (data_mp_quan_s'length-1 downto 0 => '1')) then
							data_mp_quan_s <= (others => '0');		
						else
							data_mp_quan_s <= std_logic_vector(resize(unsigned(data_mp_quan_s) + to_unsigned(1, data_mp_quan_s'length), data_mp_quan_s'length));
						end if;
						
						if ((img_coord_out_s.x >= 4) and (img_coord_out_s.x <= 5)) then
							err_lim_upd_en_s <= '1';
							err_lim_val_s	 <= std_logic_vector(resize(unsigned(err_lim_val_s) + to_unsigned(10, err_lim_val_s'length), err_lim_val_s'length));
						else
							err_lim_upd_en_s <= '0';
						end if;
					end if;
				end if;
			end if;	
		end if;
	end process p_mp_quan_update;
	
	-- Process to stop the simulation when finished (to use 'wait' statements, the process cannot be clocked)
	p_stop_sim : process is
	begin
		-- After one complete image, wait statement and request to finish the simulation
		wait until ((img_coord_out_s.x >= NX_C-1) and (img_coord_out_s.y >= NY_C-1) and (img_coord_out_s.z >= NZ_C-1));
		wait for 10 ns;
		flag_stop <= '1';
		
		-- Two possibilities to stop the simulation without VUnit framework (only the "assert" line really works, anyway)
		-- assert FALSE Report "Simulation Finished" severity FAILURE;
		-- std.env.finish;
	end process p_stop_sim;	
	
	-- Entity to control the image coordinates
	i_image_coord_control : image_coord_control
	generic map(
		SMPL_ORDER_G	=> tb_cfg.SMPL_ORDER_G
	)
	port map(
		clock_i			=> clock_s,
		reset_i			=> reset_s,

		enable_i		=> flag_start,
		enable_o		=> enable_in_s,
		
		image_end_i		=> '0',
		img_coord_o		=> img_coord_in_s
	);

	-- Hybrid Encoder top entity
	i_hybrid_wrapper : hybrid_wrapper
	generic map(
		SMPL_ORDER_G		=> tb_cfg.SMPL_ORDER_G,
		MAX_NUM_BITS_G		=> MAX_NUM_BITS_C,
		FIDEL_CTRL_TYPE_G	=> tb_cfg.FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G	=> tb_cfg.ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> tb_cfg.REL_ERR_BAND_TYPE_G
	)
	port map(
		clock_i			 => clock_s,
		reset_i			 => reset_s,
		
		enable_i		 => enable_in_s,
		enable_o		 => enable_out_s,
		img_coord_i		 => img_coord_in_s,
		img_coord_o		 => img_coord_out_s,
		
		err_lim_upd_en_i => err_lim_upd_en_s,
		err_lim_val_i	 => err_lim_val_s,
		data_mp_quan_i	 => data_mp_quan_s,
		
		enc_num_bits_o	 => enc_num_bits_s,
		rvd_codeword_o	 => rvd_codeword_s
	);

end architecture behavioural;