--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		05/04/2021
--------------------------------------------------------------------------------
-- IP name:		smpl_adap_statistic
--
-- Description: The adaptive code selection statistics used to select the
--				codeword for each mapped quantizer index.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.utils_predictor.all;
use work.comp_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;

entity smpl_adap_statistic is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	 : std_logic_vector(1 downto 0)
	);
	port (
		clock_i			 : in  std_logic;
		reset_i			 : in  std_logic;
		
		enable_i		 : in  std_logic;
		enable_o		 : out std_logic;
		img_coord_i		 : in  img_coord_t;
		img_coord_o		 : out img_coord_t;
		
		err_lim_upd_en_i : in  std_logic;
		err_lim_upd_en_o : out std_logic;
		err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
		err_lim_val_o	 : out std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_o	 : out std_logic_vector(D_C-1 downto 0);
		
		accu_o			 : out std_logic_vector(D_C*2-1 downto 0);	-- "Σz(t)" (accumulator)
		counter_o		 : out std_logic_vector(D_C-1 downto 0)		-- "Γ(t)"  (counter)
	);
end entity smpl_adap_statistic;

architecture behavioural of smpl_adap_statistic is
	-- Accumulator and counter initial values
	constant COUNTER_INIT_C	: std_logic_vector(D_C-1 downto 0)	 := std_logic_vector(to_unsigned(2**Yo_C, D_C));
	constant ACCU_INIT_C	: std_logic_vector(D_C*2-1 downto 0) := std_logic_vector(to_unsigned((to_integer(unsigned(COUNTER_INIT_C)) * (3 * 2**(K_C + 6) - 49)) / 2**7, D_C*2));
	-- Number of delay samples to get t-1 (or W position), for BSQ/BIP/BIL orders (delays table defined inside "samples_store" IP)
	constant NEXT_W_C		: integer	  := locate_position(SMPL_ORDER_G, 1, NZ_C, 1);
	constant NEXT_SAMPLE_C	: integer	  := 1;
	
	signal enable_s			: std_logic   := '0';
	signal img_coord_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal k1_s				: integer	  := K_C;
	
	signal err_lim_upd_en_s	: std_logic	  := '0';
	signal err_lim_val_s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_mp_quan_s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	
	signal accu_s			: std_logic_vector(D_C*2-1 downto 0) := (others => '0');
	signal accu_old_s		: std_logic_vector(D_C*2-1 downto 0) := (others => '0');
	signal counter_s		: std_logic_vector(D_C-1 downto 0)	 := COUNTER_INIT_C;
	signal counter_old_s	: std_logic_vector(D_C-1 downto 0)	 := COUNTER_INIT_C;
	-- "accu_s" and "counter_s" are updated AFTER coding each input sample (so "?z(t-1)" is used)
	
begin
	-- Shift register to compute previous "err_lim_upd_en_i" signal, independently from the samples order
	i_err_lim_upd_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => 1,
		REG_SIZE_G	 => NEXT_SAMPLE_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => enable_s,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => img_coord_s,

		data_i(0)	 => err_lim_upd_en_i,
		data_o(0)	 => err_lim_upd_en_s
	);
	
	-- Shift register to compute previous "err_lim_val_i" signal, independently from the samples order
	i_err_lim_val_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => NEXT_SAMPLE_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => err_lim_val_i,
		data_o		 => err_lim_val_s
	);
	
	-- Shift register to compute previous "data_mp_quan_i" signal, independently from the samples order
	i_mp_quan_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => NEXT_SAMPLE_C,
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => data_mp_quan_i,
		data_o		 => data_mp_quan_s
	);

	-- Shift register to compute previous "counter_s" signal, independently from the samples order
	i_cnt_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C,
		REG_SIZE_G	 => NEXT_SAMPLE_C,
		INIT_VAL_G	 => COUNTER_INIT_C
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => counter_s,
		data_o		 => counter_old_s
	);

	-- Shift register to compute "t-1" from "accu_s" signal, depending on the samples order
	i_accu_prev_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => D_C*2,
		REG_SIZE_G	 => iif(NEXT_W_C <= 2, NEXT_W_C+1, NEXT_W_C-1),
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => accu_s,
		data_o		 => accu_old_s
	);
	
	-- Calculation of accumulator Σz(t) and counter Γ(t)
	p_calc_accu_cnt : process(clock_i) is
		variable k1_v	 : integer := K_C;
		variable comp1_v : std_logic_vector(D_C*2-1 downto 0) := (others => '0');
		variable comp2_v : std_logic_vector(D_C-1 downto 0)	  := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				k1_v	  := K_C;
				k1_s	  <= K_C;
				comp1_v	  := (others => '0');
				comp2_v	  := (others => '0');
				accu_s	  <= (others => '0');
				counter_s <= COUNTER_INIT_C;
			else	-- Apart from "enable_i", this IP does nothing while "error limit values" are being encoded
				if (enable_i = '1' and err_lim_upd_en_i = '0') then
					if (img_coord_i.t = 0) then			-- Initial values (requires "k'z" calculation!)
						k1_v	   := iif(K2_ARR_C(img_coord_i.z) <= 30-D_C, K2_ARR_C(img_coord_i.z), 2*K2_ARR_C(img_coord_i.z)+D_C-30);
						k1_s	   <= k1_v;
						accu_s	   <= std_logic_vector(to_unsigned((to_integer(unsigned(COUNTER_INIT_C)) * (3 * 2**(k1_v + 6) - 49)) / 2**7, D_C*2));
						counter_s  <= COUNTER_INIT_C;
					else 	-- img_coord_i.t > 0
						-- "accu_s" and "counter_s" are updated AFTER coding each input sample (so "?z(t-1)" is used -> "data_mp_quan_i")
						comp1_v	   := std_logic_vector(resize(unsigned(accu_old_s) + resize(unsigned(data_mp_quan_i), D_C*2), D_C*2));
						comp2_v	   := std_logic_vector(resize(unsigned(counter_old_s) + to_unsigned(1, D_C), D_C));
						if (unsigned(counter_old_s) < to_unsigned(2**Y_C-1, D_C)) then
							accu_s <= comp1_v;
							-- Counter is updated only at the end of the spectral band...
							if (img_coord_i.z >= NZ_C-1) then
								counter_s <= comp2_v;
							end if;
						else	-- This can only mean "counter_old_s = 2**Y_C-1", the rescaling (5.4.3.2.3.4)
							accu_s <= std_logic_vector(shift_usg(resize(unsigned(comp1_v) + to_unsigned(1, D_C*2), D_C*2), -1));
							-- Counter is updated only at the end of the spectral band...
							if (img_coord_i.z >= NZ_C-1) then
								counter_s <= std_logic_vector(shift_usg(unsigned(comp2_v), -1));
							end if;
						end if;
					end if;
				end if;
			end if;
		end if;
	end process p_calc_accu_cnt;
	
	-- Outputs
	enable_o		 <= enable_s;
	img_coord_o		 <= img_coord_s;
	err_lim_upd_en_o <= err_lim_upd_en_s;
	err_lim_val_o	 <= err_lim_val_s;
	data_mp_quan_o	 <= data_mp_quan_s;
	accu_o			 <= accu_old_s;
	counter_o		 <= counter_s;
end architecture behavioural;