--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		05/04/2021
--------------------------------------------------------------------------------
-- IP name:		smpl_adap_wrapper
--
-- Description: Top entity for the Sample-Adaptive Entropy Coder
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;
use work.comp_encoder.all;

entity smpl_adap_wrapper is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- Maximum number of bits (size) of every encoded data sample
		MAX_NUM_BITS_G		: integer;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic
	);
	port (
		clock_i			 : in  std_logic;
		reset_i			 : in  std_logic;

		enable_i		 : in  std_logic;
		enable_o		 : out std_logic;
		img_coord_i		 : in  img_coord_t;
		img_coord_o		 : out img_coord_t;
		
		err_lim_upd_en_i : in  std_logic;
		err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
		
		enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		codeword_o		 : out std_logic_vector(Umax_C+D_C-1 downto 0)	-- "Rk(j)" (codeword)
	);
end entity smpl_adap_wrapper;

architecture behavioural of smpl_adap_wrapper is
	signal enable1_s		: std_logic   := '0';
	signal enable2_s		: std_logic   := '0';
	signal img_coord1_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord2_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal err_lim_upd_en_s	: std_logic	  := '0';
	signal err_lim_val_s	: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	signal data_mp_quan_s	: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	
	signal accu_s			: std_logic_vector(D_C*2-1 downto 0)		  := (others => '0');
	signal counter_s		: std_logic_vector(D_C-1 downto 0)			  := (others => '0');
	signal enc_num_bits_s	: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal codeword_s		: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	
begin
	i_smpl_adap_statistic : smpl_adap_statistic
	generic map(
		SMPL_ORDER_G	 => SMPL_ORDER_G
	)
	port map(
		clock_i			 => clock_i,
		reset_i			 => reset_i,		
		
		enable_i		 => enable_i,
		enable_o		 => enable1_s,
		img_coord_i		 => img_coord_i,
		img_coord_o		 => img_coord1_s,
		
		err_lim_upd_en_i => err_lim_upd_en_i,
		err_lim_upd_en_o => err_lim_upd_en_s,
		err_lim_val_i	 => err_lim_val_i,
		err_lim_val_o	 => err_lim_val_s,
		data_mp_quan_i	 => data_mp_quan_i,
		data_mp_quan_o	 => data_mp_quan_s,
		
		accu_o			 => accu_s,
		counter_o		 => counter_s
	);

	i_smpl_adap_coding : smpl_adap_gpo2_code
	generic map(
		SMPL_ORDER_G		=> SMPL_ORDER_G,
		MAX_NUM_BITS_G		=> MAX_NUM_BITS_G,
		FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G
	)
	port map(
		clock_i			 => clock_i,
		reset_i			 => reset_i,
		
		enable_i		 => enable1_s,
		enable_o		 => enable2_s,
		img_coord_i		 => img_coord1_s,
		img_coord_o		 => img_coord2_s,
		
		err_lim_upd_en_i => err_lim_upd_en_s,
		err_lim_val_i	 => err_lim_val_s,
		data_mp_quan_i	 => data_mp_quan_s,
		
		accu_i			 => accu_s,
		counter_i		 => counter_s,
		
		enc_num_bits_o	 => enc_num_bits_s,
		codeword_o		 => codeword_s
	);

	-- Outputs
	enable_o		<= enable2_s;
	img_coord_o		<= img_coord2_s;
	enc_num_bits_o	<= enc_num_bits_s;
	codeword_o		<= codeword_s;
end architecture behavioural;