--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		09/06/2021
--------------------------------------------------------------------------------
-- IP name:		top_enc_body
--
-- Description: Defines the body part of the compressed image
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;
use work.comp_encoder.all;

entity top_enc_body is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- 00: Sample-Adaptive Entropy, 01: Hybrid Entropy, 10: Block-Adaptive Entropy
		ENCODER_TYPE_G		: std_logic_vector(1 downto 0);
		-- Maximum number of bits (size) of every encoded data sample
		MAX_NUM_BITS_G		: integer;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic
	);
	port (
		clock_i			 : in  std_logic;
		reset_i			 : in  std_logic;
		
		enable_i		 : in  std_logic;
		enable_o		 : out std_logic;
		img_coord_i		 : in  img_coord_t;
		img_coord_o		 : out img_coord_t;

		err_lim_upd_en_i : in  std_logic;
		err_lim_val_i	 : in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i	 : in  std_logic_vector(D_C-1 downto 0);
		
		enc_num_bits_o	 : out std_logic_vector(MAX_NUM_BITS_G-1 downto 0);
		enc_body_data_o	 : out std_logic_vector(Umax_C+D_C-1 downto 0)
	);
end entity top_enc_body;

architecture behavioural of top_enc_body is	
	signal enable_s			: std_logic   := '0';
	signal img_coord_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal enc_num_bits_s	: std_logic_vector(MAX_NUM_BITS_G-1 downto 0) := (others => '0');
	signal enc_body_data_s	: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');

begin
	g_enc_body_type : case ENCODER_TYPE_G generate
		when SMPL_ADAPT_C =>
			i_smpl_adap_wrapper : smpl_adap_wrapper
			generic map(
				SMPL_ORDER_G		=> SMPL_ORDER_G,
				MAX_NUM_BITS_G		=> MAX_NUM_BITS_G,
				FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
				ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
				REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G
			)
			port map(
				clock_i			 => clock_i,
				reset_i			 => reset_i,

				enable_i		 => enable_i,
				enable_o		 => enable_s,
				img_coord_i		 => img_coord_i,
				img_coord_o		 => img_coord_s,

				err_lim_upd_en_i => err_lim_upd_en_i,
				err_lim_val_i	 => err_lim_val_i,
				data_mp_quan_i	 => data_mp_quan_i,
		
				enc_num_bits_o	 => enc_num_bits_s,
				codeword_o		 => enc_body_data_s
			);

		when HYBRID_C =>
			i_hybrid_wrapper : hybrid_wrapper
			generic map(
				SMPL_ORDER_G		=> SMPL_ORDER_G,
				MAX_NUM_BITS_G		=> MAX_NUM_BITS_G,
				FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
				ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
				REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G
			)
			port map(
				clock_i			 => clock_i,
				reset_i			 => reset_i,

				enable_i		 => enable_i,
				enable_o		 => enable_s,
				img_coord_i		 => img_coord_i,
				img_coord_o		 => img_coord_s,

				err_lim_upd_en_i => err_lim_upd_en_i,
				err_lim_val_i	 => err_lim_val_i,
				data_mp_quan_i	 => data_mp_quan_i,
				
				enc_num_bits_o	 => enc_num_bits_s,
				rvd_codeword_o	 => enc_body_data_s
			);
		
		-- when BLK_ADAPT_C =>
			-- To be implemented!
		
		when others =>		-- No encoder type selected!
			enable_s		<= '0';
			img_coord_s		<= reset_img_coord(SMPL_ORDER_G);
			enc_num_bits_s	<= (others => '0');
			enc_body_data_s	<= (others => '0');
			
	end generate g_enc_body_type;

	-- Outputs
	enable_o		<= enable_s;
	img_coord_o		<= img_coord_s;
	enc_num_bits_o	<= enc_num_bits_s;
	enc_body_data_o	<= enc_body_data_s;
end architecture behavioural;