--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		25/02/2021
--------------------------------------------------------------------------------
-- IP name:		metadata_encod
--
-- Description: Defines the "Encoder Metadata" from compressed image header part
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.utils_image.all;
use work.param_image.all;

use work.utils_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;

entity metadata_encod is
	generic (
		-- "00": Sample-Adaptive Entropy, "01": Hybrid Entropy, "10": Block-Adaptive Entropy
		ENCODER_TYPE_G			: std_logic_vector(1 downto 0);
		-- Flag to add the "Accumulator Initialization Table"
		ACCU_INIT_TABLE_FLAG_G	: std_logic;
		-- 1: Restricted set of code options used, 0: Restricted set of code options NOT used
		RESTRICT_CODE_G			: std_logic
	);
	port (		
		md_enc_width_o			: out integer;
		md_enc_data_o			: out std_logic_vector(511 downto 0)
	);
end entity metadata_encod;

architecture behavioural of metadata_encod is
	-- Constants
	constant ACCU_INIT_TABLE_FLAG_C : std_logic := iif(ACCU_INIT_TABLE_FLAG_G = '1' and K_C < 0, '1', '0');
	
	-- Returns the length of the selected encoder
	function encoder_length(enc_type : in std_logic_vector; enc1 : in mdata_enc_smpl_adapt_t; enc2 : in mdata_enc_hybrid_t; enc3 : in mdata_enc_block_adapt_t) return integer is
		variable enc_length_v : integer := 0;
	begin
		if (enc_type = SMPL_ADAPT_C) then		-- "00": Sample-Adaptive Entropy
			enc_length_v := enc1.total_width;
		elsif (enc_type = HYBRID_C) then		-- "01": Hybrid Entropy
			enc_length_v := enc2.total_width;
		elsif (enc_type = BLK_ADAPT_C) then		-- "10": Block-Adaptive Entropy
			enc_length_v := enc3.total_width;
		else
			enc_length_v := 0;
        end if;

		return enc_length_v;
	end function;
	
	-- Returns the serialized data of the selected encoder
	function select_encoder(enc_type : in std_logic_vector; enc1 : in mdata_enc_smpl_adapt_t; enc2 : in mdata_enc_hybrid_t; enc3 : in mdata_enc_block_adapt_t; accu_init_en_in : in std_logic) return mdata_enc_t is
		-- Top encoder header type, already defining the width that the data is going to have...
		variable sel_encoder_v : mdata_enc_t(enc_subtype_data(encoder_length(enc_type, enc1, enc2, enc3)-1 downto 0));
	begin
		if (enc_type = SMPL_ADAPT_C) then		-- "00": Sample-Adaptive Entropy
			sel_encoder_v.enc_subtype_data	:= serial_mdata_enc_smpl_adapt(enc1, accu_init_en_in);
			sel_encoder_v.total_width		:= enc1.total_width;

		elsif (enc_type = HYBRID_C) then		-- "01": Hybrid Entropy
			sel_encoder_v.enc_subtype_data	:= serial_mdata_enc_hybrid(enc2);
			sel_encoder_v.total_width		:= enc2.total_width;

		elsif (enc_type = BLK_ADAPT_C) then		-- "10": Block-Adaptive Entropy
			sel_encoder_v.enc_subtype_data	:= serial_mdata_enc_block_adapt(enc3);
			sel_encoder_v.total_width		:= enc3.total_width;

		else
			sel_encoder_v.enc_subtype_data	:= (others => '0');
			sel_encoder_v.total_width		:= 0;
		end if;

		return sel_encoder_v;
	end function;

	-- Returns the size of the Accumulator Initialization Table
	function get_accu_init_size(flag_in : in std_logic) return integer is
		variable size_v		: integer := 0;
		variable pad_bits_v : integer := 0;
	begin
		if (flag_in = '0') then
			size_v := 0;
		else
			size_v := K2_ARR_C'length * 4;	-- As each position of the array is encoded within 4 bits...
		end if;
		
		-- In the same way as with the table, and only if necessary, size is increased until reaching next byte boundary.
		-- NOTE: "pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition...
		pad_bits_v := size_v mod 8;
		if (pad_bits_v /= 0) then
			size_v := size_v + (8 - pad_bits_v);
		end if;
		
		return size_v;
	end function get_accu_init_size;

	-- Creation of the Accumulator Initialization Table subblock
	function create_accu_init_table(flag_in : in std_logic) return std_logic_vector is
		variable accu_init_table_v	: std_logic_vector(1023 downto 0) := (others => '0');
		variable pointer_v			: integer := 0;
		variable pad_bits_v			: integer := 0;
	begin
		if (flag_in = '0') then
			return accu_init_table_v;		-- All '0's if table not enabled (wont be used later anyway)
		else
			for i in 0 to (NZ_C-1) loop
				accu_init_table_v(pointer_v+4-1 downto pointer_v) := std_logic_vector(to_unsigned(K2_ARR_C(i), 4));
				pointer_v := pointer_v + 4;
			end loop;
		end if;
		
		-- If necessary, fills with 0s until reach the next byte boundary ("pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition)
		pad_bits_v := pointer_v mod 8;
		accu_init_table_v((8-pad_bits_v)+pointer_v-1 downto 0) := padding_bits(accu_init_table_v(pointer_v-1 downto 0), (8-pad_bits_v), 0);

		-- The output signal has a fixed width, so here I cannot simply output the meaningful data, but everything
		-- return accu_init_table_v((8-pad_bits_v)+pointer_v-1 downto 0);
		return accu_init_table_v;
	end function create_accu_init_table;
	
	-- Record "Block Adaptive Entropy" sub-structure from "Entropy Coder Metadata" (Table 5-15)
	constant MDATA_ENC_BLOCK_ADAPT_C : mdata_enc_block_adapt_t := (
		reserved_1				=> (others => '0'),
		block_size				=> J_C,
		restr_code_opt_flag		=> iif(D_C <= 4 and RESTRICT_CODE_G = '1', "1", "0"),
		ref_smpl_interval		=> std_logic_vector(to_unsigned(Rs_C mod (2**12), 12)),
		total_width				=> iif(ENCODER_TYPE_G = BLK_ADAPT_C, 16, 0)
	);
	
	-- Record "Hybrid Entropy" sub-structure from "Entropy Coder Metadata" (Table 5-14)
	constant MDATA_ENC_HYBRID_C : mdata_enc_hybrid_t := (
		unary_len_limit			=> std_logic_vector(to_unsigned(Umax_C mod (2**5), 5)),
		resc_count_size			=> std_logic_vector(to_unsigned(Y_C-4, 3)),
		init_count_exp			=> std_logic_vector(to_unsigned(Yo_C mod (2**3), 3)),
		reserved_1				=> (others => '0'),
		total_width				=> iif(ENCODER_TYPE_G = HYBRID_C, 16, 0)
	);

	-- Record "Sample Adaptive Entropy" sub-structure from "Entropy Coder Metadata" (Table 5-13)
	constant MDATA_ENC_SMPL_ADAPT_C : mdata_enc_smpl_adapt_t := (
		unary_len_limit			=> std_logic_vector(to_unsigned(Umax_C mod (2**5), 5)),
		resc_count_size			=> std_logic_vector(to_unsigned(Y_C-4, 3)),
		init_count_exp			=> std_logic_vector(to_unsigned(Yo_C mod (2**3), 3)),
		accu_init_const			=> iif(K_C > 0, std_logic_vector(to_unsigned(K_C, 4)), "1111"),
		accu_init_table_flag	=> (others => ACCU_INIT_TABLE_FLAG_C),
		accu_init_table_size	=> get_accu_init_size(ACCU_INIT_TABLE_FLAG_C),
		accu_init_table			=> create_accu_init_table(ACCU_INIT_TABLE_FLAG_C),
		total_width				=> iif(ENCODER_TYPE_G = SMPL_ADAPT_C, 16, 0) +
									iif(ENCODER_TYPE_G = SMPL_ADAPT_C, iif(ACCU_INIT_TABLE_FLAG_C = '1', get_accu_init_size(ACCU_INIT_TABLE_FLAG_C), 0), 0)
	);

	-- Record "Encoder Metadata" structure (Additional Table)
	-- NOTE: Do not confuse "encoded with n-bits" vs "encoded mod 2**n with n-bits"!!
	constant MDATA_ENC_C	: mdata_enc_t := select_encoder(ENCODER_TYPE_G, MDATA_ENC_SMPL_ADAPT_C, MDATA_ENC_HYBRID_C, MDATA_ENC_BLOCK_ADAPT_C, ACCU_INIT_TABLE_FLAG_C);

	-- Encoder Metadata generation
	constant OUT_MD_ENC_WIDTH_C : integer := MDATA_ENC_C.total_width;
	constant OUT_MD_ENC_DATA_C	: std_logic_vector(OUT_MD_ENC_WIDTH_C-1 downto 0) := serial_mdata_enc(MDATA_ENC_C);

begin
	-- Outputs
	md_enc_width_o													<= OUT_MD_ENC_WIDTH_C;
	md_enc_data_o(md_enc_data_o'length-1 downto OUT_MD_ENC_WIDTH_C)	<= (others => '0');
	md_enc_data_o(OUT_MD_ENC_WIDTH_C-1 downto 0)					<= OUT_MD_ENC_DATA_C;

end architecture behavioural;