--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		20/02/2021
--------------------------------------------------------------------------------
-- IP name:		metadata_img
--
-- Description: Defines the "Image Metadata" from compressed image header part
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.supl_info_tables.all;
use work.utils_image.all;

use work.utils_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;

entity metadata_img is
	generic (
		-- 0: "signed" type samples, 1: "unsigned" type samples
		SMPL_TYPE_G			: std_logic;
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 00: Sample-Adaptive Entropy, 01: Hybrid Entropy, 10: Block-Adaptive Entropy
		ENCODER_TYPE_G		: std_logic_vector(1 downto 0);
		-- User Defined Data
		UDEF_DATA_G			: std_logic_vector(7 downto 0);
		-- Array with values -> 00: unsigned integer, 01: signed integer, 10: float
		SUPL_TABLE_TYPE_G	: supl_table_type_t;
		-- Array with values -> From 0 to 15 (check Table 3-1)
		SUPL_TABLE_PURPOSE_G: supl_table_purpose_t;
		-- Array with values -> 00: zero-dimensional, 01: one-dimensional, 10: two-dimensional-zx, 11: two-dimensional-yx
		SUPL_TABLE_STRUCT_G	: supl_table_struct_t;
		-- Array with values -> Suplementary User-Defined Data
		SUPL_TABLE_UDATA_G	: supl_table_udata_t
	);
	port (		
		md_img_width_o		: out integer;
		md_img_data_o		: out std_logic_vector(2047 downto 0)
	);
end entity metadata_img;

architecture behavioural of metadata_img is
	-- Returns the size of a specific Table Data Subblock
	function get_tdata_size(tdata_type_in : in std_logic_vector; tdata_struct_in : in std_logic_vector; index_in : in integer) return integer is
		variable size_v		: integer := 0;
		variable pad_bits_v : integer := 0;
	begin
		if (tdata_type_in = USG_TYPE_C) then		-- unsigned table type
			
			-- Initial values in the table
			size_v := 5;
			
			if (tdata_struct_in = DIM0_C) then			-- 00: zero-dimensional
				size_v := size_v + SUPL_TBL_0DIM_USG_ARR_C(index_in)'length;
					
			elsif (tdata_struct_in = DIM1_C) then		-- 01: one-dimensional
				size_v := size_v + (SUPL_TBL_1DIM_USG_ARR_C(index_in)'length * SUPL_TBL_1DIM_USG_ARR_C(index_in)(0)'length);

			elsif (tdata_struct_in = DIM2_ZX_C) then	-- 10: two-dimensional-zx
				size_v := size_v + (SUPL_TBL_2DIM_ZX_USG_ARR_C(index_in)'length * SUPL_TBL_2DIM_ZX_USG_ARR_C(index_in)(0)'length * SUPL_TBL_2DIM_ZX_USG_ARR_C(index_in)(0)(0)'length);

			elsif (tdata_struct_in = DIM2_YX_C) then	-- 11: two-dimensional-yx
				size_v := size_v + (SUPL_TBL_2DIM_YX_USG_ARR_C(index_in)'length * SUPL_TBL_2DIM_YX_USG_ARR_C(index_in)(0)'length * SUPL_TBL_2DIM_YX_USG_ARR_C(index_in)(0)(0)'length);

			else										-- NON-acceptable table structure
				size_v := 0;
			end if;

		elsif (tdata_type_in = SGD_TYPE_C) then		-- signed table type
			
			-- Initial values in the table
			size_v := 5;
			
			if (tdata_struct_in = DIM0_C) then			-- 00: zero-dimensional
				size_v := size_v + SUPL_TBL_0DIM_SGD_ARR_C(index_in)'length;
				
			elsif (tdata_struct_in = DIM1_C) then		-- 01: one-dimensional
				size_v := size_v + (SUPL_TBL_1DIM_SGD_ARR_C(index_in)'length * SUPL_TBL_1DIM_SGD_ARR_C(index_in)(0)'length);

			elsif (tdata_struct_in = DIM2_ZX_C) then	-- 10: two-dimensional-zx
				size_v := size_v + (SUPL_TBL_2DIM_ZX_SGD_ARR_C(index_in)'length * SUPL_TBL_2DIM_ZX_SGD_ARR_C(index_in)(0)'length * SUPL_TBL_2DIM_ZX_SGD_ARR_C(index_in)(0)(0)'length);

			elsif (tdata_struct_in = DIM2_YX_C) then	-- 11: two-dimensional-yx
				size_v := size_v + (SUPL_TBL_2DIM_YX_SGD_ARR_C(index_in)'length * SUPL_TBL_2DIM_YX_SGD_ARR_C(index_in)(0)'length * SUPL_TBL_2DIM_YX_SGD_ARR_C(index_in)(0)(0)'length);

			else										-- NON-acceptable table structure
				size_v := 0;
			end if;
			
		elsif (tdata_type_in = FLO_TYPE_C) then		-- float table type
		
			-- Initial values in the table
			size_v := 5 + 3 + DE_C;
		
			if (tdata_struct_in = DIM0_C) then			-- 00: zero-dimensional
				size_v := size_v + 1 + DE_C + DF_C;
				
			elsif (tdata_struct_in = DIM1_C) then		-- 01: one-dimensional
				size_v := size_v + (SUPL_TBL_1DIM_FLO_ARR_C(index_in)'length * (1 + DE_C + DF_C));

			elsif (tdata_struct_in = DIM2_ZX_C) then	-- 10: two-dimensional-zx
				size_v := size_v + (SUPL_TBL_2DIM_ZX_FLO_ARR_C(index_in)'length * SUPL_TBL_2DIM_ZX_FLO_ARR_C(index_in)(0)'length * (1 + DE_C + DF_C));

			elsif (tdata_struct_in = DIM2_YX_C) then	-- 11: two-dimensional-yx
				size_v := size_v + (SUPL_TBL_2DIM_YX_FLO_ARR_C(index_in)'length * SUPL_TBL_2DIM_YX_FLO_ARR_C(index_in)(0)'length * (1 + DE_C + DF_C));
			
			else										-- NON-acceptable table structure
				size_v := 0;
			end if;
		
		else	-- NON-acceptable table type
			size_v := 0;
		end if;

		-- In the same way as with the table, and only if necessary, size is increased until reaching next byte boundary.
		-- NOTE: "pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition...
		pad_bits_v := size_v mod 8;
		if (pad_bits_v /= 0) then
			size_v := size_v + (8 - pad_bits_v);
		end if;

		return size_v;
	end function get_tdata_size;

	-- Creates the Table Data Subblock
	function create_tdata_subblock(tdata_type_in : in std_logic_vector; tdata_struct_in : in std_logic_vector; index_in : in integer) return std_logic_vector is
		variable tdata_subblock_v : std_logic_vector(1023 downto 0) := (others => '0');
		variable pointer_v	: integer := 0;
		variable pad_bits_v	: integer := 0;
	begin
		if (tdata_type_in = USG_TYPE_C) then		-- unsigned table type
			
			-- (Un)signed table bit depth (DI_C) encoded
			tdata_subblock_v(4 downto pointer_v) := std_logic_vector(to_unsigned(DI_C mod (2**5), 5));
			pointer_v := 5;
			
			if (tdata_struct_in = DIM0_C) then			-- 00: zero-dimensional
				tdata_subblock_v(pointer_v+DI_C-1 downto pointer_v) := std_logic_vector(SUPL_TBL_0DIM_USG_ARR_C(index_in));
				pointer_v := pointer_v + DI_C;
				
			elsif (tdata_struct_in = DIM1_C) then		-- 01: one-dimensional
				for i in 0 to (NZ_C-1) loop
					tdata_subblock_v(pointer_v+DI_C-1 downto pointer_v) := std_logic_vector(SUPL_TBL_1DIM_USG_ARR_C(index_in)(i));
					pointer_v := pointer_v + DI_C;
				end loop;

			elsif (tdata_struct_in = DIM2_ZX_C) then	-- 10: two-dimensional-zx
				for i in 0 to (NZ_C-1) loop
					for j in 0 to (NX_C-1) loop
						tdata_subblock_v(pointer_v+DI_C-1 downto pointer_v) := std_logic_vector(SUPL_TBL_2DIM_ZX_USG_ARR_C(index_in)(i)(j));
						pointer_v := pointer_v + DI_C;
					end loop;
				end loop;

			elsif (tdata_struct_in = DIM2_YX_C) then	-- 11: two-dimensional-yx
				for i in 0 to (NY_C-1) loop
					for j in 0 to (NX_C-1) loop
						tdata_subblock_v(pointer_v+DI_C-1 downto pointer_v) := std_logic_vector(SUPL_TBL_2DIM_YX_USG_ARR_C(index_in)(i)(j));
						pointer_v := pointer_v + DI_C;
					end loop;
				end loop;

			else										-- NON-acceptable table structure
				tdata_subblock_v := (others => '0');
				return tdata_subblock_v;
			end if;

		elsif (tdata_type_in = SGD_TYPE_C) then		-- signed table type
			
			-- (Un)signed table bit depth (DI_C) encoded
			tdata_subblock_v(4 downto pointer_v) := std_logic_vector(to_unsigned(DI_C mod (2**5), 5));
			pointer_v := 5;
			
			if (tdata_struct_in = DIM0_C) then			-- 00: zero-dimensional
				tdata_subblock_v(pointer_v+DI_C-1 downto pointer_v) := std_logic_vector(SUPL_TBL_0DIM_SGD_ARR_C(index_in));
				pointer_v := pointer_v + DI_C;
				
			elsif (tdata_struct_in = DIM1_C) then		-- 01: one-dimensional
				for i in 0 to (NZ_C-1) loop
					tdata_subblock_v(pointer_v+DI_C-1 downto pointer_v) := std_logic_vector(SUPL_TBL_1DIM_SGD_ARR_C(index_in)(i));
					pointer_v := pointer_v + DI_C;
				end loop;

			elsif (tdata_struct_in = DIM2_ZX_C) then	-- 10: two-dimensional-zx
				for i in 0 to (NZ_C-1) loop
					for j in 0 to (NX_C-1) loop
						tdata_subblock_v(pointer_v+DI_C-1 downto pointer_v) := std_logic_vector(SUPL_TBL_2DIM_ZX_SGD_ARR_C(index_in)(i)(j));
						pointer_v := pointer_v + DI_C;
					end loop;
				end loop;

			elsif (tdata_struct_in = DIM2_YX_C) then	-- 11: two-dimensional-yx
				for i in 0 to (NY_C-1) loop
					for j in 0 to (NX_C-1) loop
						tdata_subblock_v(pointer_v+DI_C-1 downto pointer_v) := std_logic_vector(SUPL_TBL_2DIM_YX_SGD_ARR_C(index_in)(i)(j));
						pointer_v := pointer_v + DI_C;
					end loop;
				end loop;

			else										-- NON-acceptable table structure
				tdata_subblock_v := (others => '0');
				return tdata_subblock_v;
			end if;
			
		elsif (tdata_type_in = FLO_TYPE_C) then		-- float table type
		
			-- Float table significand bit depth (DF_C) encoded
			tdata_subblock_v(pointer_v+5-1 downto pointer_v) := std_logic_vector(to_unsigned(DF_C, 5));
			pointer_v := 5;
			
			-- Float table exponent bit depth (DE_C) encoded
			tdata_subblock_v(pointer_v+3-1 downto pointer_v) := std_logic_vector(to_unsigned(DE_C mod (2**3), 3));
			pointer_v := pointer_v + 3;
			
			-- Float table exponent bias (BETA_C) encoded
			tdata_subblock_v(pointer_v+DE_C-1 downto pointer_v) := std_logic_vector(to_unsigned(BETA_C, DE_C));
			pointer_v := pointer_v + DE_C;

			if (tdata_struct_in = DIM0_C) then			-- 00: zero-dimensional
				tdata_subblock_v(pointer_v+1-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_0DIM_FLO_ARR_C(index_in).b, 1));
				pointer_v := pointer_v + 1;
				tdata_subblock_v(pointer_v+DE_C-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_0DIM_FLO_ARR_C(index_in).alpha, DE_C));
				pointer_v := pointer_v + DE_C;
				tdata_subblock_v(pointer_v+DF_C-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_0DIM_FLO_ARR_C(index_in).j, DF_C));
				pointer_v := pointer_v + DF_C;
				
			elsif (tdata_struct_in = DIM1_C) then		-- 01: one-dimensional
				for i in 0 to (NZ_C-1) loop
					tdata_subblock_v(pointer_v+1-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_1DIM_FLO_ARR_C(index_in)(i).b, 1));
					pointer_v := pointer_v + 1;
					tdata_subblock_v(pointer_v+DE_C-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_1DIM_FLO_ARR_C(index_in)(i).alpha, DE_C));
					pointer_v := pointer_v + DE_C;
					tdata_subblock_v(pointer_v+DF_C-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_1DIM_FLO_ARR_C(index_in)(i).j, DF_C));
					pointer_v := pointer_v + DF_C;
				end loop;

			elsif (tdata_struct_in = DIM2_ZX_C) then	-- 10: two-dimensional-zx
				for i in 0 to (NZ_C-1) loop
					for j in 0 to (NX_C-1) loop
						tdata_subblock_v(pointer_v+1-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_2DIM_ZX_FLO_ARR_C(index_in)(i)(j).b, 1));
						pointer_v := pointer_v + 1;
						tdata_subblock_v(pointer_v+DE_C-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_2DIM_ZX_FLO_ARR_C(index_in)(i)(j).alpha, DE_C));
						pointer_v := pointer_v + DE_C;
						tdata_subblock_v(pointer_v+DF_C-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_2DIM_ZX_FLO_ARR_C(index_in)(i)(j).j, DF_C));
						pointer_v := pointer_v + DF_C;
					end loop;
				end loop;

			elsif (tdata_struct_in = DIM2_YX_C) then	-- 11: two-dimensional-yx
				for i in 0 to (NY_C-1) loop
					for j in 0 to (NX_C-1) loop
						tdata_subblock_v(pointer_v+1-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_2DIM_YX_FLO_ARR_C(index_in)(i)(j).b, 1));
						pointer_v := pointer_v + 1;
						tdata_subblock_v(pointer_v+DE_C-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_2DIM_YX_FLO_ARR_C(index_in)(i)(j).alpha, DE_C));
						pointer_v := pointer_v + DE_C;
						tdata_subblock_v(pointer_v+DF_C-1 downto pointer_v) := std_logic_vector(to_unsigned(SUPL_TBL_2DIM_YX_FLO_ARR_C(index_in)(i)(j).j, DF_C));
						pointer_v := pointer_v + DF_C;
					end loop;
				end loop;

			else										-- NON-acceptable table structure
				tdata_subblock_v := (others => '0');
				return tdata_subblock_v;
			end if;

		else	-- NON-acceptable table type
			tdata_subblock_v := (others => '0');
			return tdata_subblock_v;
		end if;		

		-- If necessary, fills with 0s until reach the next byte boundary ("pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition)
		pad_bits_v := pointer_v mod 8;
		tdata_subblock_v((8-pad_bits_v)+pointer_v-1 downto 0) := padding_bits(tdata_subblock_v(pointer_v-1 downto 0), (8-pad_bits_v), 0);
		
		-- The output signal has a fixed width, so here I cannot simply output the meaningful data, but everything
		-- return tdata_subblock_v((8-pad_bits_v)+pointer_v-1 downto 0);
		return tdata_subblock_v;
	end function create_tdata_subblock;

	-- Returns the total size of all instantiated Supplementary Inforamtion Tables
	function get_supl_tables_size(supl_tables_in : in mdata_img_supl_info_arr_t) return integer is
		variable supl_tables_size_v : integer := 0;
	begin
		for i in 0 to (supl_tables_in'length-1) loop
			supl_tables_size_v := supl_tables_size_v + supl_tables_in(i).total_width;
		end loop;
		
		return supl_tables_size_v;
	end function get_supl_tables_size;

	-- Creates the Supplementary Inforamtion Tables structure
	function create_supl_tables(num_tables_in : in integer) return mdata_img_supl_info_arr_t is
		variable supl_tables_v : mdata_img_supl_info_arr_t(0 to num_tables_in-1);
		variable tdata_size_v  : integer := 0;
	begin
		for i in 0 to (num_tables_in-1) loop
			-- Compiler requires a fixed size for "table_data_subblock", so such size is managed independently
			tdata_size_v := get_tdata_size(SUPL_TABLE_TYPE_G(i), SUPL_TABLE_STRUCT_G(i), i);

			-- Supplementary table is filled
			supl_tables_v(i).table_type			 := SUPL_TABLE_TYPE_G(i);
			supl_tables_v(i).reserved_1			 := (others => '0');
			supl_tables_v(i).table_purpose		 := SUPL_TABLE_PURPOSE_G(i);
			supl_tables_v(i).reserved_2			 := (others => '0');
			supl_tables_v(i).table_structure	 := SUPL_TABLE_STRUCT_G(i);
			supl_tables_v(i).reserved_3			 := (others => '0');
			supl_tables_v(i).supl_user_def_data	 := SUPL_TABLE_UDATA_G(i);
			supl_tables_v(i).table_data_size	 := tdata_size_v;
			supl_tables_v(i).table_data_subblock := create_tdata_subblock(SUPL_TABLE_TYPE_G(i), SUPL_TABLE_STRUCT_G(i), i);
			supl_tables_v(i).total_width		 := 16 + tdata_size_v;
		end loop;
		
		return supl_tables_v;
	end function create_supl_tables;
	
	-- Auxiliary constant to avoid the array below fail if "TAU_C=0"
	constant TAU_0_C : integer := iif(TAU_C < 1, 1, TAU_C);
	-- Record "Supplementary Information" sub-structure from "Image Metadata" (Table 5-4)
	-- NOTE: If TAU_C=0, array will be built anyway (so no compilation problems), but will not be serialized in the corresponding function...
	constant MDATA_IMG_SUPL_INFO_ARR_C : mdata_img_supl_info_arr_t(0 to TAU_0_C-1) := create_supl_tables(TAU_0_C);

	-- Record "Essential" sub-structure from "Image Metadata" (Table 5-3)
	constant MDATA_IMG_ESSEN_C : mdata_img_essential_t := (
		udef_data			=> UDEF_DATA_G,
		x_size				=> std_logic_vector(to_unsigned(NX_C mod (2**16), 16)),
		y_size				=> std_logic_vector(to_unsigned(NY_C mod (2**16), 16)),
		z_size				=> std_logic_vector(to_unsigned(NZ_C mod (2**16), 16)),
		smpl_type			=> (others => SMPL_TYPE_G),
		reserved_1			=> (others => '0'),
		larg_dyn_rng_flag	=> (others => iif(D_C > 16, '1', '0')),
		dyn_range			=> std_logic_vector(to_unsigned(D_C mod (2**4), 4)),
		smpl_enc_order		=> (others => iif(SMPL_ORDER_G = BSQ_C, '1', '0')),
		sub_frm_intlv_depth => std_logic_vector(to_unsigned(iif(SMPL_ORDER_G = BSQ_C, 0, iif(SMPL_ORDER_G = BIL_C, M_BIL_C, M_BIP_C)) mod (2**16), 16)),
		reserved_2			=> (others => '0'),
		out_word_size		=> std_logic_vector(to_unsigned(B_C mod (2**3), 3)),
		entropy_coder_type	=> ENCODER_TYPE_G,
		reserved_3			=> (others => '0'),
		quant_fidel_ctrl_mth=> FIDEL_CTRL_TYPE_G,
		reserved_4			=> (others => '0'),
		supl_info_table_cnt	=> std_logic_vector(to_unsigned(TAU_C, 4)),
		total_width			=> 96	-- The previous fields are a total of: 12 bytes * 8 bits/byte = 96 bits
	);

	-- Record "Image Metadata" structure (Table 5-2)
	-- NOTE: Do not confuse "encoded with n-bits" vs "encoded mod 2**n with n-bits"!!
	constant MDATA_IMG_C : mdata_img_t := (
		essential			=> MDATA_IMG_ESSEN_C,
		supl_info_arr		=> MDATA_IMG_SUPL_INFO_ARR_C,
		total_width			=> MDATA_IMG_ESSEN_C.total_width + iif(TAU_C > 0, get_supl_tables_size(MDATA_IMG_SUPL_INFO_ARR_C), 0)
	);
	
	-- Image Metadata generation
	constant OUT_MD_IMG_WIDTH_C : integer := MDATA_IMG_C.total_width;
	constant OUT_MD_IMG_DATA_C	: std_logic_vector(OUT_MD_IMG_WIDTH_C-1 downto 0) := serial_mdata_img(MDATA_IMG_C);

begin
	-- Outputs
	md_img_width_o													<= OUT_MD_IMG_WIDTH_C;
	md_img_data_o(md_img_data_o'length-1 downto OUT_MD_IMG_WIDTH_C)	<= (others => '0');
	md_img_data_o(OUT_MD_IMG_WIDTH_C-1 downto 0)					<= OUT_MD_IMG_DATA_C;

end architecture behavioural;