--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		20/02/2021
--------------------------------------------------------------------------------
-- IP name:		metadata_pred
--
-- Description: Defines the "Predictor Metadata" from compressed image header part
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;
use work.error_limit_values_table.all;

use work.types_encoder.all;
use work.utils_encoder.all;

entity metadata_pred is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G			: std_logic_vector(1 downto 0);
		-- 1: Full prediction mode, 0: Reduced prediction mode
		PREDICT_MODE_G			: std_logic;
		-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
		LSUM_TYPE_G				: std_logic_vector(1 downto 0);
		-- 1: Custom weight init, 0: Default weight init
		W_INIT_TYPE_G			: std_logic;
		-- 1: enabled, 0: disabled
		PER_ERR_LIM_UPD_G		: std_logic;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G		: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G		: std_logic;
		REL_ERR_BAND_TYPE_G		: std_logic;
		-- Flag to add the "Weight Initialization Table"
		W_INIT_TABL_FLAG_G		: std_logic;
		-- Flag to add the "Weight Exponent Offset Table"
		W_EXP_OFF_TABL_FLAG_G	: std_logic;
		-- Flag to add the "Damping Table"
		DAMP_TABLE_FLAG_G		: std_logic;
		-- Flag to add the "Offset Table"
		OFFSET_TABLE_FLAG_G		: std_logic
	);
	port (
		md_pred_width_o			: out integer;
		md_pred_data_o			: out std_logic_vector(2047 downto 0)
	);
end entity metadata_pred;

architecture behavioural of metadata_pred is
	-- Absolute/relative error limit values (initial values) are included in the "Encoder Header" ONLY if
	-- "Periodic Error Limit Updating" disabed. Otherwise, they are ignored here, but encoded in the "Encoder Body".
	
	-- 4.9.1.2: If lossless fidelity control is used, ѱz = 0 for all spectral bands z
	constant PSI_0_ARR_C			: array_integer_t(0 to NZ_C-1) := (others => 0);
	constant PSI_FIX_ARR_C			: array_integer_t(0 to NZ_C-1) := iif(FIDEL_CTRL_TYPE_G /= LOSSLESS_C, PSI_ARR_C, PSI_0_ARR_C);
	
	constant W_INIT_TABL_FLAG_C		: std_logic	:= iif(W_INIT_TABL_FLAG_G = '1' and W_INIT_TYPE_G = '1', '1', '0');
	constant W_EXP_OFF_TABL_FLAG_C	: std_logic	:= iif(W_EXP_OFF_TABL_FLAG_G = '1' and not((check_matrix_all_pos_same(Ci_MTX_C, 0) = '1') and (check_array_all_pos_same(C_ARR_C, 0) = '1')), '1', '0');

	constant DAMP_TABLE_FLAG_C		: std_logic := iif(DAMP_TABLE_FLAG_G = '1' and check_array_all_pos_same(FI_ARR_C, FI_ARR_C(0)) = '1', '1', '0');
	constant OFFSET_TABLE_FLAG_C	: std_logic	:= iif(OFFSET_TABLE_FLAG_G = '1' and check_array_all_pos_same(PSI_FIX_ARR_C, PSI_FIX_ARR_C(0)) = '1', '1', '0');
	
	-- Returns the size of the weight initialization table
	function get_winit_size(flag_in : in std_logic) return integer is
		variable cz_v		: integer := iif(PREDICT_MODE_G = '1', MAX_CZ_C, MAX_CZ_C-3);
		variable pad_bits_v	: integer := 0;
		variable size_v		: integer := 0;
	begin
		if (flag_in = '0') then
			size_v := 0;
		else		-- Instead of "LAMBDA_MTX_C(0)'length", the "cz_v" value is used (which depends on whether "Full" or "Reduced Prediction Mode")
			size_v := LAMBDA_MTX_C'length * cz_v * LAMBDA_MTX_C(0)(0)'length;
		end if;
		
		-- In the same way as with the table, and only if necessary, size is increased until reaching next byte boundary.
		-- NOTE: "pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition...
		pad_bits_v := size_v mod 8;
		if (pad_bits_v /= 0) then
			size_v := size_v + (8 - pad_bits_v);
		end if;
		
		return size_v;
	end function get_winit_size;
	
	-- Creation of the weight initialization table
	function create_w_init_table(flag_in : in std_logic) return std_logic_vector is
		variable w_init_table_v	: std_logic_vector(2047 downto 0) := (others => '0');
		variable pointer_v		: integer := 0;
		variable pad_bits_v		: integer := 0;
		variable j_init_v		: integer := iif(PREDICT_MODE_G = '1', 0, 3);	-- To handle MAX_CZ_C with whether "Full" or "Reduced Prediction Mode"
	begin
		if (flag_in = '0') then
			return w_init_table_v;		-- All '0's if table not enabled (wont be used later anyway)
		else
			for i in 0 to (NZ_C-1) loop
				for j in j_init_v to (MAX_CZ_C-1) loop		-- This value is already in two's complement (comes from "signed" type)
					w_init_table_v(pointer_v+Q_C-1 downto pointer_v) := std_logic_vector(LAMBDA_MTX_C(i)(j));
					pointer_v := pointer_v + Q_C;
				end loop;
			end loop;
		end if;

		-- If necessary, fills with 0s until reach the next byte boundary ("pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition)
		pad_bits_v := pointer_v mod 8;
		w_init_table_v((8-pad_bits_v)+pointer_v-1 downto 0) := padding_bits(w_init_table_v(pointer_v-1 downto 0), (8-pad_bits_v), 0);
		
		-- The output signal has a fixed width, so here I cannot simply output the meaningful data, but everything
		-- return w_init_table_v((8-pad_bits_v)+pointer_v-1 downto 0);
		return w_init_table_v;
	end function create_w_init_table;

	-- Returns the size of the weight exponent offset table
	function get_wexp_size(flag_in : in std_logic) return integer is
		variable pad_bits_v	: integer := 0;
		variable size_v		: integer := 0;
	begin
		if (flag_in = '0') then
			size_v := 0;
		else	-- Each value inside table is encoded with 4 bits.
			size_v := (C_ARR_C'length * 4) + (Ci_MTX_C'length * Ci_MTX_C(0)'length * 4);
		end if;
		
		-- In the same way as with the table, and only if necessary, size is increased until reaching next byte boundary.
		-- NOTE: "pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition...
		pad_bits_v := size_v mod 8;
		if (pad_bits_v /= 0) then
			size_v := size_v + (8 - pad_bits_v);
		end if;
		
		return size_v;
	end function get_wexp_size;
	
	-- Creation of the weight exponent offset table
	function create_w_exp_off_table(flag_in : in std_logic) return std_logic_vector is
		variable w_exp_off_table_v	: std_logic_vector(1023 downto 0) := (others => '0');
		variable pointer_v			: integer := 0;
		variable pad_bits_v			: integer := 0;
	begin
		if (flag_in = '0') then
			return w_exp_off_table_v;		-- All '0's if table not enabled (wont be used later anyway)
		else
			for i in 0 to (NZ_C-1) loop
				if (PREDICT_MODE_G = '1') then		-- This must be handled as two's complement ("signed")
					w_exp_off_table_v(pointer_v+4-1 downto pointer_v) := std_logic_vector(to_signed(C_ARR_C(i), 4));
					pointer_v := pointer_v + 4;
				end if;

				for j in 0 to (MAX_PZ_C-1) loop		-- This must be handled as two's complement ("signed")
					w_exp_off_table_v(pointer_v+4-1 downto pointer_v) := std_logic_vector(to_signed(Ci_MTX_C(i)(j), 4));
					pointer_v := pointer_v + 4;
				end loop;
			end loop;
		end if;
		
		-- If necessary, fills with 0s until reach the next byte boundary ("pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition)
		pad_bits_v := pointer_v mod 8;
		w_exp_off_table_v((8-pad_bits_v)+pointer_v-1 downto 0) := padding_bits(w_exp_off_table_v(pointer_v-1 downto 0), (8-pad_bits_v), 0);

		-- The output signal has a fixed width, so here I cannot simply output the meaningful data, but everything
		-- return w_exp_off_table_v((8-pad_bits_v)+pointer_v-1 downto 0);
		return w_exp_off_table_v;
	end function create_w_exp_off_table;
	
	-- Returns the size of the Absolute Error Limit Values subblock
	function get_abserr_size(band_type_in : in std_logic; err_lim : in err_lim_t) return integer is
		variable pad_bits_v	: integer := 0;
		variable size_v		: integer := 0;
	begin
		if (band_type_in = '1') then	-- band-dependent
			size_v := err_lim.abs_arr'length * DA_C;	-- Each value inside table is encoded with DA_C bits.
		else							-- band-independent
			size_v := DA_C;
		end if;
		
		-- In the same way as with the table, and only if necessary, size is increased until reaching next byte boundary.
		-- NOTE: "pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition...
		pad_bits_v := size_v mod 8;
		if (pad_bits_v /= 0) then
			size_v := size_v + (8 - pad_bits_v);
		end if;
		
		return size_v;
	end function get_abserr_size;
	
	-- Creation of the Absolute Error Limit Values subblock
	function create_abs_err_lim_val_subblock(band_type_in : in std_logic; err_lim : in err_lim_t) return std_logic_vector is
		variable abs_err_lim_val_v	: std_logic_vector(1023 downto 0) := (others => '0');
		variable pointer_v			: integer := 0;
		variable pad_bits_v			: integer := 0;
	begin
		if (band_type_in = '1') then	-- band-dependent
			for i in 0 to (NZ_C-1) loop
				abs_err_lim_val_v(pointer_v+DA_C-1 downto pointer_v) := std_logic_vector(to_unsigned(err_lim.abs_arr(i), DA_C));
				pointer_v := pointer_v + DA_C;
			end loop;
		else							-- band-independent
			abs_err_lim_val_v(DA_C-1 downto 0) := std_logic_vector(to_unsigned(err_lim.abs_c, DA_C));
			pointer_v := pointer_v + DA_C;
		end if;
		
		-- If necessary, fills with 0s until reach the next byte boundary ("pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition)
		pad_bits_v := pointer_v mod 8;
		abs_err_lim_val_v((8-pad_bits_v)+pointer_v-1 downto 0) := padding_bits(abs_err_lim_val_v(pointer_v-1 downto 0), (8-pad_bits_v), 0);

		-- The output signal has a fixed width, so here I cannot simply output the meaningful data, but everything
		-- return abs_err_lim_val_v((8-pad_bits_v)+pointer_v-1 downto 0);
		return abs_err_lim_val_v;
	end function create_abs_err_lim_val_subblock;
	
	-- Returns size of the Relative Error Limit Values subblock
	function get_relerr_size(band_type_in : in std_logic; err_lim : in err_lim_t) return integer is
		variable pad_bits_v	: integer := 0;
		variable size_v		: integer := 0;
	begin
		if (band_type_in = '1') then	-- band-dependent
			size_v := err_lim.rel_arr'length * DR_C;	-- Each value inside table is encoded with DR_C bits.
		else							-- band-independent
			size_v := DR_C;
		end if;
		
		-- In the same way as with the table, and only if necessary, size is increased until reaching next byte boundary.
		-- NOTE: "pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition...
		pad_bits_v := size_v mod 8;
		if (pad_bits_v /= 0) then
			size_v := size_v + (8 - pad_bits_v);
		end if;
		
		return size_v;
	end function get_relerr_size;
	
	-- Creation of the Relative Error Limit Values subblock
	function create_rel_err_lim_val_subblock(band_type_in : in std_logic; err_lim : in err_lim_t) return std_logic_vector is
		variable rel_err_lim_val_v	: std_logic_vector(1023 downto 0) := (others => '0');
		variable pointer_v			: integer := 0;
		variable pad_bits_v			: integer := 0;
	begin
		if (band_type_in = '1') then	-- band-dependent
			for i in 0 to (NZ_C-1) loop
				rel_err_lim_val_v(pointer_v+DR_C-1 downto pointer_v) := std_logic_vector(to_unsigned(err_lim.rel_arr(i), DR_C));
				pointer_v := pointer_v + DR_C;
			end loop;
		else							-- band-independent
			rel_err_lim_val_v(DR_C-1 downto 0) := std_logic_vector(to_unsigned(err_lim.rel_c, DR_C));
			pointer_v := pointer_v + DR_C;
		end if;
		
		-- If necessary, fills with 0s until reach the next byte boundary ("pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition)
		pad_bits_v := pointer_v mod 8;
		rel_err_lim_val_v((8-pad_bits_v)+pointer_v-1 downto 0) := padding_bits(rel_err_lim_val_v(pointer_v-1 downto 0), (8-pad_bits_v), 0);
		
		-- The output signal has a fixed width, so here I cannot simply output the meaningful data, but everything
		-- return rel_err_lim_val_v((8-pad_bits_v)+pointer_v-1 downto 0);
		return rel_err_lim_val_v;
	end function create_rel_err_lim_val_subblock;
	
	-- Returns the size of the Damping Table subblock
	function get_damp_size(flag_in : in std_logic) return integer is
		variable pad_bits_v	: integer := 0;
		variable size_v		: integer := 0;
	begin
		if (flag_in = '0') then
			size_v := 0;
		else		-- Each value inside table is encoded with THETA_C bits.
			size_v := FI_ARR_C'length * THETA_C;
		end if;
		
		-- In the same way as with the table, and only if necessary, size is increased until reaching next byte boundary.
		-- NOTE: "pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition...
		pad_bits_v := size_v mod 8;
		if (pad_bits_v /= 0) then
			size_v := size_v + (8 - pad_bits_v);
		end if;
		
		return size_v;
	end function get_damp_size;
	
	-- Creation of the Damping Table subblock
	function create_damp_table_subblock(flag_in : in std_logic) return std_logic_vector is
		variable damp_table_v	: std_logic_vector(1023 downto 0) := (others => '0');
		variable pointer_v		: integer := 0;
		variable pad_bits_v		: integer := 0;
	begin
		if (flag_in = '0') then
			return damp_table_v;		-- All '0's if table not enabled (wont be used later anyway)
		else
			for i in 0 to (NZ_C-1) loop
				damp_table_v(pointer_v+THETA_C-1 downto pointer_v) := std_logic_vector(to_unsigned(FI_ARR_C(i), THETA_C));
				pointer_v := pointer_v + THETA_C;
			end loop;
		end if;

		-- If necessary, fills with 0s until reach the next byte boundary ("pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition)
		pad_bits_v := pointer_v mod 8;
		damp_table_v((8-pad_bits_v)+pointer_v-1 downto 0) := padding_bits(damp_table_v(pointer_v-1 downto 0), (8-pad_bits_v), 0);
		
		-- The output signal has a fixed width, so here I cannot simply output the meaningful data, but everything
		-- return damp_table_v((8-pad_bits_v)+pointer_v-1 downto 0);
		return damp_table_v;
	end function create_damp_table_subblock;

	-- Returns the size of the Offset Table subblock
	function get_offset_size(flag_in : in std_logic) return integer is
		variable pad_bits_v	: integer := 0;
		variable size_v		: integer := 0;
	begin
		if (flag_in = '0') then
			size_v := 0;
		else		-- Each value inside table is encoded with THETA_C bits.
			size_v := PSI_FIX_ARR_C'length * THETA_C;
		end if;
		
		-- In the same way as with the table, and only if necessary, size is increased until reaching next byte boundary.
		-- NOTE: "pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition...
		pad_bits_v := size_v mod 8;
		if (pad_bits_v /= 0) then
			size_v := size_v + (8 - pad_bits_v);
		end if;
		
		return size_v;
	end function get_offset_size;
	
	-- Creation of the Offset Table subblock
	function create_offset_table_subblock(flag_in : in std_logic) return std_logic_vector is
		variable offset_table_v	: std_logic_vector(1023 downto 0) := (others => '0');
		variable pointer_v		: integer := 0;
		variable pad_bits_v		: integer := 0;
	begin
		if (flag_in = '0') then
			return offset_table_v;		-- All '0's if table not enabled (wont be used later anyway)
		else
			for i in 0 to (NZ_C-1) loop
				offset_table_v(pointer_v+THETA_C-1 downto pointer_v) := std_logic_vector(to_unsigned(PSI_FIX_ARR_C(i), THETA_C));
				pointer_v := pointer_v + THETA_C;
			end loop;
		end if;

		-- If necessary, fills with 0s until reach the next byte boundary ("pad_bits_v" are the extra bits above 8, so the difference with 8 must be the addition)
		pad_bits_v := pointer_v mod 8;
		offset_table_v((8-pad_bits_v)+pointer_v-1 downto 0) := padding_bits(offset_table_v(pointer_v-1 downto 0), (8-pad_bits_v), 0);

		-- The output signal has a fixed width, so here I cannot simply output the meaningful data, but everything
		-- return offset_table_v((8-pad_bits_v)+pointer_v-1 downto 0);
		return offset_table_v;
	end function create_offset_table_subblock;
	
	-- Record "Sample Representative" sub-structure from "Predictor Metadata" (Table 5-12)
	constant MDATA_PRED_SMPL_REPR_C : mdata_pred_smpl_repr_t := (
		reserved_1					=> (others => '0'),
		smpl_repr_resolution		=> std_logic_vector(to_unsigned(THETA_C, 3)),
		reserved_2					=> (others => '0'),
		band_var_damp_flag			=> (others => check_array_all_pos_same(FI_ARR_C, FI_ARR_C(0))),
		damp_table_flag				=> (others => DAMP_TABLE_FLAG_C),
		reserved_3					=> (others => '0'),
		fixed_damp_value			=> iif(check_array_all_pos_same(FI_ARR_C, FI_ARR_C(0)) = '0', std_logic_vector(to_unsigned(FI_ARR_C(0), 4)), "0000"),
		reserved_4					=> (others => '0'),
		band_var_offset_flag		=> (others => check_array_all_pos_same(PSI_FIX_ARR_C, PSI_FIX_ARR_C(0))),
		offset_table_flag			=> (others => OFFSET_TABLE_FLAG_C),
		reserved_5					=> (others => '0'),
		fixed_offset_value			=> iif(check_array_all_pos_same(PSI_FIX_ARR_C, PSI_FIX_ARR_C(0)) = '0', std_logic_vector(to_unsigned(PSI_FIX_ARR_C(0), 4)), "0000"),
		damp_table_size				=> get_damp_size(DAMP_TABLE_FLAG_C),
		damp_table_subblock			=> create_damp_table_subblock(DAMP_TABLE_FLAG_C),
		offset_table_size			=> get_offset_size(OFFSET_TABLE_FLAG_C),
		offset_table_subblock		=> create_offset_table_subblock(OFFSET_TABLE_FLAG_C),
		total_width					=> iif(THETA_C > 0, 24, 0) +
										iif(DAMP_TABLE_FLAG_C = '1', get_damp_size(DAMP_TABLE_FLAG_C), 0) +
										iif(OFFSET_TABLE_FLAG_C = '1', get_offset_size(OFFSET_TABLE_FLAG_C), 0)
	);

	-- Record "Relative Error Limit" sub-structure from "Predictor Metadata" (Table 5-11)
	constant MDATA_PRED_REL_ERR_LIM_C : mdata_pred_rel_err_limit_t := (
		reserved_1					=> (others => '0'),
		rel_err_limit_assig_meth	=> (others => REL_ERR_BAND_TYPE_G),
		reserved_2					=> (others => '0'),
		rel_err_limit_bit_depth		=> std_logic_vector(to_unsigned(DR_C mod (2**4), 4)),
		rel_err_limit_val_size		=> get_relerr_size(REL_ERR_BAND_TYPE_G, ERR_LIM_VAL_ARR_C(0)),
		rel_err_limit_val_subblock	=> create_rel_err_lim_val_subblock(REL_ERR_BAND_TYPE_G, ERR_LIM_VAL_ARR_C(0)),
		total_width					=> iif(FIDEL_CTRL_TYPE_G = REL_ERR_C or FIDEL_CTRL_TYPE_G = ABS_REL_C, 8, 0) +
										iif((FIDEL_CTRL_TYPE_G = REL_ERR_C or FIDEL_CTRL_TYPE_G = ABS_REL_C) and PER_ERR_LIM_UPD_G = '0', get_relerr_size(REL_ERR_BAND_TYPE_G, ERR_LIM_VAL_ARR_C(0)), 0)
	);
	
	-- Record "Absolute Error Limit" sub-structure from "Predictor Metadata" (Table 5-10)
	constant MDATA_PRED_ABS_ERR_LIM_C : mdata_pred_abs_err_limit_t := (
		reserved_1					=> (others => '0'),
		abs_err_limit_assig_meth	=> (others => ABS_ERR_BAND_TYPE_G),
		reserved_2					=> (others => '0'),
		abs_err_limit_bit_depth		=> std_logic_vector(to_unsigned(DA_C mod (2**4), 4)),
		abs_err_limit_val_size		=> get_abserr_size(ABS_ERR_BAND_TYPE_G, ERR_LIM_VAL_ARR_C(0)),
		abs_err_limit_val_subblock	=> create_abs_err_lim_val_subblock(ABS_ERR_BAND_TYPE_G, ERR_LIM_VAL_ARR_C(0)),
		total_width					=> iif(FIDEL_CTRL_TYPE_G = ABS_ERR_C or FIDEL_CTRL_TYPE_G = ABS_REL_C, 8, 0) +
										iif((FIDEL_CTRL_TYPE_G = ABS_ERR_C or FIDEL_CTRL_TYPE_G = ABS_REL_C) and PER_ERR_LIM_UPD_G = '0', get_abserr_size(ABS_ERR_BAND_TYPE_G, ERR_LIM_VAL_ARR_C(0)), 0)
	);
	
	-- Record "Error Limit Update Period" sub-structure from "Predictor Metadata" (Table 5-9)
	constant MDATA_PRED_ERR_LIM_UPD_PER_C : mdata_pred_err_limit_upd_period_t := (
		reserved_1					=> (others => '0'),
		per_upd_flag				=> (others => PER_ERR_LIM_UPD_G),
		reserved_2					=> (others => '0'),
		upd_period_exp				=> iif(PER_ERR_LIM_UPD_G = '1', std_logic_vector(to_unsigned(U_C, 4)), "0000"),
		total_width					=> iif(SMPL_ORDER_G /= BSQ_C, 8, 0)
	);

	-- Record "Quantization" sub-structure from "Predictor Metadata" (Table 5-8)
	constant MDATA_PRED_QUANT_C : mdata_pred_quant_t := (
		err_limit_upd_period		=> MDATA_PRED_ERR_LIM_UPD_PER_C,
		absol_err_limit				=> MDATA_PRED_ABS_ERR_LIM_C,
		relat_err_limit				=> MDATA_PRED_REL_ERR_LIM_C,
		total_width					=> iif(SMPL_ORDER_G /= BSQ_C, MDATA_PRED_ERR_LIM_UPD_PER_C.total_width, 0) +
										iif(FIDEL_CTRL_TYPE_G = ABS_ERR_C or FIDEL_CTRL_TYPE_G = ABS_REL_C, MDATA_PRED_ABS_ERR_LIM_C.total_width, 0) + 
										iif(FIDEL_CTRL_TYPE_G = REL_ERR_C or FIDEL_CTRL_TYPE_G = ABS_REL_C, MDATA_PRED_REL_ERR_LIM_C.total_width, 0)
	);

	-- Record "Weight tables" sub-structure from "Predictor Metadata" (Table 5-7)
	constant MDATA_PRED_WEIGHT_TABLES_C : mdata_pred_weight_tables_t := (
		w_init_table_size			=> get_winit_size(W_INIT_TABL_FLAG_C),
		w_init_table				=> create_w_init_table(W_INIT_TABL_FLAG_C),
		w_exp_off_table_size		=> get_wexp_size(W_EXP_OFF_TABL_FLAG_C),
		w_exp_off_table				=> create_w_exp_off_table(W_EXP_OFF_TABL_FLAG_C),
		total_width					=> iif(W_INIT_TABL_FLAG_C = '1', get_winit_size(W_INIT_TABL_FLAG_C), 0) +
										iif(W_EXP_OFF_TABL_FLAG_C = '1', get_wexp_size(W_EXP_OFF_TABL_FLAG_C), 0)
	);

	-- Record "Primary" sub-structure from "Predictor Metadata" (Table 5-6)
	constant MDATA_PRED_PRIMARY_C : mdata_pred_primary_t := (
		reserved_1					=> (others => '0'),
		smpl_repr_flag				=> (others => iif(THETA_C > 0, '1', '0')),
		num_pred_bands				=> std_logic_vector(to_unsigned(P_C, 4)),
		pred_mode					=> (others => not PREDICT_MODE_G),		-- Documentation uses here the opposite logic as I do
		w_exp_offset_flag			=> (others => iif((check_matrix_all_pos_same(Ci_MTX_C, 0) = '0') and (check_array_all_pos_same(C_ARR_C, 0) = '0'), '0', '1')),
		lsum_type					=> LSUM_TYPE_G,
		register_size				=> std_logic_vector(to_unsigned(Re_C mod (2**6), 6)),
		w_comp_res					=> std_logic_vector(to_unsigned(OMEGA_C-4, 4)),
		w_upd_scal_exp_chng_int		=> std_logic_vector(to_unsigned(log2((2**T_INC_C)-4), 4)), -- Unlike docu, T_INC_C is just the power value, not the final one, so always use 2**T_INC_C
		w_upd_scal_exp_init_param	=> std_logic_vector(to_unsigned(V_MIN_C+6, 4)),
		w_upd_scal_exp_final_param	=> std_logic_vector(to_unsigned(V_MAX_C+6, 4)),
		w_exp_off_table_flag		=> (others => W_EXP_OFF_TABL_FLAG_C),
		w_init_method				=> (others => W_INIT_TYPE_G),
		w_init_table_flag			=> (others => W_INIT_TABL_FLAG_C),
		w_init_res					=> iif(W_INIT_TYPE_G = '1', std_logic_vector(to_unsigned(Q_C, 5)), "00000"),
		total_width					=> 40
	);

	-- Record "Predictor Metadata" structure (Table 5-5)
	-- NOTE: Do not confuse "encoded with n-bits" vs "encoded mod 2**n with n-bits"!!
	constant MDATA_PRED_C : mdata_pred_t := (
		primary						=> MDATA_PRED_PRIMARY_C,
		weight_tables				=> MDATA_PRED_WEIGHT_TABLES_C,
		quantization				=> MDATA_PRED_QUANT_C,
		smpl_repr					=> MDATA_PRED_SMPL_REPR_C,
		total_width					=> MDATA_PRED_PRIMARY_C.total_width + 
										iif(W_INIT_TABL_FLAG_C = '1' or W_EXP_OFF_TABL_FLAG_C = '1', MDATA_PRED_WEIGHT_TABLES_C.total_width, 0) +
										iif(FIDEL_CTRL_TYPE_G /= LOSSLESS_C, MDATA_PRED_QUANT_C.total_width, 0) + 
										iif(THETA_C > 0, MDATA_PRED_SMPL_REPR_C.total_width, 0)
	);
	
	-- Predictor Metadata generation
	constant OUT_MD_PRED_WIDTH_C : integer := MDATA_PRED_C.total_width;
	constant OUT_MD_PRED_DATA_C  : std_logic_vector(OUT_MD_PRED_WIDTH_C-1 downto 0) := serial_mdata_pred(MDATA_PRED_C, W_INIT_TABL_FLAG_C, W_EXP_OFF_TABL_FLAG_C, FIDEL_CTRL_TYPE_G, SMPL_ORDER_G, PER_ERR_LIM_UPD_G, THETA_C, DAMP_TABLE_FLAG_C, OFFSET_TABLE_FLAG_C);

begin	
	-- Outputs
	md_pred_width_o														<= OUT_MD_PRED_WIDTH_C;
	md_pred_data_o(md_pred_data_o'length-1 downto OUT_MD_PRED_WIDTH_C)	<= (others => '0');
	md_pred_data_o(OUT_MD_PRED_WIDTH_C-1 downto 0)						<= OUT_MD_PRED_DATA_C;

end architecture behavioural;