--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		14/10/2021
--------------------------------------------------------------------------------
-- IP name:		parallel_sync_fifos
--
-- Description: Two FIFOs to stall 'data' and 'image coordinates' when something
--				else (controlled by a 'condition' signal) must be sent out.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

-- Xilinx Parameterized Macro (XPM), version 2019.1
library xpm;
use xpm.vcomponents.all;

entity parallel_sync_fifos is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	: std_logic_vector(1 downto 0);
		-- FIFO data width (for writing [input] & reading [output] channels)
		DATA_WIDTH_G	: integer;
		-- FIFO data depth (for writing [input] & reading [output] channels). MUST be a power of 2, and at least 16!!
		DATA_DEPTH_G	: integer;
		-- Number of parallel sync FIFOs to store data (apart from the 'image coordinates')
		NUM_PAR_FIFOS_G	: integer
	);
	port (
		clock_i		 	: in  std_logic;
		reset_i		 	: in  std_logic;
		
		enable_i	 	: in  std_logic;
		enable_o	 	: out std_logic;
		img_coord_i	 	: in  img_coord_t;
		img_coord_o	 	: out img_coord_t;
		
		read_fifo_i	 	: in  std_logic;
		init_fifo_o		: out std_logic;
		
		data_arr_i	 	: in  array_slv_t(0 to NUM_PAR_FIFOS_G-1)(DATA_WIDTH_G-1 downto 0);
		data_arr_o	 	: out array_slv_t(0 to NUM_PAR_FIFOS_G-1)(DATA_WIDTH_G-1 downto 0)
	);
end entity parallel_sync_fifos;

architecture behavioural of parallel_sync_fifos is
	-- FIFOs configuration
	constant FIFO_READ_MODE_C		: string  := "fwft";
	constant FIFO_MEMORY_TYPE_C		: string  := "auto";
	
	constant FIFO_WR_DATA_WIDTH_C	: integer := DATA_WIDTH_G;
	constant FIFO_RD_DATA_WIDTH_C	: integer := DATA_WIDTH_G;
	constant FIFO_WR_DATA_DEPTH_C	: integer := DATA_DEPTH_G;		-- MUST be at least 16, and a power of 2!!
	constant FIFO_RD_DATA_DEPTH_C	: integer := FIFO_WR_DATA_DEPTH_C * FIFO_WR_DATA_WIDTH_C / FIFO_RD_DATA_WIDTH_C;
	constant FIFO_WR_DATA_COUNT_C 	: integer := log2(FIFO_WR_DATA_DEPTH_C) + 1;
	constant FIFO_RD_DATA_COUNT_C	: integer := log2(FIFO_RD_DATA_DEPTH_C) + 1;
	
	-- Extra bit for the coordinates in order to work with "signed" (as intermediate coordinate value '-1' exists, but no more negative values are generated)
	constant FIFO_X_COORD_WIDTH_C	: integer := log2(NX_C) + 1;
	constant FIFO_Y_COORD_WIDTH_C	: integer := log2(NY_C) + 1;
	constant FIFO_Z_COORD_WIDTH_C	: integer := log2(NZ_C) + 1;

	constant FIFO_WR_COORD_WIDTH_C	: integer := FIFO_X_COORD_WIDTH_C + FIFO_Y_COORD_WIDTH_C + FIFO_Z_COORD_WIDTH_C;
	constant FIFO_RD_COORD_WIDTH_C	: integer := FIFO_X_COORD_WIDTH_C + FIFO_Y_COORD_WIDTH_C + FIFO_Z_COORD_WIDTH_C;
	constant FIFO_WR_COORD_DEPTH_C	: integer := DATA_DEPTH_G;		-- MUST be at least 16, and a power of 2!!
	constant FIFO_RD_COORD_DEPTH_C	: integer := FIFO_WR_COORD_DEPTH_C * FIFO_WR_COORD_WIDTH_C / FIFO_RD_COORD_WIDTH_C;
	constant FIFO_WR_COORD_COUNT_C	: integer := log2(FIFO_WR_COORD_DEPTH_C) + 1;
	constant FIFO_RD_COORD_COUNT_C	: integer := log2(FIFO_RD_COORD_DEPTH_C) + 1;

	-- Signals declaration
	signal wr_rst_busy_data_arr_s	: std_logic_vector(NUM_PAR_FIFOS_G-1 downto 0) := (others => '0');
	signal rd_rst_busy_data_arr_s	: std_logic_vector(NUM_PAR_FIFOS_G-1 downto 0) := (others => '0');
	signal wr_en_data_arr_s			: std_logic_vector(NUM_PAR_FIFOS_G-1 downto 0) := (others => '0');
	signal rd_en_data_arr_s			: std_logic_vector(NUM_PAR_FIFOS_G-1 downto 0) := (others => '0');
	signal full_data_arr_s			: std_logic_vector(NUM_PAR_FIFOS_G-1 downto 0) := (others => '0');
	signal empty_data_arr_s			: std_logic_vector(NUM_PAR_FIFOS_G-1 downto 0) := (others => '0');
	signal wr_data_count_arr_s		: array_slv_t(0 to NUM_PAR_FIFOS_G-1)(FIFO_WR_DATA_COUNT_C-1 downto 0) := (others => (others => '0'));
	signal rd_data_count_arr_s		: array_slv_t(0 to NUM_PAR_FIFOS_G-1)(FIFO_RD_DATA_COUNT_C-1 downto 0) := (others => (others => '0'));

	signal wr_rst_busy_coord_s		: std_logic	:= '0';
	signal rd_rst_busy_coord_s		: std_logic	:= '0';
	signal wr_en_coord_s			: std_logic	:= '0';
	signal rd_en_coord_s			: std_logic	:= '0';
	signal full_coord_s				: std_logic	:= '0';
	signal empty_coord_s			: std_logic	:= '0';
	signal wr_coord_count_s			: std_logic_vector(FIFO_WR_COORD_COUNT_C-1 downto 0) := (others => '0');
	signal rd_coord_count_s			: std_logic_vector(FIFO_RD_COORD_COUNT_C-1 downto 0) := (others => '0');
	
	signal enable_s					: std_logic	:= '0';
	signal serial_img_coord_in_s	: std_logic_vector(FIFO_X_COORD_WIDTH_C+FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C-1 downto 0) := (others => '0');
	signal serial_img_coord_out_s	: std_logic_vector(FIFO_X_COORD_WIDTH_C+FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C-1 downto 0) := (others => '0');
	signal extr_img_coord_s			: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal data_arr_s				: array_slv_t(0 to NUM_PAR_FIFOS_G-1)(DATA_WIDTH_G-1 downto 0) := (others => (others => '0'));

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_fifos_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s <= '0';
			else
				enable_s <= enable_i;
			end if;
		end if;
	end process p_fifos_delay;
	
	-- This array of FIFOs store the processed encoder body data while the encoder header is still outputting its data
	g_sync_fifo_data : for i in 0 to NUM_PAR_FIFOS_G-1 generate 
		i_sync_fifo_data : xpm_fifo_sync	-- xpm_fifo_async
		generic map(
			FIFO_MEMORY_TYPE	=> FIFO_MEMORY_TYPE_C,	-- Designate FIFO memory primitive (resource type). "auto": Allow Vivado to choose, "block": Block RAM FIFO, "distributed": Distributed RAM FIFO
			FIFO_WRITE_DEPTH	=> FIFO_WR_DATA_DEPTH_C,-- Defines the FIFO Write Depth, must be power of two.
			WRITE_DATA_WIDTH	=> FIFO_WR_DATA_WIDTH_C,-- Defines the width of the write data port (din)
			READ_MODE			=> FIFO_READ_MODE_C,	-- "std": standard read mode / "fwft": First-Word-Fall-Through read mode
			FIFO_READ_LATENCY	=> 0,					-- Number of output register stages in the read data path (if READ_MODE = "fwft", the only valid value is 0).
			FULL_RESET_VALUE	=> 0,					-- Sets full, almost_full and prog_full to FULL_RESET_VALUE during reset.
			USE_ADV_FEATURES	=> "0404",				-- Enables/Disables "data_valid", "almost_empty", "rd_data_count", "prog_empty", "underflow", "wr_ack", "almost_full", "wr_data_count", "prog_full", "overflow" signals.
			READ_DATA_WIDTH		=> FIFO_RD_DATA_WIDTH_C,-- Defines the width of the read data port (dout)
			WR_DATA_COUNT_WIDTH	=> FIFO_WR_DATA_COUNT_C,-- Specifies the width of wr_data_count. To reflect the correct value, the width should be log2(FIFO_WRITE_DEPTH)+1.
			PROG_FULL_THRESH	=> 10,					-- Specifies the maximum number of write words in the FIFO at or above which prog_full is asserted.
			RD_DATA_COUNT_WIDTH	=> FIFO_RD_DATA_COUNT_C,-- Specifies the width of rd_data_count. To reflect the correct value, the width should be log2(FIFO_READ_DEPTH)+1.
			PROG_EMPTY_THRESH	=> 10,					-- Specifies the minimum number of read words in the FIFO at or below which prog_empty is asserted.
			DOUT_RESET_VALUE	=> "0",					-- Reset value of read data path.
			ECC_MODE			=> "no_ecc",			-- "no_ecc": Disables ECC, "en_ecc": Enables both ECC Encoder and Decoder
			WAKEUP_TIME			=> 0,					-- 0: Disable sleep, 2: Use Sleep Pin
			SIM_ASSERT_CHK		=> 0					-- 0: Disable simulation message reporting, 1: Enable simulation message reporting
			-- CDC_SYNC_STAGES	=> 2,					-- Specifies the number of synchronization stages on the CDC path
			-- RELATED_CLOCKS	=> 1					-- Specifies if the "wr_clk" and "rd_clk" are related having the same source but different clock ratios
		)
		port map(
			-- rd_clk		=> clock_i,					-- 1-bit input: Read clock: Used for read operation. rd_clk must be a free running clock.
			wr_clk			=> clock_i,					-- 1-bit input: Write clock: Used for write operation. wr_clk must be a free running clock.
			rst				=> reset_i,					-- 1-bit input: Reset: Must be synchronous to wr_clk. The clock(s) can be unstable at the time of applying reset, but reset must be released only after the clock(s) is/are stable.
			wr_en			=> wr_en_data_arr_s(i),		-- 1-bit input: Write Enable: If the FIFO is not full, asserting this signal causes data (on din) to be written to the FIFO. Must be held active-low when rst or wr_rst_busy is active high.
			din				=> data_arr_i(i),			-- WRITE_DATA_WIDTH-bit input: Write Data: The input data bus used when writing the FIFO.
			full			=> full_data_arr_s(i),		-- 1-bit output: Full Flag: When asserted, this signal indicates that the FIFO is full. Write requests are ignored when the FIFO is full, initiating a write when the FIFO is full is not destructive to the contents of the FIFO.
			prog_full		=> open,					-- 1-bit output: Programmable Full: This signal is asserted when the number of words in the FIFO is greater than or equal to the programmable full threshold value. It is de-asserted when the number of words in the FIFO is less than the programmable full threshold value.
			wr_data_count	=> wr_data_count_arr_s(i),	-- WR_DATA_COUNT_WIDTH-bit output: Write Data Count: This bus indicates the number of words written into the FIFO.
			overflow		=> open,					-- 1-bit output: Overflow: This signal indicates that a write request (wren) during the prior clock cycle was rejected, because the FIFO is full. Overflowing the FIFO is not destructive to the contents of the FIFO.
			wr_rst_busy		=> wr_rst_busy_data_arr_s(i), -- 1-bit output: Write Reset Busy: Active-High indicator that the FIFO write domain is currently in a reset state.
			almost_full		=> open,					-- 1-bit output: Almost Full: When asserted, this signal indicates that only one more write can be performed before the FIFO is full.
			wr_ack			=> open,					-- 1-bit output: Write Acknowledge: This signal indicates that a write request (wr_en) during the prior clock cycle is succeeded.
			rd_en			=> rd_en_data_arr_s(i),		-- 1-bit input: Read Enable: If the FIFO is not empty, asserting this signal causes data (on dout) to be read from the FIFO. Must be held active-low when rd_rst_busy is active high.
			dout			=> data_arr_s(i),			-- READ_DATA_WIDTH-bit output: Read Data: The output data bus is driven when reading the FIFO.
			empty			=> empty_data_arr_s(i),		-- 1-bit output: Empty Flag: When asserted, this signal indicates that the FIFO is empty. Read requests are ignored when the FIFO is empty, initiating a read while empty is not destructive to the FIFO.
			prog_empty		=> open,					-- 1-bit output: Programmable Empty: This signal is asserted when the number of words in the FIFO is less than or equal to the programmable empty threshold value. It is de-asserted when the number of words in the FIFO exceeds the programmable empty threshold value.
			rd_data_count	=> rd_data_count_arr_s(i),	-- RD_DATA_COUNT_WIDTH-bit output: Read Data Count: This bus indicates the number of words read from the FIFO.
			underflow		=> open,					-- 1-bit output: Underflow: Indicates that the read request (rd_en) during the previous clock cycle was rejected because the FIFO is empty. Under flowing the FIFO is not destructive to the FIFO.
			rd_rst_busy		=> rd_rst_busy_data_arr_s(i), -- 1-bit output: Read Reset Busy: Active-High indicator that the FIFO read domain is currently in a reset state.
			almost_empty	=> open,					-- 1-bit output: Almost Empty : When asserted, this signal indicates that only one more read can be performed before the FIFO goes to empty.
			data_valid		=> open,					-- 1-bit output: Read Data Valid: When asserted, this signal indicates that valid data is available on the output bus (dout).
			injectdbiterr	=> '0',						-- 1-bit input: Double Bit Error Injection: Injects a double bit error if the ECC feature is used on block RAMs or UltraRAM macros.
			injectsbiterr	=> '0',						-- 1-bit input: Single Bit Error Injection: Injects a single bit error if the ECC feature is used on block RAMs or UltraRAM macros.
			sbiterr			=> open,					-- 1-bit output: Single Bit Error: Indicates that the ECC decoder detected and fixed a single-bit error.
			dbiterr			=> open,					-- 1-bit output: Double Bit Error: Indicates that the ECC decoder detected a double-bit error and data in the FIFO core is corrupted.
			sleep			=> '0'						-- 1-bit input: Dynamic power saving: If sleep is High, the memory/fifo block is in power saving mode.
		);
		-- "wr_en"/"rd_en" should not be toggled when "rst" or "wr_rst_busy" or "rd_rst_busy" is asserted...
		wr_en_data_arr_s(i) <= '1' when (wr_rst_busy_data_arr_s(i) = '0' and full_data_arr_s(i) = '0') else '0';
		rd_en_data_arr_s(i) <= '1' when (rd_rst_busy_data_arr_s(i) = '0' and empty_data_arr_s(i) = '0' and read_fifo_i = '1') else '0';
	end generate g_sync_fifo_data;

	-- This FIFO stores the image coordinates while the encoder header is still outputting its data
	i_sync_fifo_coord : xpm_fifo_sync		-- xpm_fifo_async
	generic map(
		FIFO_MEMORY_TYPE	=> FIFO_MEMORY_TYPE_C,	 -- Designate FIFO memory primitive (resource type). "auto": Allow Vivado to choose, "block": Block RAM FIFO, "distributed": Distributed RAM FIFO
		FIFO_WRITE_DEPTH	=> FIFO_WR_COORD_DEPTH_C,-- Defines the FIFO Write Depth, must be power of two.
		WRITE_DATA_WIDTH	=> FIFO_WR_COORD_WIDTH_C,-- Defines the width of the write data port (din)
		READ_MODE			=> FIFO_READ_MODE_C,	 -- "std": standard read mode / "fwft": First-Word-Fall-Through read mode
		FIFO_READ_LATENCY	=> 0,					 -- Number of output register stages in the read data path (if READ_MODE = "fwft", the only valid value is 0).
		FULL_RESET_VALUE	=> 0,					 -- Sets full, almost_full and prog_full to FULL_RESET_VALUE during reset.
		USE_ADV_FEATURES	=> "0404",				 -- Enables/Disables "data_valid", "almost_empty", "rd_data_count", "prog_empty", "underflow", "wr_ack", "almost_full", "wr_data_count", "prog_full", "overflow" signals.
		READ_DATA_WIDTH		=> FIFO_RD_COORD_WIDTH_C,-- Defines the width of the read data port (dout)
		WR_DATA_COUNT_WIDTH	=> FIFO_WR_COORD_COUNT_C,-- Specifies the width of wr_data_count. To reflect the correct value, the width should be log2(FIFO_WRITE_DEPTH)+1.
		PROG_FULL_THRESH	=> 10,					 -- Specifies the maximum number of write words in the FIFO at or above which prog_full is asserted.
		RD_DATA_COUNT_WIDTH	=> FIFO_RD_COORD_COUNT_C,-- Specifies the width of rd_data_count. To reflect the correct value, the width should be log2(FIFO_READ_DEPTH)+1.
		PROG_EMPTY_THRESH	=> 10,					 -- Specifies the minimum number of read words in the FIFO at or below which prog_empty is asserted.
		DOUT_RESET_VALUE	=> "0",					 -- Reset value of read data path.
		ECC_MODE			=> "no_ecc",			 -- "no_ecc": Disables ECC, "en_ecc": Enables both ECC Encoder and Decoder
		WAKEUP_TIME			=> 0,					 -- 0: Disable sleep, 2: Use Sleep Pin
		SIM_ASSERT_CHK		=> 0					 -- 0: Disable simulation message reporting, 1: Enable simulation message reporting
		-- CDC_SYNC_STAGES	=> 2,					 -- Specifies the number of synchronization stages on the CDC path
		-- RELATED_CLOCKS	=> 1					 -- Specifies if the "wr_clk" and "rd_clk" are related having the same source but different clock ratios
	)
	port map(
		-- rd_clk		=> clock_i,				-- 1-bit input: Read clock: Used for read operation. rd_clk must be a free running clock.
		wr_clk			=> clock_i,				-- 1-bit input: Write clock: Used for write operation. wr_clk must be a free running clock.
		rst				=> reset_i,				-- 1-bit input: Reset: Must be synchronous to wr_clk. The clock(s) can be unstable at the time of applying reset, but reset must be released only after the clock(s) is/are stable.
		wr_en			=> wr_en_coord_s,		-- 1-bit input: Write Enable: If the FIFO is not full, asserting this signal causes data (on din) to be written to the FIFO. Must be held active-low when rst or wr_rst_busy is active high.
		din				=> serial_img_coord_in_s, -- WRITE_DATA_WIDTH-bit input: Write Data: The input data bus used when writing the FIFO.
		full			=> full_coord_s,		-- 1-bit output: Full Flag: When asserted, this signal indicates that the FIFO is full. Write requests are ignored when the FIFO is full, initiating a write when the FIFO is full is not destructive to the contents of the FIFO.
		prog_full		=> open,				-- 1-bit output: Programmable Full: This signal is asserted when the number of words in the FIFO is greater than or equal to the programmable full threshold value. It is de-asserted when the number of words in the FIFO is less than the programmable full threshold value.
		wr_data_count	=> wr_coord_count_s,	-- WR_DATA_COUNT_WIDTH-bit output: Write Data Count: This bus indicates the number of words written into the FIFO.
		overflow		=> open,				-- 1-bit output: Overflow: This signal indicates that a write request (wren) during the prior clock cycle was rejected, because the FIFO is full. Overflowing the FIFO is not destructive to the contents of the FIFO.
		wr_rst_busy		=> wr_rst_busy_coord_s,	-- 1-bit output: Write Reset Busy: Active-High indicator that the FIFO write domain is currently in a reset state.
		almost_full		=> open,				-- 1-bit output: Almost Full: When asserted, this signal indicates that only one more write can be performed before the FIFO is full.
		wr_ack			=> open,				-- 1-bit output: Write Acknowledge: This signal indicates that a write request (wr_en) during the prior clock cycle is succeeded.
		rd_en			=> rd_en_coord_s,		-- 1-bit input: Read Enable: If the FIFO is not empty, asserting this signal causes data (on dout) to be read from the FIFO. Must be held active-low when rd_rst_busy is active high.
		dout			=> serial_img_coord_out_s, -- READ_DATA_WIDTH-bit output: Read Data: The output data bus is driven when reading the FIFO.
		empty			=> empty_coord_s,		-- 1-bit output: Empty Flag: When asserted, this signal indicates that the FIFO is empty. Read requests are ignored when the FIFO is empty, initiating a read while empty is not destructive to the FIFO.
		prog_empty		=> open,				-- 1-bit output: Programmable Empty: This signal is asserted when the number of words in the FIFO is less than or equal to the programmable empty threshold value. It is de-asserted when the number of words in the FIFO exceeds the programmable empty threshold value.
		rd_data_count	=> rd_coord_count_s,	-- RD_DATA_COUNT_WIDTH-bit output: Read Data Count: This bus indicates the number of words read from the FIFO.
		underflow		=> open,				-- 1-bit output: Underflow: Indicates that the read request (rd_en) during the previous clock cycle was rejected because the FIFO is empty. Under flowing the FIFO is not destructive to the FIFO.
		rd_rst_busy		=> rd_rst_busy_coord_s,	-- 1-bit output: Read Reset Busy: Active-High indicator that the FIFO read domain is currently in a reset state.
		almost_empty	=> open,				-- 1-bit output: Almost Empty : When asserted, this signal indicates that only one more read can be performed before the FIFO goes to empty.
		data_valid		=> open,				-- 1-bit output: Read Data Valid: When asserted, this signal indicates that valid data is available on the output bus (dout).
		injectdbiterr	=> '0',					-- 1-bit input: Double Bit Error Injection: Injects a double bit error if the ECC feature is used on block RAMs or UltraRAM macros.
		injectsbiterr	=> '0',					-- 1-bit input: Single Bit Error Injection: Injects a single bit error if the ECC feature is used on block RAMs or UltraRAM macros.
		sbiterr			=> open,				-- 1-bit output: Single Bit Error: Indicates that the ECC decoder detected and fixed a single-bit error.
		dbiterr			=> open,				-- 1-bit output: Double Bit Error: Indicates that the ECC decoder detected a double-bit error and data in the FIFO core is corrupted.
		sleep			=> '0'					-- 1-bit input: Dynamic power saving: If sleep is High, the memory/fifo block is in power saving mode.
	);
	-- "wr_en"/"rd_en" should not be toggled when "rst" or "wr_rst_busy" or "rd_rst_busy" is asserted...
	wr_en_coord_s <= '1' when (wr_rst_busy_coord_s = '0' and full_coord_s = '0') else '0';
	rd_en_coord_s <= '1' when (rd_rst_busy_coord_s = '0' and empty_coord_s = '0' and read_fifo_i = '1') else '0';

	-- Coordinates are serialized to be write into (and read out to be extracted) within a dedicated FIFO
	-- NOTE: "signed" is used instead of "unsigned" because the coordinate value "-1" exists (only as intermediate value, anyway) 
	serial_img_coord_in_s(FIFO_X_COORD_WIDTH_C+FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C-1 downto FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C) <= std_logic_vector(to_signed(img_coord_i.x, FIFO_X_COORD_WIDTH_C));
	serial_img_coord_in_s(FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C-1 downto FIFO_Z_COORD_WIDTH_C)											 <= std_logic_vector(to_signed(img_coord_i.y, FIFO_Y_COORD_WIDTH_C));
	serial_img_coord_in_s(FIFO_Z_COORD_WIDTH_C-1 downto 0)																					 <= std_logic_vector(to_signed(img_coord_i.z, FIFO_Z_COORD_WIDTH_C));
	extr_img_coord_s <= (
		x => to_integer(signed(serial_img_coord_out_s(FIFO_X_COORD_WIDTH_C+FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C-1 downto FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C))),
		y => to_integer(signed(serial_img_coord_out_s(FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C-1 downto FIFO_Z_COORD_WIDTH_C))),
		z => to_integer(signed(serial_img_coord_out_s(FIFO_Z_COORD_WIDTH_C-1 downto 0))),
		t => to_integer(signed(serial_img_coord_out_s(FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C-1 downto FIFO_Z_COORD_WIDTH_C))) * NX_C + to_integer(signed(serial_img_coord_out_s(FIFO_X_COORD_WIDTH_C+FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C-1 downto FIFO_Y_COORD_WIDTH_C+FIFO_Z_COORD_WIDTH_C)))
	);

	-- Outputs
	enable_o	<= enable_s;
	img_coord_o	<= extr_img_coord_s	when read_fifo_i = '1' else reset_img_coord(SMPL_ORDER_G);	-- To avoid reading strange values out from the FIFOs...
	data_arr_o	<= data_arr_s		when read_fifo_i = '1' else (others => (others => '0'));	-- ...they are only outputted when "read_fifo_i = '1'"
	init_fifo_o	<= not(and(wr_rst_busy_data_arr_s) and and(rd_rst_busy_data_arr_s) and wr_rst_busy_coord_s and rd_rst_busy_coord_s);	-- To inform when all FIFOs are ready to use
end architecture behavioural;