--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		09/06/2021
--------------------------------------------------------------------------------
-- IP name:		top_encoder
--
-- Description: Top entity for the "encoder" module
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.types_predictor.all;
use work.utils_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;
use work.comp_encoder.all;

entity top_encoder is
	generic (
		-- 0: "signed" type samples, 1: "unsigned" type samples
		SMPL_TYPE_G				: std_logic;
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G			: std_logic_vector(1 downto 0);
		-- 0: Encoder Header disabled, 1: Encoder Header enabled
		HEADER_EN_G				: std_logic;
		-- 1: Full prediction mode, 0: Reduced prediction mode
		PREDICT_MODE_G			: std_logic;
		-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
		LSUM_TYPE_G				: std_logic_vector(1 downto 0);
		-- 1: Custom weight init, 0: Default weight init
		W_INIT_TYPE_G			: std_logic;
		-- 1: enabled, 0: disabled
		PER_ERR_LIM_UPD_G		: std_logic;
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G		: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G		: std_logic;
		REL_ERR_BAND_TYPE_G		: std_logic;
		-- 00: Sample-Adaptive Entropy, 01: Hybrid Entropy, 10: Block-Adaptive Entropy
		ENCODER_TYPE_G			: std_logic_vector(1 downto 0);
		-- Flag to add the "Weight Initialization Table"
		W_INIT_TABL_FLAG_G		: std_logic;
		-- Flag to add the "Weight Exponent Offset Table"
		W_EXP_OFF_TABL_FLAG_G	: std_logic;
		-- Flag to add the "Damping Table"
		DAMP_TABLE_FLAG_G		: std_logic;
		-- Flag to add the "Offset Table"
		OFFSET_TABLE_FLAG_G		: std_logic;
		-- Flag to add the "Accumulator Initialization Table"
		ACCU_INIT_TABLE_FLAG_G	: std_logic;
		-- 1: Restricted set of code options used, 0: Restricted set of code options NOT used
		RESTRICT_CODE_G			: std_logic;
		-- User Defined Data
		UDEF_DATA_G				: std_logic_vector(7 downto 0);
		-- Array with values -> 00: unsigned integer, 01: signed integer, 10: float
		SUPL_TABLE_TYPE_G		: supl_table_type_t;
		-- Array with values -> From 0 to 15 (check Table 3-1)
		SUPL_TABLE_PURPOSE_G	: supl_table_purpose_t;
		-- Array with values -> 00: zero-dimensional, 01: one-dimensional, 10: two-dimensional-zx, 11: two-dimensional-yx
		SUPL_TABLE_STRUCT_G		: supl_table_struct_t;
		-- Array with values -> Suplementary User-Defined Data
		SUPL_TABLE_UDATA_G		: supl_table_udata_t;
		-- 0: Little Endian order, 1: Big Endian order
		ENDIANNESS_G			: std_logic
	);
	port (
		clock_i					: in  std_logic;
		reset_i					: in  std_logic;

		enable_i				: in  std_logic;
		enable_o				: out std_logic;
		img_coord_i				: in  img_coord_t;
		img_coord_o				: out img_coord_t;
		
		init_fifo_o				: out std_logic;
		image_start_o			: out std_logic;			-- Flag to inform a new image is coming out
		image_end_o				: out std_logic;			-- Flag to inform current image has been finished
		
		err_lim_upd_en_i	    : in  std_logic;
		err_lim_val_i			: in  std_logic_vector(D_C-1 downto 0);
		data_mp_quan_i			: in  std_logic_vector(D_C-1 downto 0);
		valid_data_o			: out std_logic;			-- Flag to inform valid data is in the output line
		output_data_o			: out std_logic_vector(B_C*8-1 downto 0)
	);
end entity top_encoder;

architecture behavioural of top_encoder is
	-- Maximum number of bits (size) of every encoded data sample
	constant MAX_NUM_BITS_C		: integer	  := log2(Umax_C + D_C);

	-- Signals declaration
	signal enable_body_s		: std_logic   := '0';
	signal enable_top_s			: std_logic   := '0';
	signal img_coord_body_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord_top_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal enc_header_width_s	: integer	  := 0;
	signal enc_header_data_s	: std_logic_vector(4095 downto 0)			  := (others => '0');
	signal enc_body_num_bits_s	: std_logic_vector(MAX_NUM_BITS_C-1 downto 0) := (others => '0');
	signal enc_body_data_s		: std_logic_vector(Umax_C+D_C-1 downto 0)	  := (others => '0');
	signal init_fifo_s			: std_logic	  := '0';
	
	signal image_start_s		: std_logic	  := '0';
	signal image_end_s			: std_logic	  := '0';
	signal valid_data_s			: std_logic	  := '0';
	signal packed_data_s		: std_logic_vector(B_C*8-1 downto 0)		  := (others => '0');
	signal output_data_s		: std_logic_vector(B_C*8-1 downto 0)		  := (others => '0');

begin
	-- "Encoder Header" IP is only instantiated if configured so
	g_gen_header : if (HEADER_EN_G = '1') generate
		i_top_enc_header : top_enc_header
		generic map(
			SMPL_TYPE_G				=> SMPL_TYPE_G,
			SMPL_ORDER_G			=> SMPL_ORDER_G,
			PREDICT_MODE_G			=> PREDICT_MODE_G,
			LSUM_TYPE_G				=> LSUM_TYPE_G,
			W_INIT_TYPE_G			=> W_INIT_TYPE_G,
			PER_ERR_LIM_UPD_G		=> PER_ERR_LIM_UPD_G,
			FIDEL_CTRL_TYPE_G		=> FIDEL_CTRL_TYPE_G,
			ABS_ERR_BAND_TYPE_G		=> ABS_ERR_BAND_TYPE_G,
			REL_ERR_BAND_TYPE_G		=> REL_ERR_BAND_TYPE_G,
			ENCODER_TYPE_G			=> ENCODER_TYPE_G,
			W_INIT_TABL_FLAG_G		=> W_INIT_TABL_FLAG_G,
			W_EXP_OFF_TABL_FLAG_G	=> W_EXP_OFF_TABL_FLAG_G,
			DAMP_TABLE_FLAG_G		=> DAMP_TABLE_FLAG_G,
			OFFSET_TABLE_FLAG_G		=> OFFSET_TABLE_FLAG_G,
			ACCU_INIT_TABLE_FLAG_G	=> ACCU_INIT_TABLE_FLAG_G,
			RESTRICT_CODE_G			=> RESTRICT_CODE_G,
			UDEF_DATA_G				=> UDEF_DATA_G,
			SUPL_TABLE_TYPE_G		=> SUPL_TABLE_TYPE_G,
			SUPL_TABLE_PURPOSE_G	=> SUPL_TABLE_PURPOSE_G,
			SUPL_TABLE_STRUCT_G		=> SUPL_TABLE_STRUCT_G,
			SUPL_TABLE_UDATA_G		=> SUPL_TABLE_UDATA_G
		)
		port map(
			enc_header_width_o	=> enc_header_width_s,
			enc_header_data_o	=> enc_header_data_s
		);
	else generate
		enc_header_width_s <= 0;
		enc_header_data_s  <= (others => '0');
	end generate g_gen_header;
	
	-- "Encoder Body" IP instantiation
	i_top_enc_body : top_enc_body
	generic map(
		SMPL_ORDER_G		=> SMPL_ORDER_G,
		ENCODER_TYPE_G		=> ENCODER_TYPE_G,
		MAX_NUM_BITS_G		=> MAX_NUM_BITS_C,
		FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G
	)
	port map(
		clock_i				=> clock_i,
		reset_i				=> reset_i or image_end_s,
		
		enable_i			=> enable_i,
		enable_o			=> enable_body_s,
		img_coord_i			=> img_coord_i,
		img_coord_o			=> img_coord_body_s,
		
		err_lim_upd_en_i	=> err_lim_upd_en_i,
		err_lim_val_i		=> err_lim_val_i,
		data_mp_quan_i		=> data_mp_quan_i,
		
		enc_num_bits_o		=> enc_body_num_bits_s,
		enc_body_data_o		=> enc_body_data_s
	);
	
	-- "Packer" IP to take "Encoder Header & Body" IPs data and pack it
	i_packer : top_packer
	generic map(
		SMPL_ORDER_G		=> SMPL_ORDER_G,
		HEADER_EN_G			=> HEADER_EN_G,
		MAX_NUM_BITS_G		=> MAX_NUM_BITS_C
	)
	port map(
		clock_i				=> clock_i,
		reset_i				=> reset_i,

		enable_header_i		=> enable_i,
		enable_body_i		=> enable_body_s,
		enable_o			=> enable_top_s,
		img_coord_header_i	=> img_coord_i,
		img_coord_body_i	=> img_coord_body_s,
		img_coord_o			=> img_coord_top_s,
		
		enc_header_width_i	=> enc_header_width_s,
		enc_header_data_i	=> enc_header_data_s,
		enc_body_num_bits_i	=> enc_body_num_bits_s,
		enc_body_data_i		=> enc_body_data_s,
		init_fifo_o			=> init_fifo_s,
		
		image_start_o		=> image_start_s,
		image_end_o			=> image_end_s,
		valid_data_o		=> valid_data_s,
		packed_data_o		=> packed_data_s
	);
	
	-- Outputs the data with the right Endiannes
	p_endiannes : process(reset_i, packed_data_s) is
	begin
		if (reset_i = '1') then
			output_data_s <= (others => '0');
		else
			if (ENDIANNESS_G = '1' or B_C = 1) then		-- If 'Big Endian order' or only 'one-byte output width', no need to reorder anything
				output_data_s <= packed_data_s;
			else	-- ENDIANNESS_G = '0'				-- If 'Little Endian order'
				for i in 0 to B_C-1 loop
					output_data_s((B_C-i)*8-1 downto (B_C-i-1)*8) <= packed_data_s((i+1)*8-1 downto i*8);
				end loop;
			end if;
		end if;
	end process p_endiannes;
	
	-- Outputs
	enable_o		<= enable_top_s;
	img_coord_o		<= img_coord_top_s;
	init_fifo_o		<= init_fifo_s;
	image_start_o	<= image_start_s;
	image_end_o		<= image_end_s;
	valid_data_o	<= valid_data_s;
	output_data_o	<= output_data_s;
end architecture behavioural;