library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

-- Table E-2: Predictor Quantities
package param_predictor is

	constant P_C	 : integer range 0 to 15			:= work.utils_image.min(3, NZ_C);	-- Number of spectral bands used for prediction
	constant MAX_PZ_C: integer range 0 to 15			:= P_C;			-- Maximum number of previous spectral bands for prediction in band Z
	constant MAX_CZ_C: integer range 0 to MAX_PZ_C+3	:= P_C + 3;		-- Maximum number of local diff. values for prediction in band Z
	
	constant DA_C	 : integer range 1 to (work.utils_image.min(D_C-1, 16)) := 8;	-- Absolute error limit bit depth
	constant A_C	 : integer range 0 to (2**DA_C-1)	:= 10;			-- Absolute error limit constant
	constant DR_C	 : integer range 1 to (work.utils_image.min(D_C-1, 16)) := 8;	-- Relative error limit bit depth
	constant R_C	 : integer range 0 to (2**DR_C-1)	:= 20;			-- Relative error limit constant
	
	constant U_C	 : integer range 0 to 9 			:= 2;			-- Error limit update period exponent
	
	constant THETA_C : integer range 0 to 4 			:= 0;			-- Sample representative resolution
	constant FI_ARR_C : array_integer_t(0 to NZ_C-1) := (others => 0);	-- range 0 to (2**THETA_C-1). Sample representative damping
	constant PSI_ARR_C: array_integer_t(0 to NZ_C-1) := (others => 0);	-- range 0 to (2**THETA_C-1). Sample representative offset
	
	constant C_ARR_C : array_integer_t(0 to NZ_C-1)	 := (others => 0);	-- range -6 to 5. Intra-band weight exponent offsets
	constant Ci_MTX_C: matrix_integer_t(0 to NZ_C-1)(0 to MAX_PZ_C-1) := (others => (others => 0));	-- integer range -6 to 5. Inter-band weight exponent offsets
	
	-- Max. number for "V_MIN_C" should be "V_MAX_C"!!
	constant V_MIN_C : integer range -6 to 9 			:= -6;			-- Initial weight update scaling exponent parameters
	constant V_MAX_C : integer range V_MIN_C to 9 		:= 9;			-- Final weight update scaling exponent parameters
	constant T_INC_C : integer range 4 to 11	 		:= 4;			-- Weight update scaling exponent change interval (THIS NUMBER IS JUST THE POWER VALUE [2**T_INC_C])
	
	constant OMEGA_C : integer range 4 to 19			:= 19;			-- Weight resolution
	constant W_MIN_C : signed(OMEGA_C+3-1 downto 0)		:= (OMEGA_C+3-1 => '1', others => '0');	-- Minimum possible weight value
	constant W_MAX_C : signed(OMEGA_C+3-1 downto 0)		:= (OMEGA_C+3-1 => '0', others => '1');	-- Maximum possible weight value
	constant Q_C	 : integer range 3 to (OMEGA_C+3) 	:= 8;			-- Weight initialization resolution
	constant Re_C	 : integer range (work.utils_image.max(32, D_C+OMEGA_C+2)) to 64 := 64;		-- Register size in bits, used in prediction calculation

	-- Weight initialization vector (in two's complement)
	constant LAMBDA_MTX_C : matrix_signed_t(0 to NZ_C-1)(0 to MAX_CZ_C-1)(Q_C-1 downto 0) := (others => (others => (Q_C/2-1 => '1', others => '0')));

end package param_predictor;