--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		14/08/2021
--------------------------------------------------------------------------------
-- IP name:		tb_adder
--
-- Description: Testbench for the "adder" module
--
--
--------------------------------------------------------------------------------

-- Library instantiated, then declaration of files/functions to use
library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

-- A reference to VUnit (and related) set of libraries and packages
library vunit_lib;
context vunit_lib.vunit_context;
context vunit_lib.com_context;
context vunit_lib.data_types_context;

-- Components and packages related to the Image IP
use vunit_lib.param_image.all;
use vunit_lib.types_image.all;
use vunit_lib.utils_image.all;
use vunit_lib.comp_image.all;

-- Components and packages related to the Predictor IP
use vunit_lib.param_predictor.all;
use vunit_lib.types_predictor.all;
use vunit_lib.utils_predictor.all;
use vunit_lib.comp_predictor.all;

entity tb_adder is
    generic (
        encoded_tb_cfg	: string;
        runner_cfg		: string
    );
end entity tb_adder;

architecture behavioural of tb_adder is
	-- Record type to pack all signals (from Python script) together
	type tb_cfg_t is record
		SMPL_TYPE_G	 : std_logic;
		SMPL_ORDER_G : std_logic_vector(1 downto 0);
	end record tb_cfg_t;

	-- Function to decode the Python signals and connect them into the VHDL testbench
	impure function decode(encoded_tb_cfg : string) return tb_cfg_t is
	begin
	return (
		SMPL_TYPE_G  => std_logic'value(get(encoded_tb_cfg, "SMPL_TYPE_PY")),
		SMPL_ORDER_G => std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "SMPL_ORDER_PY")), 2))
	);
	end function decode;

	-- Constant of type "tb_cfg_t" and initialized by "decode()" function
	constant tb_cfg			: tb_cfg_t	:= decode(encoded_tb_cfg);
	
	signal clock_s			: std_logic	:= '0';
	signal reset_s			: std_logic := '1';
	
	signal img_en_s			: std_logic := '0';
	signal pred_en_s		: std_logic := '0';
	signal enable_out_s		: std_logic := '0';
	signal img_coord_in_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_out_s	: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal pred_coord_s		: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);

	signal data_s0_s		: std_logic_vector(D_C-1 downto 0)	 := (others => '0');	-- "sz(t)"  (original sample)
	signal data_s3_1s		: std_logic_vector(D_C+1-1 downto 0) := (others => '0');	-- "s^z(t)" (predicted sample)
	signal data_s3_2s		: std_logic_vector(D_C+1-1 downto 0) := (others => '0');	-- "s^z(t)" (predicted sample)
	signal data_s4_lsb_1s	: std_logic := '0';
	signal data_s4_lsb_2s	: std_logic := '0';
	signal en_res_s			: std_logic := '0';
	signal data_res_s		: std_logic_vector(D_C+2-1 downto 0) := (others => '0');	-- "/\z(t)" (prediction residual)

	signal flag_start, flag_stop : std_logic := '0';

begin
	reset_s <= '0' after 20 ns;
	clock_s <= not clock_s after 6.4 ns;	 -- Main clock frequency: 78125000 Hz
	
	-- Simulation MAIN process (if this one finishes, the whole simulation too)
	test_runner : process is
	begin
		-- Process starts by setting up VUnit using test_runner_setup procedure
		test_runner_setup(runner, runner_cfg);

		while test_suite loop	-- Pay very attention to the testcase name, arrows or accents prevent the system to work...
			if run("Top Adder Block") then
				info("Running test case = " & to_string(running_test_case));

				flag_start <= '1';
			end if;
			
			wait until (flag_stop = '1');
		end loop;

		-- The process ends with the test_runner_cleanup procedure which will force the simulation to stop
		test_runner_cleanup(runner);
	end process;

	-- Set Watchdog with limited time (this number must be higher as any "wait for" statement used in the testbench)
	test_runner_watchdog(runner, 1 ms);

	-- Input signals stimulus calculation
	p_inputs_update : process(clock_s) is
	begin
		if (rising_edge(clock_s)) then
			if (reset_s = '1') then
				pred_en_s		<= '0';
				pred_coord_s	<= reset_img_coord(tb_cfg.SMPL_ORDER_G);
				data_s0_s		<= (others => '0');
				data_s3_1s		<= (others => '0');
				data_s4_lsb_1s	<= '0';
			else
				if (flag_start = '1') then
					if ((img_coord_out_s.x < NX_C-1) or (img_coord_out_s.y < NY_C-1) or (img_coord_out_s.z < NZ_C-1)) then
						pred_en_s		<= enable_out_s;
						pred_coord_s	<= img_coord_out_s;
						
						if (data_s3_1s = (data_s3_1s'length-1 downto 0 => '1')) then
							data_s3_1s	<= (others => '0');		
						else
							data_s3_1s	<= std_logic_vector(resize(signed(data_s3_1s) + to_signed(3, data_s3_1s'length), data_s3_1s'length));
						end if;
						
						if (data_s0_s = (data_s0_s'length-1 downto 0 => '1')) then
							data_s0_s	<= (others => '0');		
						else		-- "data_s0_s" should update faster than "data_s3_1s"
							data_s0_s	<= std_logic_vector(resize(signed(data_s0_s) + to_signed(5, data_s0_s'length), data_s0_s'length));
						end if;
						
						if (img_coord_out_s.z = 0) then
							data_s4_lsb_1s <= not data_s4_lsb_1s;
						end if;
					end if;
				end if;
			end if;	
		end if;
	end process p_inputs_update;
	
	-- Process to stop the simulation when finished (to use 'wait' statements, the process cannot be clocked)
	p_stop_sim : process is
	begin
		-- After one complete image, wait statement and request to finish the simulation
		wait until (en_res_s = '1' and (img_coord_out_s.x >= NX_C-1) and (img_coord_out_s.y >= NY_C-1) and (img_coord_out_s.z >= NZ_C-1));
		wait for 100 ns;
		flag_stop <= '1';
		
		-- Two possibilities to stop the simulation without VUnit framework (only the "assert" line really works, anyway)
		-- assert FALSE Report "Simulation Finished" severity FAILURE;
		-- std.env.finish;
	end process p_stop_sim;	
	
	-- Entity to control the image coordinates
	i_image_coord_control : image_coord_control
	generic map(
		SMPL_ORDER_G => tb_cfg.SMPL_ORDER_G
	)
	port map(
		clock_i		 => clock_s,
		reset_i		 => reset_s,

		enable_i	 => flag_start,
		enable_o	 => img_en_s,
		
		image_end_i	 => '0',
		img_coord_o	 => img_coord_in_s
	);

	-- Mapper top entity
	i_adder : adder
	generic map(
		SMPL_ORDER_G	=> tb_cfg.SMPL_ORDER_G,
		SMPL_TYPE_G		=> tb_cfg.SMPL_TYPE_G
	)	
	port map(
		clock_i			=> clock_s,
		reset_i			=> reset_s,
		
		img_en_i		=> img_en_s,
		enable_o		=> enable_out_s,
		img_coord_i		=> img_coord_in_s,
		img_coord_o		=> img_coord_out_s,
		
		pred_en_i		=> pred_en_s,
		pred_coord_i	=> pred_coord_s,
		data_s3_i		=> data_s3_1s,
		data_s3_o		=> data_s3_2s,
		data_s4_lsb_i	=> data_s4_lsb_1s,
		data_s4_lsb_o	=> data_s4_lsb_2s,
		
		data_s0_i		=> data_s0_s,
		en_res_o		=> en_res_s,
		data_res_o		=> data_res_s
	);

end architecture behavioural;