--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		01/11/2020
--------------------------------------------------------------------------------
-- IP name:		dbl_res_pred_error
--
-- Description: Computes the double-resolution prediction error "ez(t)"
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.utils_predictor.all;

entity dbl_res_pred_error is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	: std_logic_vector(1 downto 0)
	);
	port (
		clock_i			: in  std_logic;
		reset_i			: in  std_logic;
		
		enable_i		: in  std_logic;
		enable_o		: out std_logic;
		img_coord_i		: in  img_coord_t;
		img_coord_o		: out img_coord_t;
		
		data_s1_i		: in  std_logic_vector(D_C-1 downto 0);		-- "s'z(t)" (clipped quantizer bin center)
		data_s4_i		: in  std_logic_vector(D_C+2-1 downto 0);	-- "s~z(t)" (double-resolution predicted sample)
		pred_err_bit_o	: out integer range -1 to 1					-- "ez(t)"  (double-resolution prediction error) bit
	);
end entity dbl_res_pred_error;

architecture behavioural of dbl_res_pred_error is
	signal enable_s			: std_logic	  := '0';
	signal img_coord_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal pred_err_val_s	: std_logic_vector(D_C+2-1 downto 0) := (others => '0');
	signal pred_err_bit_s	: integer range -1 to 1 := 1;
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_data_pred_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_data_pred_delay;
	
	-- Double-resolution prediction error value (ez(t)) calculation
	p_pred_err_calc : process(clock_i) is
		variable pred_err_val_v : std_logic_vector(D_C+2-1 downto 0) := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then					
				pred_err_val_v := (others => '0');
				pred_err_val_s <= (others => '0');
				pred_err_bit_s <= 1;
			else
				if (enable_i = '1') then
					pred_err_val_v := std_logic_vector(resize(shift_sgn(resize(signed(data_s1_i), D_C+4), 1) - resize(signed(data_s4_i), D_C+4), D_C+2));
					pred_err_val_s <= pred_err_val_v;
					pred_err_bit_s <= sgnp(signed(pred_err_val_v));
				end if;
			end if;
		end if;
	end process p_pred_err_calc;
	
	-- Outputs
	enable_o		<= enable_s;
	img_coord_o		<= img_coord_s;
	pred_err_bit_o	<= pred_err_bit_s;
end architecture behavioural;