--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		01/11/2020
--------------------------------------------------------------------------------
-- IP name:		high_res_pred_smpl
--
-- Description: High-resolution predicted sample value "s)z(t)" calculation
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.types_predictor.all;
use work.utils_predictor.all;
use work.param_predictor.all;

entity high_res_pred_smpl is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G : std_logic_vector(1 downto 0);
		-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
		SMPL_LIMIT_G : smpl_lim_t
	);
	port (
		clock_i		: in  std_logic;
		reset_i		: in  std_logic;
		
		enable_i	: in  std_logic;
		enable_o	: out std_logic;
		img_coord_i	: in  img_coord_t;
		img_coord_o	: out img_coord_t;
																-- "d^z(t)" (predicted central local difference)
		data_pred_cldiff_i : in std_logic_vector(MAX_CZ_C+D_C+4+OMEGA_C+3-1 downto 0);
		data_lsum_i	: in  std_logic_vector(D_C+4-1 downto 0);	-- "σz(t)"  (local sum)
		data_s6_o	: out std_logic_vector(Re_C-1 downto 0)		-- "s)z(t)" (high-resolution predicted sample)
	);
end entity high_res_pred_smpl;

architecture behavioural of high_res_pred_smpl is
	constant PW_OM1_C	: std_logic_vector(OMEGA_C+3-1 downto 0) := (OMEGA_C+1 => '1', others => '0');
	
	signal enable_s		: std_logic := '0';
	signal img_coord_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal data_s6_s	: std_logic_vector(Re_C-1 downto 0) := (others => '0');
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_high_res_pred_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_high_res_pred_delay;
	
	-- High-resolution predicted sample value (s)z(t)) calculation
	p_high_res_pred_smpl_calc : process(clock_i) is
		variable comp1_v : std_logic_vector(D_C+6-1 downto 0)	:= (others => '0');
		variable comp2_v : std_logic_vector(Re_C+4-1 downto 0)	:= (others => '0');
		variable comp3_v, comp4_v, comp5_v, comp6_v : std_logic_vector(Re_C-1 downto 0) := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_v   := (others => '0');
				comp2_v   := (others => '0');
				comp3_v   := (others => '0');
				comp4_v   := (others => '0');
				comp5_v   := (others => '0');
				comp6_v   := (others => '0');
				data_s6_s <= (others => '0');
			else
				if (enable_i = '1') then
					comp1_v	  := std_logic_vector(resize(resize(signed(data_lsum_i), D_C+6) - shift_sgn(resize(signed(SMPL_LIMIT_G.mid), D_C+6), 2), D_C+6));
					comp2_v   := std_logic_vector(resize(resize(signed(data_pred_cldiff_i), Re_C+4) + shift_sgn(resize(signed(comp1_v), Re_C+4), OMEGA_C), Re_C+4));
					comp3_v   := std_logic_vector(mod_R(signed(comp2_v), Re_C));
					comp4_v	  := std_logic_vector(resize(resize(signed(comp3_v), Re_C) + shift_sgn(resize(signed(SMPL_LIMIT_G.mid), Re_C), OMEGA_C+2) + resize(signed(PW_OM1_C), Re_C), Re_C));
					comp5_v	  := std_logic_vector(resize(shift_sgn(resize(signed(SMPL_LIMIT_G.min), Re_C), OMEGA_C+2), Re_C));
					comp6_v	  := std_logic_vector(resize(shift_sgn(resize(signed(SMPL_LIMIT_G.max), Re_C), OMEGA_C+2) + resize(signed(PW_OM1_C), Re_C), Re_C));
					data_s6_s <= std_logic_vector(clip(signed(comp4_v), signed(comp5_v), signed(comp6_v)));
				end if;
			end if;
		end if;
	end process p_high_res_pred_smpl_calc;

	-- Outputs
	enable_o	<= enable_s;
	img_coord_o	<= img_coord_s;
	data_s6_o	<= data_s6_s;
end architecture behavioural;