--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		24/10/2020
--------------------------------------------------------------------------------
-- IP name:		local_differences
--
-- Description: Computes the local differences
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.types_predictor.all;
use work.utils_predictor.all;

entity local_differences is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	: std_logic_vector(1 downto 0);
		-- 1: Full prediction mode, 0: Reduced prediction mode
		PREDICT_MODE_G	: std_logic
	);
	port (
		clock_i		  : in  std_logic;
		reset_i		  : in  std_logic;
		
		enable_i	  : in  std_logic;
		enable_o	  : out std_logic;
		img_coord_i	  : in  img_coord_t;
		img_coord_o	  : out img_coord_t;
		
		data_lsum_i	  : in  std_logic_vector(D_C+4-1 downto 0);
		data_s2_pos_i : in  s2_pos_t;
		ldiff_pos_o	  : out ldiff_pos_t
	);
end entity local_differences;

architecture behavioural of local_differences is
	signal enable_s		: std_logic	  := '0';
	signal img_coord_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal cldiff_s		: std_logic_vector(D_C+4-1 downto 0) := (others => '0');
	signal nldiff_s		: std_logic_vector(D_C+4-1 downto 0) := (others => '0');
	signal wldiff_s		: std_logic_vector(D_C+4-1 downto 0) := (others => '0');
	signal nwldiff_s	: std_logic_vector(D_C+4-1 downto 0) := (others => '0');

begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_local_diff_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_local_diff_delay;
	
	-- Central local difference calculation
	p_cnt_ldiff_calc : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				cldiff_s <= (others => '0');
			else
				if (enable_i = '1') then
					if (img_coord_i.t > 0 and img_coord_i.z /= NZ_C-1) then
						cldiff_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.cur), D_C+4), 2) - signed(data_lsum_i), D_C+4));
					else		-- Case X=0 and Y=0 here defined, but anyway not used later on...
						cldiff_s <= (others => '0');
					end if;
				end if;
			end if;
		end if;
	end process p_cnt_ldiff_calc;
	
	-- Directional local difference are only used under "Full Prediction Mode"
	g_direct_ldiff_calc : if (PREDICT_MODE_G = '1') generate
		-- North local difference calculation
		p_nldiff_calc : process(clock_i) is
		begin
			if rising_edge(clock_i) then
				if (reset_i = '1') then
					nldiff_s <= (others => '0');
				else
					if (enable_i = '1') then
						if (img_coord_i.t > 0) then
							if (img_coord_i.y > 0) then
								nldiff_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.n), D_C+4), 2) - signed(data_lsum_i), D_C+4));
							else	-- img_coord_i.y = 0
								nldiff_s <= (others => '0');
							end if;
						else		-- Case X=0 and Y=0 here defined, but anyway not used later on...
							nldiff_s <= (others => '0');
						end if;
					end if;
				end if;
			end if;
		end process p_nldiff_calc;
		
		-- West local difference calculation
		p_wldiff_calc : process(clock_i) is
		begin
			if rising_edge(clock_i) then
				if (reset_i = '1') then
					wldiff_s <= (others => '0');
				else
					if (enable_i = '1') then
						if (img_coord_i.t > 0) then
							if (img_coord_i.y > 0 and img_coord_i.x > 0) then
								wldiff_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.w), D_C+4), 2) - signed(data_lsum_i), D_C+4));
							elsif (img_coord_i.y > 0 and img_coord_i.x = 0) then
								wldiff_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.n), D_C+4), 2) - signed(data_lsum_i), D_C+4));
							else	-- img_coord_i.y = 0
								wldiff_s <= (others => '0');
							end if;
						else		-- Case X=0 and Y=0 here defined, but anyway not used later on...
							wldiff_s <= (others => '0');
						end if;
					end if;
				end if;
			end if;
		end process p_wldiff_calc;
		
		-- North-West local difference calculation
		p_nwldiff_calc : process(clock_i) is
		begin
			if rising_edge(clock_i) then
				if (reset_i = '1') then
					nwldiff_s <= (others => '0');
				else
					if (enable_i = '1') then
						if (img_coord_i.t > 0) then
							if (img_coord_i.y > 0 and img_coord_i.x > 0) then
								nwldiff_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.nw), D_C+4), 2) - signed(data_lsum_i), D_C+4));
							elsif (img_coord_i.y > 0 and img_coord_i.x = 0) then
								nwldiff_s <= std_logic_vector(resize(shift_sgn(resize(signed(data_s2_pos_i.n), D_C+4), 2) - signed(data_lsum_i), D_C+4));
							else	-- img_coord_i.y = 0
								nwldiff_s <= (others => '0');
							end if;
						else		-- Case X=0 and Y=0 here defined, but anyway not used later on...
							nwldiff_s <= (others => '0');
						end if;
					end if;
				end if;
			end if;
		end process p_nwldiff_calc;
	else generate		-- "Reduced Prediction Mode"
		nldiff_s  <= (others => '0');
		wldiff_s  <= (others => '0');
		nwldiff_s <= (others => '0');
	end generate g_direct_ldiff_calc;
	
	-- Outputs
	enable_o	<= enable_s;
	img_coord_o	<= img_coord_s;
	ldiff_pos_o <= (
		c	=> cldiff_s,
		n	=> nldiff_s,
		w	=> wldiff_s,
		nw	=> nwldiff_s
	);
end architecture behavioural;