--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		24/10/2020
--------------------------------------------------------------------------------
-- IP name:		shift_register
--
-- Description: Array to delay output values a certain time.
--				Delay time = array size, and with a new value per clock cycle.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.types_image.all;
use work.utils_image.all;

entity shift_register is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G : std_logic_vector(1 downto 0);
		DATA_SIZE_G	 : integer;
		REG_SIZE_G	 : integer;
		INIT_VAL_G	 : std_logic_vector(DATA_SIZE_G-1 downto 0)
	);
	port (
		clock_i		: in  std_logic;
		reset_i		: in  std_logic;
		
		enable_i	: in  std_logic;
		enable_o	: out std_logic;
		img_coord_i	: in  img_coord_t;
		img_coord_o	: out img_coord_t;

		data_i		: in  std_logic_vector(DATA_SIZE_G-1 downto 0);
		data_o		: out std_logic_vector(DATA_SIZE_G-1 downto 0)
	);
end entity shift_register;

architecture behavioural of shift_register is
	-- '1' is subtracted here because there is already one extra cycle delay to output the signal
	constant REG_SIZE_C		: integer	  := iif(REG_SIZE_G-1 > 0, REG_SIZE_G-1, 0);
	-- Constant to avoid the array "data_arr_s" crashing if the size is 0
	constant SECU_SIZE_C	: integer	  := iif(REG_SIZE_C > 0, REG_SIZE_C, 2);
	
	signal enable_s			: std_logic   := '0';
	signal img_coord_s		: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal data_s			: std_logic_vector(DATA_SIZE_G-1 downto 0)				  := INIT_VAL_G;
	
	signal counter_s		: integer range 0 to REG_SIZE_C							  := 0;
	-- Initial read index = lowest position / Initial write index = highest position
	signal enable_arr_s		: std_logic_vector(SECU_SIZE_C-1 downto 0)				  := (others => '0');
	signal img_coord_arr_s	: array_img_coord_t(0 to SECU_SIZE_C-1)					  := (others => reset_img_coord(SMPL_ORDER_G));
	signal data_arr_s		: array_slv_t(0 to SECU_SIZE_C-1)(DATA_SIZE_G-1 downto 0) := (others => INIT_VAL_G);

begin
	p_shift_reg_array : process(clock_i) is
	begin
		if (rising_edge(clock_i)) then
			if (reset_i = '1') then
				enable_s		<= '0';
				img_coord_s		<= reset_img_coord(SMPL_ORDER_G);
				data_s			<= INIT_VAL_G;
				counter_s		<= 0;
				enable_arr_s	<= (others => '0');
				img_coord_arr_s	<= (others => reset_img_coord(SMPL_ORDER_G));
				data_arr_s		<= (others => INIT_VAL_G);
			else
				if (REG_SIZE_C > 0) then
					-- Shift register values are moved to the right, and incoming value goes to the lowest position (left)
					-- NOTE: Directions for the array are the opposite in respect of normal std_logic_vector
					enable_arr_s(enable_arr_s'high downto 1)	<= enable_arr_s(enable_arr_s'high-1 downto 0);
					img_coord_arr_s(1 to img_coord_arr_s'high)	<= img_coord_arr_s(0 to img_coord_arr_s'high-1);
					data_arr_s(1 to data_arr_s'high)			<= data_arr_s(0 to data_arr_s'high-1);
					enable_arr_s(0)		<= enable_i;
					img_coord_arr_s(0)	<= img_coord_i;
					data_arr_s(0)		<= data_i;
					-- Outcoming value is the highest position from the shift register
					enable_s	<= enable_arr_s(enable_arr_s'high);
					img_coord_s	<= img_coord_arr_s(img_coord_arr_s'high);
					data_s		<= data_arr_s(data_arr_s'high);
				else
					-- If delay is just 1 clock cycle, input is directly outputted
					enable_s	<= enable_i;
					img_coord_s	<= img_coord_i;
					data_s		<= data_i;
				end if;
				
				if (counter_s < REG_SIZE_C) then
					counter_s <= counter_s + 1;
				else
					counter_s <= 0;
				end if;
			end if;
		end if;
	end process p_shift_reg_array;
	
	-- Outputs
	enable_o	<= enable_s;
	img_coord_o	<= img_coord_s;
	data_o		<= data_s;
end architecture behavioural;