--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		01/11/2020
--------------------------------------------------------------------------------
-- IP name:		weight_upd_scal_exp
--
-- Description: Computes the weight update scaling exponent "p(t)"
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;

entity weight_upd_scal_exp is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G : std_logic_vector(1 downto 0)
	);
	port (
		clock_i		 : in  std_logic;
		reset_i		 : in  std_logic;
		
		enable_i	 : in  std_logic;
		enable_o	 : out std_logic;
		img_coord_i	 : in  img_coord_t;
		img_coord_o	 : out img_coord_t;
		
		w_scal_exp_o : out integer range -127 to 127		-- "p(t)" (weight update scaling exponent)
	);
end entity weight_upd_scal_exp;

architecture behavioural of weight_upd_scal_exp is
	-- Constants declaration
	constant T_INC_VAL_C : integer		:= 2**T_INC_C;	-- Unlike docu, T_INC_C is just the power value, not the final one, so always use 2**T_INC_C
	
	-- Signals declaration
	signal enable_s		 : std_logic	:= '0';
	signal img_coord_s	 : img_coord_t	:= reset_img_coord(SMPL_ORDER_G);
	
	signal w_scal_exp_s	 : integer range -127 to 127 := V_MIN_C + D_C - OMEGA_C;
	
begin
	-- Input values delayed to synchronize them with the next modules in chain
	p_w_scal_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_w_scal_delay;
	
	-- Weight update scaling exponent value (p(t)) calculation	
	p_w_scal_exp_calc : process(clock_i) is
		variable comp1_v : integer range -NX_C*NY_C to NX_C*NY_C := 0;
		variable comp2_v : integer range V_MIN_C	to V_MAX_C	 := 0;
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_v		 := 0;
				comp2_v		 := 0;
				w_scal_exp_s <= V_MIN_C + D_C - OMEGA_C;
			else
				if (enable_i = '1') then
					if (img_coord_i.t > 0) then
						comp1_v		 := division_down(img_coord_i.t - NX_C, T_INC_VAL_C);
						comp2_v		 := clip(V_MIN_C + comp1_v, V_MIN_C, V_MAX_C);
						w_scal_exp_s <= comp2_v + D_C - OMEGA_C;
					else		-- This case is not defined on documentation, so likely not used later on
						comp1_v		 := 0;
						comp2_v		 := 0;
						w_scal_exp_s <= V_MIN_C + D_C - OMEGA_C;
					end if;
				end if;
			end if;
		end if;
	end process p_w_scal_exp_calc;

	-- Outputs
	enable_o	 <= enable_s;
	img_coord_o	 <= img_coord_s;
	w_scal_exp_o <= w_scal_exp_s;
end architecture behavioural;