--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		06/11/2020
--------------------------------------------------------------------------------
-- IP name:		weights_vector
--
-- Description: Computes the weight vector "Wz(t)"
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;
use work.comp_predictor.all;

entity weights_vector is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G	: std_logic_vector(1 downto 0);
		-- 1: Full prediction mode, 0: Reduced prediction mode
		PREDICT_MODE_G	: std_logic;
		-- 1: Custom weight init, 0: Default weight init
		W_INIT_TYPE_G	: std_logic
	);
	port (
		clock_i			: in  std_logic;
		reset_i			: in  std_logic;
		
		enable_i		: in  std_logic;
		enable_o		: out std_logic;
		img_coord_i		: in  img_coord_t;
		img_coord_o		: out img_coord_t;
		
		pred_err_bit_i	: in  integer range -1 to 1;							 -- "ez(t)" (double-resolution prediction error) bit
		w_scal_exp_i	: in  integer range -127 to 127;						 -- "p(t)"  (weight update scaling exponent)
		ldiff_vect_i	: in  array_slv_t(0 to MAX_CZ_C-1)(D_C+4-1 downto 0);	 -- "Uz(t)" (local difference vector)
		weight_vect_o	: out array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0) -- "Wz(t)" (weight vector)
	);
end entity weights_vector;

architecture behavioural of weights_vector is
	-- Default weight values initialization function
	pure function init_def_weight_vec return array_slv_t is
		variable def_weight_vect_v : array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0) := (others => (others => '0'));
	begin
		-- Directional weight values (wN, wW, wNW) are all set to 0
		def_weight_vect_v(0 to 2) := (others => (others => '0'));

		-- Normal weight values (OMEGA_C will never surpass 32 bits, so one can work here with integer values...)
		def_weight_vect_v(3) := std_logic_vector(to_unsigned(division_down(7 * (2**OMEGA_C), 8), OMEGA_C+3));
		for i in 4 to (def_weight_vect_v'length-1) loop
			def_weight_vect_v(i) := std_logic_vector(resize(division_down(unsigned(def_weight_vect_v(i-1)), to_unsigned(8, OMEGA_C+3)), OMEGA_C+3));
		end loop;

		return def_weight_vect_v;
	end function init_def_weight_vec;

	-- Custom weight values initialization function
	pure function init_cust_weight_vec(coord_z : in integer) return array_slv_t is
		constant PW_OM2Q_C : signed(OMEGA_C+4-Q_C-1 downto 0) := (OMEGA_C+2-Q_C => '1', others => '0');
		
		variable cust_weight_vect_v : array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0) := (others => (others => '0'));
	begin
		for i in 0 to (cust_weight_vect_v'length-1) loop	-- Here there was an useless "round_up" (without division), so it was removed
			cust_weight_vect_v(i) := std_logic_vector(resize(shift_sgn(resize(LAMBDA_MTX_C(coord_z)(i), OMEGA_C+3), OMEGA_C+3-Q_C) + resize(PW_OM2Q_C, OMEGA_C+3) - to_signed(1, OMEGA_C+3), OMEGA_C+3));
		end loop;

		return cust_weight_vect_v;
	end function init_cust_weight_vec;
	
	-- Number of delay samples to get t-1 (or W position), for BSQ/BIP/BIL orders (delays table defined inside "samples_store" IP)
	constant NEXT_W_C			 : integer		:= locate_position(SMPL_ORDER_G, 1, NZ_C, 1);
	
	-- Signals declaration
	signal enable_s				 : std_logic	:= '0';
	signal img_coord_s			 : img_coord_t	:= reset_img_coord(SMPL_ORDER_G);

	-- The maximum number of previous spectral bands for prediction (MAX_PZ_C+3) are calculated, but only some of them (until "pz_i") will be used later
	signal weight_vect_s		 : array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0) := (others => (others => '0'));
	signal weight_vect_old_s	 : array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0) := (others => (others => '0'));
	signal weight_vect_ser_s	 : std_logic_vector(MAX_CZ_C*(OMEGA_C+3)-1 downto 0)  := (others => '0');
	signal weight_vect_old_ser_s : std_logic_vector(MAX_CZ_C*(OMEGA_C+3)-1 downto 0)  := (others => '0');
	signal weight_vect_next_s	 : array_slv_t(0 to MAX_CZ_C-1)(OMEGA_C+3-1 downto 0) := (others => (others => '0'));
	signal weight_vect_next_ser_s: std_logic_vector(MAX_CZ_C*(OMEGA_C+3)-1 downto 0)  := (others => '0');

begin
	-- Input values delayed one clock cycle to synchronize them with the next modules in chain
	p_weight_vect_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s	<= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_weight_vect_delay;

	-- Shift register to compute "t-1" from "weight_vect_s" signal, depending on the samples order
	i_shift_weight_old_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => MAX_CZ_C * (OMEGA_C+3),
		REG_SIZE_G	 => iif(NEXT_W_C <= 2, 8+3, 8+NZ_C+1),	-- Magic number
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => weight_vect_ser_s,
		data_o		 => weight_vect_old_ser_s
	);
	-- Functions to serialize/parallelize the 'weight vector' and 'old weight vector'
	weight_vect_ser_s <= arr_slv_2_sgl_slv(weight_vect_s);
	weight_vect_old_s <= sgl_slv_2_arr_slv(weight_vect_old_ser_s, OMEGA_C+3);
	
	-- Weight vector value (Wz(t)) calculation
	p_weight_vect_calc : process(clock_i) is
		variable comp1_arr_v : array_integer_t(0 to MAX_CZ_C-1)							:= (others => 0);	-- range -6 to 6
		variable comp2_arr_v : array_signed_t(0 to MAX_CZ_C-1)(OMEGA_C+D_C-1 downto 0)	:= (others => (others => '0'));
		variable comp3_arr_v : array_signed_t(0 to MAX_CZ_C-1)(OMEGA_C+D_C-1 downto 0)	:= (others => (others => '0'));
		variable comp4_arr_v : array_signed_t(0 to MAX_CZ_C-1)(OMEGA_C+D_C-1 downto 0)	:= (others => (others => '0'));
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_arr_v   := (others => 0);
				comp2_arr_v   := (others => (others => '0'));
				comp3_arr_v	  := (others => (others => '0'));
				comp4_arr_v	  := (others => (others => '0'));
				weight_vect_s <= (others => (others => '0'));
			else
				if (enable_i = '1') then
					-- Condition "z>0 and 'Reduced Prediction Mode'" not included here, as already added in the next IP in chain...
					if (img_coord_i.t = 0) then
						if (W_INIT_TYPE_G = '1') then
							weight_vect_s <= init_cust_weight_vec(img_coord_i.z);
						else	-- W_INIT_TYPE_G = '0'
							weight_vect_s <= init_def_weight_vec;
						end if;
					else	-- img_coord_i.t > 0
						for i in 0 to weight_vect_s'length-1 loop
							-- Under "Reduced prediction mode", ONLY directional weight values (wN, wW, wNW) are directly set to 0
							if (PREDICT_MODE_G = '0' and i <= 2) then
								weight_vect_s(i)	<= (others => '0');
							else	-- Central weight values (or the whole array under "Full prediction mode") calculation
								if (i <= 2) then
									comp1_arr_v(i)	:= -1 * (w_scal_exp_i + C_ARR_C(img_coord_i.z));
								else	-- i > 2
									-- 'i-3' refers to the difference between 'MAX_CZ_C' and 'MAX_PZ_C'
									comp1_arr_v(i)	:= -1 * (w_scal_exp_i + Ci_MTX_C(img_coord_i.z)(i-3));
								end if;
								comp2_arr_v(i)		:= resize(to_signed(pred_err_bit_i, D_C+4) * signed(ldiff_vect_i(i)), OMEGA_C+D_C);
								comp3_arr_v(i)		:= shift_sgn(comp2_arr_v(i), comp1_arr_v(i));
								comp4_arr_v(i)		:= shift_sgn(resize(comp3_arr_v(i) + to_signed(1, OMEGA_C+D_C), OMEGA_C+D_C), -1);
								weight_vect_s(i)	<= std_logic_vector(resize(clip(resize(resize(signed(weight_vect_old_s(i)), OMEGA_C+D_C) + resize(signed(comp4_arr_v(i)), OMEGA_C+D_C), OMEGA_C+D_C), resize(W_MIN_C, OMEGA_C+D_C), resize(W_MAX_C, OMEGA_C+D_C)), OMEGA_C+3));
							end if;
						end loop;
					end if;
				end if;
			end if;
		end if;
	end process p_weight_vect_calc;
	
	-- Delays the output as requested to match Johann's code
	i_shift_weight_next_t : shift_register
	generic map(
		SMPL_ORDER_G => SMPL_ORDER_G,
		DATA_SIZE_G	 => MAX_CZ_C * (OMEGA_C+3),
		REG_SIZE_G	 => iif(NEXT_W_C <= 2, 4+3, 4+NZ_C+1),	-- Magic number
		INIT_VAL_G	 => (others => '0')
	)
	port map(
		clock_i		 => clock_i,
		reset_i		 => reset_i,
		
		enable_i	 => enable_i,
		enable_o	 => open,
		img_coord_i	 => img_coord_i,
		img_coord_o	 => open,

		data_i		 => weight_vect_ser_s,
		data_o		 => weight_vect_next_ser_s
	);
	-- Function to parallelize the 'weight vector'
	weight_vect_next_s <= sgl_slv_2_arr_slv(weight_vect_next_ser_s, OMEGA_C+3);
	
	-- Outputs
	enable_o	  <= enable_s;
	img_coord_o	  <= img_coord_s;
	weight_vect_o <= weight_vect_next_s;
end architecture behavioural;