--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		30/10/2020
--------------------------------------------------------------------------------
-- IP name:		fidelity_ctrl
--
-- Description: Controls the step size via an absolute and/or a relative error
--				limit (so samples can be reconstructed later with some tolerance)
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;
use work.error_limit_values_table.all;

entity fidelity_ctrl is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic;
		-- 1: enabled, 0: disabled
		PER_ERR_LIM_UPD_G	: std_logic
	);
	port (
		clock_i		: in  std_logic;
		reset_i		: in  std_logic;
		
		enable_i	: in  std_logic;
		enable_o	: out std_logic;
		img_coord_i	: in  img_coord_t;
		img_coord_o	: out img_coord_t;
		
		data_s3_i	: in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
		data_merr_o	: out std_logic_vector(D_C+1-1 downto 0)	-- "mz(t)"  (maximum error)
	);
end entity fidelity_ctrl;

architecture behavioural of fidelity_ctrl is
	signal enable_s		: std_logic   := '0';
	signal img_coord_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal data_merr_s	: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal err_lim_pos_s: integer range 0 to NY_C := 0;
	signal Az_s			: integer := 0; -- range 0 to (2**DA_C-1) := 0;		-- SIMULATION FAILS WITH THE "range" OPTION!
	signal Rz_s			: integer := 0;	-- range 0 to (2**DR_C-1) := 0;

begin	
	-- Input values delayed to synchronize them with the next modules in chain
	p_fidelity_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable_s	<= '0';
				img_coord_s <= reset_img_coord(SMPL_ORDER_G);
			else
				enable_s	<= enable_i;
				img_coord_s	<= img_coord_i;
			end if;
		end if;
	end process p_fidelity_delay;

	-- If configured so, error limit values are updated (every 2**U_C frames) inside "image_coord_control" IP. Thus, here the incoming data is simply used according to local config.
	g_err_lim_def : if (FIDEL_CTRL_TYPE_G /= LOSSLESS_C) generate
		g_err_lim_upd_en : if (PER_ERR_LIM_UPD_G = '1') generate
			err_lim_pos_s <= img_coord_i.y / (2**U_C);
		else generate	-- PER_ERR_LIM_UPD_G = '0'
			err_lim_pos_s <= 0;
		end generate g_err_lim_upd_en;
		Az_s <= ERR_LIM_VAL_ARR_C(err_lim_pos_s).abs_arr(img_coord_i.z) when (enable_i = '1' and ABS_ERR_BAND_TYPE_G = '1') else ERR_LIM_VAL_ARR_C(err_lim_pos_s).abs_c;
		Rz_s <= ERR_LIM_VAL_ARR_C(err_lim_pos_s).rel_arr(img_coord_i.z) when (enable_i = '1' and REL_ERR_BAND_TYPE_G = '1') else ERR_LIM_VAL_ARR_C(err_lim_pos_s).rel_c;
	else generate
		err_lim_pos_s <= 0;
		Az_s <= 0;
		Rz_s <= 0;
	end generate g_err_lim_def;
	
	-- Maximum error value (mz(t)) calculation
	p_merr_calc : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				data_merr_s <= (others => '0');
			else
				if (enable_i = '1') then
					if (FIDEL_CTRL_TYPE_G = LOSSLESS_C) then	-- Lossless method
						data_merr_s <= (others => '0');
					elsif (FIDEL_CTRL_TYPE_G = ABS_ERR_C) then	-- ONLY absolute error limit method
						data_merr_s <= std_logic_vector(to_signed(Az_s, D_C+1));
					elsif (FIDEL_CTRL_TYPE_G = REL_ERR_C) then	-- ONLY relative error limit method
						data_merr_s <= std_logic_vector(resize(shift_sgn(resize(to_signed(Rz_s, D_C*2) * abs(resize(signed(data_s3_i), D_C*2)), D_C*2), -D_C), D_C+1));
					else	-- FIDEL_CTRL_TYPE_G = ABS_REL_C	-- BOTH absolute and relative error limits
						data_merr_s <= std_logic_vector(resize(work.utils_image.min(to_signed(Az_s, D_C+1), resize(shift_sgn(resize(to_signed(Rz_s, D_C*2) * abs(resize(signed(data_s3_i), D_C*2)), D_C*2), -D_C), D_C+1)), D_C+1));
					end if;
				end if;
			end if;
		end if;
	end process p_merr_calc;

	-- Outputs
	enable_o	<= enable_s;
	img_coord_o	<= img_coord_s;
	data_merr_o	<= data_merr_s;
end architecture behavioural;