--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		30/10/2020
--------------------------------------------------------------------------------
-- IP name:		quantizer
--
-- Description: Quantizies the incoming input with a uniform step size
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;
use work.comp_predictor.all;
	
entity quantizer is
	generic (
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic;
		-- 1: enabled, 0: disabled
		PER_ERR_LIM_UPD_G	: std_logic
	);
	port (
		clock_i		 : in  std_logic;
		reset_i		 : in  std_logic;
		
		enable_i	 : in  std_logic;
		enable_o	 : out std_logic;
		img_coord_i	 : in  img_coord_t;
		img_coord_o	 : out img_coord_t;
		
		data_s3_i	 : in  std_logic_vector(D_C+1-1 downto 0);	-- "s^z(t)" (predicted sample)
		data_res_i	 : in  std_logic_vector(D_C+2-1 downto 0);	-- "/\z(t)" (prediction residual)
		
		data_merr_o	 : out std_logic_vector(D_C+1-1 downto 0);	-- "mz(t)"  (maximum error)
		data_quant_o : out std_logic_vector(D_C+2-1 downto 0)	-- "qz(t)"  (quantizer index)
	);
end entity quantizer;

architecture behavioural of quantizer is
	signal enable1_s	: std_logic   := '0';
	signal enable2_s	: std_logic   := '0';
	signal img_coord1_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	signal img_coord2_s	: img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal data_merr1_s	: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal data_merr2_s	: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal data_res_s	: std_logic_vector(D_C+2-1 downto 0) := (others => '0');
	signal data_quant_s	: std_logic_vector(D_C+2-1 downto 0) := (others => '0');

begin
	-- Input values delayed one clock cycle to synchronize them with the next modules in chain
	p_quant_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				enable2_s	 <= '0';
				img_coord2_s <= reset_img_coord(SMPL_ORDER_G);
				data_merr2_s <= (others => '0');
				data_res_s	 <= (others => '0');
			else
				enable2_s	 <= enable1_s;
				img_coord2_s <= img_coord1_s;
				data_merr2_s <= data_merr1_s;
				data_res_s	 <= data_res_i;
			end if;
		end if;
	end process p_quant_delay;

	-- Maximum error (mz(t)) calculation
	i_fidel_ctrl : fidelity_ctrl
	generic map(
		SMPL_ORDER_G		=> SMPL_ORDER_G,
		FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G,
		PER_ERR_LIM_UPD_G	=> PER_ERR_LIM_UPD_G
	)
	port map(
		clock_i		=> clock_i,
		reset_i		=> reset_i,
		
		enable_i	=> enable_i,
		enable_o	=> enable1_s,
		img_coord_i	=> img_coord_i,
		img_coord_o	=> img_coord1_s,
		
		data_s3_i	=> data_s3_i,
		data_merr_o	=> data_merr1_s
	);

	-- Quantizer index (qz(t)) calculation
	p_quant_calc : process(clock_i) is
		variable comp1_v : integer range -1 to 1			  := 0;
		variable comp2_v : std_logic_vector(D_C+2-1 downto 0) := (others => '0');
		variable comp3_v : std_logic_vector(D_C+4-1 downto 0) := (others => '0');
		variable comp4_v : std_logic_vector(D_C+2-1 downto 0) := (others => '0');
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				comp1_v		 := 0;
				comp2_v		 := (others => '0');
				comp3_v		 := (others => '0');
				comp4_v		 := (others => '0');
				data_quant_s <= (others => '0');
			else
				if (enable1_s = '1') then
					if (img_coord1_s.t = 0) then
						data_quant_s <= data_res_s;
					else
						comp1_v		 := sgn(signed(data_res_s));
						comp2_v		 := std_logic_vector(resize(abs(resize(signed(data_res_s), D_C+2)) + resize(signed(data_merr1_s), D_C+2), D_C+2));
						comp3_v		 := std_logic_vector(resize(shift_sgn(resize(signed(data_merr1_s), D_C+4), 1) + to_signed(1, D_C+4), D_C+4));
						comp4_v		 := std_logic_vector(resize(division_down(resize(signed(comp2_v), D_C+4), signed(comp3_v)), D_C+2));
						data_quant_s <= std_logic_vector(resize(to_signed(comp1_v, D_C+2) * signed(comp4_v), D_C+2));
					end if;
				end if;
			end if;
		end if;
	end process p_quant_calc;
	
	-- Outputs
	enable_o	 <= enable2_s;
	img_coord_o	 <= img_coord2_s;
	data_merr_o	 <= data_merr2_s;
	data_quant_o <= data_quant_s;
end architecture behavioural;