--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		07/11/2020
--------------------------------------------------------------------------------
-- IP name:		top_predictor
--
-- Description: Top entity for the "predictor" module
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;
use work.comp_predictor.all;
	
entity top_predictor is
	generic (
		-- 0: "signed" type samples, 1: "unsigned" type samples
		SMPL_TYPE_G			: std_logic;
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
		LSUM_TYPE_G			: std_logic_vector(1 downto 0);
		-- 1: Full prediction mode, 0: Reduced prediction mode
		PREDICT_MODE_G		: std_logic;
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic;
		-- 1: enabled, 0: disabled
		PER_ERR_LIM_UPD_G	: std_logic;
		-- 1: Custom weight init, 0: Default weight init
		W_INIT_TYPE_G		: std_logic
	);
	port (
		clock_i			: in  std_logic;
		reset_i			: in  std_logic;

		enable_i		: in  std_logic;
		enable_o		: out std_logic;
		img_coord_i		: in  img_coord_t;
		img_coord_o		: out img_coord_t;
		
		data_s0_i		: in  std_logic_vector(D_C-1 downto 0);	 -- "sz(t)" (original sample)
		data_mp_quan_o	: out std_logic_vector(D_C-1 downto 0)	 -- "?z(t)" (mapped quantizer index)
	);
end entity top_predictor;

architecture behavioural of top_predictor is
	------------------------------------------------------------------------------------------------------
	-- SIGNAL NAME REFERENCES:
	------------------------------------------------------------------------------------------------------
	--	s0: "sz(t)"		(original sample)
	--	s1: "s'z(t)"	(clipped quantizer bin center)
	--	s2: "s''z(t)"	(sample representative)
	--	s3: "s^z(t)"	(predicted sample)
	--	s4: "s~z(t)"	(double-resolution predicted sample)
	--	s5: "s~''z(t)"	(double-resolution sample representative)
	--	s6: "s)z(t)"	(high-resolution predicted sample)
	------------------------------------------------------------------------------------------------------

	------------------------------------------------------------------------------------------------------
	-- DESIGN CONSTANTS
	------------------------------------------------------------------------------------------------------
	-- Lower, Mid-range and Upper sample values (for either "signed" or "unsigned" types)
	constant SMPL_LIMIT_C	: smpl_lim_t := set_smpl_limits(SMPL_TYPE_G);
	-- Maximum delay to define arrays size (which COULD be different from the top IP processing-time)
	constant MAX_DELAY_C	: integer := TIME_ADDER_C + TIME_QUANTIZER_C + TIME_SMPL_REPRE_C;

	-- Array of previous spectral bands (P*z) and local differencial values (Cz) for prediction in band Z
	signal pz_arr_s			: array_integer_t(0 to MAX_DELAY_C-1) := (others => 0);
	signal cz_arr_s			: array_integer_t(0 to MAX_DELAY_C-1) := (others => 3);

	-- Enable and image coordinates to interconnect all sub-blocks
	signal enable1_s,	 enable2_s,	   enable3_s,	 enable4_s,	   enable5_s	: std_logic	:= '0';
	signal img_coord1_s, img_coord2_s, img_coord3_s, img_coord4_s, img_coord5_s : img_coord_t := reset_img_coord(SMPL_ORDER_G);
	
	signal data_s0_arr_s	: array_slv_t(0 to MAX_DELAY_C-1)(D_C-1 downto 0)	:= (others => (others => '0'));
	signal data_s3_arr1_s	: array_slv_t(0 to MAX_DELAY_C-1)(D_C+1-1 downto 0)	:= (others => (others => '0'));
	signal data_s3_arr2_s	: array_slv_t(0 to MAX_DELAY_C-1)(D_C+1-1 downto 0)	:= (others => (others => '0'));
    signal data_s6_arr_s	: array_slv_t(0 to MAX_DELAY_C-1)(Re_C-1 downto 0)	:= (others => (others => '0'));
	signal data_s4_lsb_arr_s: std_logic_vector(MAX_DELAY_C-1 downto 0)			:= (others => '0');
	
	-- For whatever reason, these arrays cannot be used to save the OUTPUT value from an IP (e.g. data_quant_o => data_quant_arr_s(0) FAIL),
	-- so independent signals are used instead, and later assigned to these arrays (e.g. data_quant_o => data_quant_s PASS).

	signal en_res_s			: std_logic := '0';
	signal data_res_s		: std_logic_vector(D_C+2-1 downto 0) := (others => '0');
	signal data_merr_s		: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal data_quant_s		: std_logic_vector(D_C+2-1 downto 0) := (others => '0');
	signal data_mp_quan_s	: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	
	signal data_s4_lsb_1s	: std_logic							 := '0';
	signal data_s4_lsb_2s	: std_logic							 := '0';
	signal data_s1_s		: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_s2_s		: std_logic_vector(D_C-1 downto 0)	 := (others => '0');
	signal data_s3_1s		: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal data_s3_2s		: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal data_s3_3s		: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal data_s3_4s		: std_logic_vector(D_C+1-1 downto 0) := (others => '0');
	signal data_s6_s		: std_logic_vector(Re_C-1 downto 0)	 := (others => '0');
	
begin
	-- Input values delayed MAX_DELAY_C clock cycles to synchronize them with the next modules in chain
	-- NOTE: Not all arrays need to use all positions, but in this way everything is within one single process
	p_predictor_delay : process(clock_i) is
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				data_s0_arr_s		 <= (others => (others => '0'));
				data_s3_arr1_s		 <= (others => (others => '0'));
				data_s3_arr2_s		 <= (others => (others => '0'));
                data_s6_arr_s		 <= (others => (others => '0'));
				data_s4_lsb_arr_s	 <= (others => '0');
			else
				data_s0_arr_s(0)	 <= data_s0_i;
				data_s3_arr1_s(0)	 <= data_s3_1s;
				data_s3_arr2_s(0)	 <= data_s3_2s;
                data_s6_arr_s(0)	 <= data_s6_s;
				data_s4_lsb_arr_s(0) <= data_s4_lsb_2s;
				
				for i in 1 to (MAX_DELAY_C-1) loop
					data_s0_arr_s(i)	 <= data_s0_arr_s(i-1);
					data_s3_arr1_s(i)	 <= data_s3_arr1_s(i-1);
					data_s3_arr2_s(i)	 <= data_s3_arr2_s(i-1);
                    data_s6_arr_s(i)	 <= data_s6_arr_s(i-1);
					data_s4_lsb_arr_s(i) <= data_s4_lsb_arr_s(i-1);
				end loop;
			end if;
		end if;
	end process p_predictor_delay;
	
	-- Calculation of Pz (Previous spectral bands for prediction in band z) and Cz (Local diff values for prediction in band z)
	-- NOTE: This process is clocked to ensure that the first position of the arrays are delayed one clock cycle, as the rest of arrays in this entity
	p_prev_spec_bands : process(clock_i) is
		variable pz_v : integer := 0;
	begin
		if rising_edge(clock_i) then
			if (reset_i = '1') then
				pz_v	 	:= 0;
				pz_arr_s	<= (others => 0);
				cz_arr_s	<= (others => 3);
			else
				pz_v		:= work.utils_image.min(img_coord_i.z, P_C);
				pz_arr_s(0) <= pz_v;
				if (PREDICT_MODE_G = '1') then
					cz_arr_s(0) <= pz_v + 3;	-- Cz for "Full Prediction Mode"
				else
					cz_arr_s(0) <= pz_v;		-- Cz for "Reduced Prediction Mode"
				end if;
				
				for i in 1 to (MAX_DELAY_C-1) loop
					pz_arr_s(i) <= pz_arr_s(i-1);
					cz_arr_s(i) <= cz_arr_s(i-1);
				end loop;
			end if;
		end if;
	end process p_prev_spec_bands;
	
	i_adder : adder
	generic map(
		SMPL_TYPE_G	 	=> SMPL_TYPE_G,
		SMPL_ORDER_G 	=> SMPL_ORDER_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i,
		
		img_en_i		=> enable_i,
		enable_o		=> enable1_s,
		img_coord_i		=> img_coord_i,		
		img_coord_o		=> img_coord1_s,
		
		pred_en_i		=> enable4_s,
		pred_coord_i	=> img_coord4_s,
		data_s3_i		=> data_s3_1s,
		data_s3_o		=> data_s3_2s,
		data_s4_lsb_i	=> data_s4_lsb_1s,
		data_s4_lsb_o	=> data_s4_lsb_2s,
		
		data_s0_i		=> data_s0_i,
		en_res_o		=> en_res_s,
		data_res_o		=> data_res_s
	);
	
	i_quantizer : quantizer
	generic map(
		SMPL_ORDER_G		=> SMPL_ORDER_G,
		FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G,
		PER_ERR_LIM_UPD_G	=> PER_ERR_LIM_UPD_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i		 	=> reset_i,
		
		enable_i	 	=> enable1_s,
		enable_o	 	=> enable2_s,
		img_coord_i	 	=> img_coord1_s,
		img_coord_o	 	=> img_coord2_s,
		
		data_s3_i	 	=> data_s3_3s,
		data_res_i	 	=> data_res_s,

		data_merr_o	 	=> data_merr_s,
		data_quant_o 	=> data_quant_s
	);
	data_s3_3s <= data_s3_arr1_s(0 + TIME_ADDER_C - 1) when (en_res_s = '0') else data_s3_2s;
	
	i_sample_repr : sample_representative
	generic map(
		SMPL_ORDER_G	  => SMPL_ORDER_G,
		SMPL_LIMIT_G	  => SMPL_LIMIT_C,
		FIDEL_CTRL_TYPE_G => FIDEL_CTRL_TYPE_G
	)
	port map(
		clock_i		 	=> clock_i,
		reset_i		 	=> reset_i or en_res_s,
		
		enable_i	 	=> enable2_s,
		enable_o	 	=> enable3_s,
		img_coord_i	 	=> img_coord2_s,
		img_coord_o	 	=> img_coord3_s,
		
		data_merr_i	 	=> data_merr_s,
		data_quant_i 	=> data_quant_s,
		data_s0_i	 	=> data_s0_arr_s(0 + TIME_ADDER_C + TIME_QUANTIZER_C - 1),
		data_s3_i	 	=> data_s3_arr1_s(0 + TIME_ADDER_C + TIME_QUANTIZER_C - 1),
		data_s6_i		=> data_s6_arr_s(0 + TIME_ADDER_C + TIME_QUANTIZER_C - 1),

		data_s1_o	 	=> data_s1_s,
		data_s2_o	 	=> data_s2_s		
	);
	
	i_prediction : prediction
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		SMPL_LIMIT_G	=> SMPL_LIMIT_C,
		LSUM_TYPE_G		=> LSUM_TYPE_G,
		PREDICT_MODE_G	=> PREDICT_MODE_G,
		W_INIT_TYPE_G	=> W_INIT_TYPE_G
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i or en_res_s,
		
		enable_i		=> enable3_s,
		enable_o		=> enable4_s,
		img_coord_i		=> img_coord3_s,
		img_coord_o		=> img_coord4_s,
		
		pz_i			=> pz_arr_s(0 + TIME_ADDER_C + TIME_QUANTIZER_C + TIME_SMPL_REPRE_C - 1),
		data_s0_i		=> data_s0_arr_s(0 + TIME_ADDER_C + TIME_QUANTIZER_C + TIME_SMPL_REPRE_C - 1),
		data_s1_i		=> data_s1_s,
		data_s2_i		=> data_s2_s,
		
		data_s3_o		=> data_s3_1s,
		data_s4_lsb_o	=> data_s4_lsb_1s,
		data_s6_o		=> data_s6_s
	);
	
	i_mapper : mapper
	generic map(
		SMPL_ORDER_G	=> SMPL_ORDER_G,
		SMPL_LIMIT_G	=> SMPL_LIMIT_C
	)
	port map(
		clock_i			=> clock_i,
		reset_i			=> reset_i or (not en_res_s),
		
		enable_i		=> enable2_s,
		enable_o		=> enable5_s,
		img_coord_i		=> img_coord2_s,
		img_coord_o		=> img_coord5_s,
		
		data_s3_i		=> data_s3_4s,
		data_s4_lsb_i	=> data_s4_lsb_arr_s(0 + TIME_QUANTIZER_C - 1),
		data_merr_i	 	=> data_merr_s,
		data_quant_i	=> data_quant_s,
		data_mp_quan_o	=> data_mp_quan_s
	);
	data_s3_4s <= data_s3_arr1_s(0 + TIME_ADDER_C + TIME_QUANTIZER_C - 1) when (en_res_s = '0') else data_s3_arr2_s(0 + TIME_QUANTIZER_C - 1);

	-- Outputs
	enable_o			<= enable5_s;
	img_coord_o			<= img_coord5_s;
	data_mp_quan_o		<= data_mp_quan_s;
end architecture behavioural;