# CCSDS123 Issue2 - VHDL 2008 Implementation
The following repository introduces an implementation of the ***CCSDS123 (Issue 2) Recommended Standard*** for ***FPGAs***, done with ***VHDL 2008***.  

This standard defines an algorithm designed to de/compress ***hyperspectral*** and ***multispectral images*** in 3D, receiving ***one single sample per clock cycle***.  

The ***Issue 2*** is an improvement of ***Issue 1***, allowing not only to do a ***lossless*** de/compression (***Issue 1***), but also a ***near-lossless*** de/compression (***Issue 2***), if configured so.

### Design notes

The present VHDL code has been designed to offer a high level of ***modularity*** and ***re-usability***, and this is why there is (almost) an independent IP per single operation to execute.  

For the sake of readability, all ***parameters***, ***types***, ***components*** and ***functions*** are declared as independent packages files, placed within folder ***Packages*** (a different one per block).  

All IPs are validated by simulation thanks to the ***VUnit framework***. Each block (and sub-block, when so) has a folder ***Simulation*** with its ***VUnit testbench + Python file*** that executes it (using ***ModelSim*** as simulator engine). Moreover, there is a ***bash file*** to automate the VUnit testbenches execution, which asks which specific block (or simply everything) to run.

There is also a ***TCL framework*** (controlled by another ***bash file***) that automates the creation of a ***Vivado (2019.1) project*** as well as the ***bitstream (.bit)*** and ***hardware description file (.hdf)*** files generation.

## Top entity

The present design is composed of 3 different parts:
1. ***Image block***: Manages the image coordinates (samples).
2. ***Predictor block***: Predicts the new samples, based on the nearby ones.
3. ***Encoder block***: Encodes the whole image as a packet, composed of a header and a body.

## 1. Image block

This block, once receives an ***enable*** signal from outside (meaning a new image is coming in), starts counting the ***image coordinates***.

## 2. Predictor block

This block is a ***closed loop*** that allows for either ***lossless*** or ***near-lossless*** de/compression. It has 5 major sub-blocks:
- Adder IP
- Quantizer IP
- Mapper IP
- Sample Representative IP
- Prediction IP

## 3. Prediction-Encoder Interconnection IP

Only instantiated if configured so (Periodic Error Limit Updating process enabled), this IP is placed between the ***Predictor*** and ***Encoder*** IPs in order to forward either the ***Predictor output*** or the ***Error Limit Values***, depending on the ***image coordinates***.

## 4. Encoder block

This block takes all ***samples*** from the ***Predictor block*** to send them out as ***packets*** of a configurable width. It has 3 major sub-blocks: a ***header*** with all ***metadata***, a ***body*** that de/compresses the ***samples*** (performed with one of the three possible ***entropy coders***) and a ***packer*** that outputs all de/compressesed ***samples*** inside ***packets***:
- Header IP
- Body IP
    - Sample-Adaptive Entropy Coder IP
    - Hybrid Entropy Coder IP
    - Block-Adaptive Entropy Coder IP
- Packer IP

#### NOTE 1

Refer to slides ***Block_diagrams.pptx*** to see a small block diagram of each sub-block IP.

#### NOTE 2

Refer to the following link for more information about the ***VUnit framework***: https://vunit.github.io/

## How to use

There are 2 bash files: one to manage the validation by simulation, and another one to automate the bitstream generation.

## 1. VUnit simulation bash file

This bash file, called ***run_simulations.sh***, is in charge of the complete VUnit testbench architecture. It accesses the selected (group of) VUnit testbench, executes it and provides the results.

Instructions for execution:
1. Install "Python3" (and include it to the PATH as well).
2. Install "pip" (python packages installer).
3. Install "VUnit" package: --> "pip install vunit_hdl".
4. Open a terminal and execute the bash file --> "sh run_simulations.sh".
5. Follow instructions in the GUI.

Nevertheless, a specific VUnit testbench can also be executed in an isolated way. Simply go to the desired ***simulation*** folder (where "run_XXX.py" and "tb_XXX.vhd" files are located) and execute:

python run_XXX.py

This will execute the ModelSim Simulator Enginer in background to perform validation, but if it is required to open the waveform, the previous command requires an additional attribute:

python run_XXX.py -g

## 2. Xilinx Vivado bash file

This bash file, called ***run_vivado_project.sh***, uses the ***TCL*** framework within ***_vivado_framework*** to call the ***Xilinx Vivado IDE***, create a project in ***VHDL-2008*** with hardware ***ZCU102***, adds all necessary source and constraint files, and performs ***synthesis*** and ***implementation***, to finally generate the ***bitstream***.

Instructions for execution:
1. Install "Xilinx Vivado 2019" (and include it to the PATH as well).
2. Open a terminal and execute the bash file --> "sh run_vivado_project.sh".
3. Follow instructions in the GUI.

In case new source/constraint files must be added, or the configuration of the ***Xilinx Vivado*** project must be modified, access the ***TCL*** framework from ***_vivado_framework*** and modify the ***TCL*** files in there.