--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		15/10/2021
--------------------------------------------------------------------------------
-- IP name:		tb_pred_enc_interconn
--
-- Description: Testbench for the "Predictor-Encoder Interconnection" module
--
--
--------------------------------------------------------------------------------

-- Library instantiated, then declaration of files/functions to use
library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

-- A reference to VUnit (and related) set of libraries and packages
library vunit_lib;
context vunit_lib.vunit_context;
context vunit_lib.com_context;
context vunit_lib.data_types_context;

-- Packages related to the Image IP
use vunit_lib.gen_labels.all;
use vunit_lib.param_image.all;
use vunit_lib.types_image.all;
use vunit_lib.utils_image.all;
use vunit_lib.comp_image.all;

-- Packages related to the Predictor IP
use vunit_lib.param_predictor.all;
use vunit_lib.types_predictor.all;
use vunit_lib.error_limit_values_table.all;

-- Components related to the Encoder IP
use vunit_lib.comp_encoder.all;

entity tb_pred_enc_interconn is
    generic (
        encoded_tb_cfg	: string;
        runner_cfg		: string
    );
end entity tb_pred_enc_interconn;

architecture behavioural of tb_pred_enc_interconn is
	-- Record type to pack all signals (from Python script) together
	type tb_cfg_t is record
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic;
	end record tb_cfg_t;
	
	-- Function to decode the Python signals and connect them into the VHDL testbench
	impure function decode(encoded_tb_cfg : string) return tb_cfg_t is
	begin
	return (
		SMPL_ORDER_G		=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "SMPL_ORDER_PY")), 2)),
		FIDEL_CTRL_TYPE_G	=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "FIDEL_CTRL_TYPE_PY")), 2)),
		ABS_ERR_BAND_TYPE_G	=> std_logic'value(get(encoded_tb_cfg, "ABS_ERR_BAND_TYPE_PY")),
		REL_ERR_BAND_TYPE_G	=> std_logic'value(get(encoded_tb_cfg, "REL_ERR_BAND_TYPE_PY"))
	);
	end function decode;
	
	-- Calculates the additional clock cycles because of the error limit values forwarding
	function num_err_lim_val_smpls(smpl_order_in : in std_logic_vector; fidel_type_in : in std_logic_vector; abs_band_in : in std_logic; rel_band_in : in std_logic) return integer is
		variable num_smpls_v : integer := 0;
	begin
		if (smpl_order_in /= BSQ_C) then
			case fidel_type_in is
				when ABS_ERR_C =>
					if (abs_band_in = '0') then
						num_smpls_v := ERR_LIM_VAL_ARR_C'length;
					else			-- abs_band_in = '1'
						num_smpls_v := ERR_LIM_VAL_ARR_C'length * ERR_LIM_VAL_ARR_C(0).abs_arr'length;
					end if;
					
				when REL_ERR_C =>
					if (rel_band_in = '0') then
						num_smpls_v := ERR_LIM_VAL_ARR_C'length;
					else			-- rel_band_in = '1'
						num_smpls_v := ERR_LIM_VAL_ARR_C'length * ERR_LIM_VAL_ARR_C(0).rel_arr'length;
					end if;

				when ABS_REL_C =>
					if (abs_band_in = '0' and rel_band_in = '0') then
						num_smpls_v := ERR_LIM_VAL_ARR_C'length * 2;
					elsif (abs_band_in = '1' and rel_band_in = '1') then
						num_smpls_v := ERR_LIM_VAL_ARR_C'length * ERR_LIM_VAL_ARR_C(0).abs_arr'length * 2;
					else
						num_smpls_v := ERR_LIM_VAL_ARR_C'length * (ERR_LIM_VAL_ARR_C(0).abs_arr'length + 1);
					end if;
				
				when others =>		-- LOSSLESS_C
					num_smpls_v := 0;
			end case;
		else
			num_smpls_v := 0;
		end if;
		
		-- The present value needs +1 to be inline with the external counter, meant to be compared
		num_smpls_v	:= num_smpls_v + 1;
		
		return num_smpls_v;
	end function num_err_lim_val_smpls;
	
	-- Constant of type "tb_cfg_t" and initialized by "decode()" function
	constant tb_cfg			  : tb_cfg_t	:= decode(encoded_tb_cfg);
	
	constant PERIOD_100M_C	  : time		:= 10 ns;
	
	signal clock_s			  : std_logic	:= '0';
	signal reset_s			  : std_logic	:= '1';
	
	signal enable_in_s		  : std_logic	:= '0';
	signal enable_out_s		  : std_logic	:= '0';
	signal img_coord_in_s	  : img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal img_coord_out_s	  : img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	
	signal clock_cyles_s	  : integer		:= 0;
	signal image_end_s		  : std_logic	:= '0';
	signal init_fifo_s		  : std_logic	:= '0';
	
    signal err_lim_upd_en_s	  : std_logic	:= '0';
	signal err_lim_val_s	  : std_logic_vector(D_C-1 downto 0) := (others => '0');
	signal data_mp_quan_in_s  : std_logic_vector(D_C-1 downto 0) := (others => '0');
	signal data_mp_quan_out_s : std_logic_vector(D_C-1 downto 0) := (others => '0');

	signal flag_start, flag_stop : std_logic := '0';

begin
	clock_s <= not clock_s after PERIOD_100M_C/2;
	reset_s <= '0' after 20 ns;
	
	-- Simulation MAIN process (if this one finishes, the whole simulation too)
	test_runner : process is
	begin
		-- Process starts by setting up VUnit using test_runner_setup procedure
		test_runner_setup(runner, runner_cfg);

		while test_suite loop		-- Pay very attention to the testcase name, arrows or accents prevent the system to work...
			if run("Predictor-Encoder Interconnection Entity") then
				info("Running test case = " & to_string(running_test_case));
				
				-- Test cases with 'SMPL_ORDER_G = BSQ_C or FIDEL_CTRL_TYPE_G = LOSSLESS_C' are skipped because they do not use the FIFO, the core of the current IP
				if (tb_cfg.SMPL_ORDER_G = BSQ_C or tb_cfg.FIDEL_CTRL_TYPE_G /= LOSSLESS_C) then
					exit;
				end if;
				
				flag_start <= '1';
			end if;
			
			wait until (flag_stop = '1');
		end loop;

		-- The process ends with the test_runner_cleanup procedure which will force the simulation to stop
		test_runner_cleanup(runner);
	end process;

	-- Set Watchdog with limited time (this number must be higher as any "wait for" statement used in the testbench)
	test_runner_watchdog(runner, 1 ms);

	-- Clock cycles counter
	p_clk_cyc_cnt : process(clock_s) is
	begin
		if (rising_edge(clock_s)) then
			if (reset_s = '1' or image_end_s = '1') then
				clock_cyles_s <= 0;
			else
				if (flag_start = '1' and init_fifo_s = '1') then
					clock_cyles_s <= clock_cyles_s + 1;
				end if;
			end if;
		end if;
	end process p_clk_cyc_cnt;
	
	-- Input signals stimulus calculation
	p_mp_quan_update : process(clock_s) is
	begin
		if (rising_edge(clock_s)) then
			if (reset_s = '1' or image_end_s = '1') then
				data_mp_quan_in_s <= (others => '0');
			else
				if (flag_start = '1' and init_fifo_s = '1') then
					if ((img_coord_out_s.x < NX_C-1) or (img_coord_out_s.y < NY_C-1) or (img_coord_out_s.z < NZ_C-1)) then						
						if (data_mp_quan_in_s = (data_mp_quan_in_s'length-1 downto 0 => '1')) then
							data_mp_quan_in_s <= (others => '0');		
						else
							data_mp_quan_in_s <= std_logic_vector(resize(signed(data_mp_quan_in_s) + to_signed(3, data_mp_quan_in_s'length), data_mp_quan_in_s'length));
						end if;
					end if;
				end if;
			end if;	
		end if;
	end process p_mp_quan_update;
	
	-- Manages the signal "image_end_s" to restart the system, when the end of the image was reached
	p_img_end : process(reset_s, img_coord_out_s) is
	begin
		if (reset_s = '1') then
			image_end_s <= '0';
		else
			if ((img_coord_out_s.x >= NX_C-1) and (img_coord_out_s.y >= NY_C-1) and (img_coord_out_s.z >= NZ_C-1)) then
				image_end_s <= '1';
			else
				image_end_s <= '0';
			end if;
		end if;
	end process p_img_end;
	
	-- Process to stop the simulation when finished (to use 'wait' statements, the process cannot be clocked)
	p_stop_sim : process is
	begin
		-- After one complete image, wait statement and request to finish the simulation
		wait until (image_end_s <= '1');
		-- Check clock cycles number is the expected one according to error limit values update
		-- check_equal(clock_cyles_s, NX_C*NY_C*NZ_C + num_err_lim_val_smpls(tb_cfg.SMPL_ORDER_G, tb_cfg.FIDEL_CTRL_TYPE_G, tb_cfg.ABS_ERR_BAND_TYPE_G, tb_cfg.REL_ERR_BAND_TYPE_G), "Image coordinates are NOT properly synchronized!");
		wait for 100 ns;
		
		-- Request to finish the simulation
		flag_stop <= '1';
		
		-- Two possibilities to stop the simulation without VUnit framework (only the "assert" line really works, anyway)
		-- assert FALSE Report "Simulation Finished" severity FAILURE;
		-- std.env.finish;
	end process p_stop_sim;
	
	-- Entity to control the image coordinates
	i_image_coord_control : image_coord_control
	generic map(
		SMPL_ORDER_G => tb_cfg.SMPL_ORDER_G
	)
	port map(
		clock_i		 => clock_s,
		reset_i		 => reset_s,

		enable_i	 => flag_start and init_fifo_s,
		enable_o	 => enable_in_s,
		
		image_end_i	 => image_end_s,
		img_coord_o	 => img_coord_in_s
	);
	
	-- Predictor-Encoder Interconnection entity (only applicable when 'SMPL_ORDER_G /= BSQ_C AND FIDEL_CTRL_TYPE_G /= LOSSLESS_C')
	i_pred_enc_interconn : pred_enc_interconn
	generic map(
		SMPL_ORDER_G		=> tb_cfg.SMPL_ORDER_G,
		FIDEL_CTRL_TYPE_G	=> tb_cfg.FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G	=> tb_cfg.ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> tb_cfg.REL_ERR_BAND_TYPE_G
	)
	port map(
		clock_i			 => clock_s,
		reset_i			 => reset_s or image_end_s,
		
		enable_i		 => enable_in_s,
		enable_o		 => enable_out_s,
		img_coord_i		 => img_coord_in_s,
		img_coord_o		 => img_coord_out_s,
		
		init_fifo_o		 => init_fifo_s,
		
		err_lim_upd_en_o => err_lim_upd_en_s,
		err_lim_val_o	 => err_lim_val_s,
		data_mp_quan_i   => data_mp_quan_in_s,
        data_mp_quan_o   => data_mp_quan_out_s
	);
	
end architecture behavioural;