--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		07/08/2021
--------------------------------------------------------------------------------
-- IP name:		tb_top_ccsds123_issue2
--
-- Description: Testbench for the "CCSDS123 Issue2 Top" module
--
--
--------------------------------------------------------------------------------

-- Library instantiated, then declaration of files/functions to use
library ieee;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;

-- A reference to VUnit (and related) set of libraries and packages
library vunit_lib;
context vunit_lib.vunit_context;
context vunit_lib.com_context;
context vunit_lib.data_types_context;

-- Components and packages related to the Image IP
use vunit_lib.param_image.all;
use vunit_lib.types_image.all;
use vunit_lib.example_vectors.all;
use vunit_lib.utils_image.all;
use vunit_lib.comp_image.all;

-- Components and packages related to the Predictor IP
use vunit_lib.param_predictor.all;
use vunit_lib.types_predictor.all;
use vunit_lib.utils_predictor.all;
use vunit_lib.comp_predictor.all;

-- Components and packages related to the Encoder IP
use vunit_lib.param_encoder.all;
use vunit_lib.types_encoder.all;
use vunit_lib.utils_encoder.all;
use vunit_lib.comp_encoder.all;

entity tb_top_ccsds123_issue2 is
    generic (
        encoded_tb_cfg	: string;
        runner_cfg		: string
    );
end entity tb_top_ccsds123_issue2;

architecture behavioural of tb_top_ccsds123_issue2 is
	-- Record type to pack all signals (from Python script) together
	type tb_cfg_t is record
		SMPL_TYPE_G			: std_logic;
		SMPL_ORDER_G		: std_logic_vector(1 downto 0);
		HEADER_EN_G			: std_logic;
		PREDICT_MODE_G		: std_logic;
		LSUM_TYPE_G			: std_logic_vector(1 downto 0);
		W_INIT_TYPE_G		: std_logic;
		PER_ERR_LIM_UPD_G	: std_logic;
		FIDEL_CTRL_TYPE_G	: std_logic_vector(1 downto 0);
		ABS_ERR_BAND_TYPE_G	: std_logic;
		REL_ERR_BAND_TYPE_G	: std_logic;
		ENCODER_TYPE_G		: std_logic_vector(1 downto 0);
		SUPL_TABLE_TYPE_G	: std_logic_vector(1 downto 0);
		SUPL_TABLE_STRUCT_G	: std_logic_vector(1 downto 0);
		ENDIANNESS_G		: std_logic;
	end record tb_cfg_t;
	
	-- Function to decode the Python signals and connect them into the VHDL testbench
	impure function decode(encoded_tb_cfg : string) return tb_cfg_t is
	begin
	return (
		SMPL_TYPE_G			=> std_logic'value(get(encoded_tb_cfg, "SMPL_TYPE_PY")),
		SMPL_ORDER_G		=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "SMPL_ORDER_PY")), 2)),
		HEADER_EN_G			=> std_logic'value(get(encoded_tb_cfg, "HEADER_EN_PY")),
		PREDICT_MODE_G		=> std_logic'value(get(encoded_tb_cfg, "PREDICT_MODE_PY")),
		LSUM_TYPE_G			=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "LSUM_TYPE_PY")), 2)),
		W_INIT_TYPE_G		=> std_logic'value(get(encoded_tb_cfg, "W_INIT_TYPE_PY")),
		PER_ERR_LIM_UPD_G	=> std_logic'value(get(encoded_tb_cfg, "PER_ERR_LIM_UPD_PY")),
		FIDEL_CTRL_TYPE_G	=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "FIDEL_CTRL_TYPE_PY")), 2)),
		ABS_ERR_BAND_TYPE_G	=> std_logic'value(get(encoded_tb_cfg, "ABS_ERR_BAND_TYPE_PY")),
		REL_ERR_BAND_TYPE_G	=> std_logic'value(get(encoded_tb_cfg, "REL_ERR_BAND_TYPE_PY")),
		ENCODER_TYPE_G		=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "ENCODER_TYPE_PY")), 2)),
		SUPL_TABLE_TYPE_G	=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "SUPL_TABLE_TYPE_PY")), 2)),
		SUPL_TABLE_STRUCT_G	=> std_logic_vector(to_unsigned(integer'value(get(encoded_tb_cfg, "SUPL_TABLE_STRUCT_PY")), 2)),
		ENDIANNESS_G		=> std_logic'value(get(encoded_tb_cfg, "ENDIANNESS_PY"))
	);
	end function decode;
	
	-- Constant of type "tb_cfg_t" and initialized by "decode()" function
	constant tb_cfg			: tb_cfg_t := decode(encoded_tb_cfg);

	-- Supplementary tables information
	constant UDEF_DATA_C		  : std_logic_vector(7 downto 0) := x"AC";
	constant SUPL_TABLE_TYPE_C	  : supl_table_type_t	 := (others => tb_cfg.SUPL_TABLE_TYPE_G);
	constant SUPL_TABLE_PURPOSE_C :	supl_table_purpose_t := (others => "0000");		-- Supl. table "Scale" type
	constant SUPL_TABLE_STRUCT_C  : supl_table_struct_t	 := (others => tb_cfg.SUPL_TABLE_STRUCT_G);
	constant SUPL_TABLE_UDATA_C	  : supl_table_udata_t	 := (others => (others => '1'));
	
	constant PERIOD_100M_C	: time := 10 ns;
	
	signal clock_s			: std_logic := '0';
	signal reset_s			: std_logic	:= '1';

	signal img_coord_s		: img_coord_t := reset_img_coord(tb_cfg.SMPL_ORDER_G);
	signal init_fifo_s		: std_logic	:= '0';
	signal image_start_s	: std_logic	:= '0';
	signal image_end_s		: std_logic	:= '0';
	signal valid_data_s		: std_logic	:= '0';

	signal i_s				: integer	:= 0;
	signal data_s0_s		: std_logic_vector(D_C-1 downto 0)	 := (others => '0');	-- "sz(t)" (original sample)
	signal output_data_s	: std_logic_vector(B_C*8-1 downto 0) := (others => '0');	-- Encoded data

	signal flag_start, flag_stop : std_logic := '0';

begin
	clock_s <= not clock_s after PERIOD_100M_C/2;
	reset_s <= '0' after 20 ns;
	
	-- Simulation MAIN process (if this one finishes, the whole simulation too)
	test_runner : process is
	begin
		-- Process starts by setting up VUnit using test_runner_setup procedure
		test_runner_setup(runner, runner_cfg);

		while test_suite loop	-- Pay very attention to the testcase name, arrows or accents prevent the system to work...
			if run("CCSDS123 Issue2 Top Entity") then
				info("Running test case = " & to_string(running_test_case));

				flag_start <= '1';
			end if;
			
			wait until (flag_stop = '1');
		end loop;

		-- The process ends with the test_runner_cleanup procedure which will force the simulation to stop
		test_runner_cleanup(runner);
	end process;

	-- Set Watchdog with limited time (this number must be higher as any "wait for" statement used in the testbench)
	test_runner_watchdog(runner, 1 ms);

	-- Input signals stimulus calculation
	p_s0_update : process(clock_s) is
	begin
		if (rising_edge(clock_s)) then
			if (reset_s = '1') then
				i_s		  <= 0;
				data_s0_s <= (others => '0');
			else
				if (flag_start = '1' and init_fifo_s = '1') then
					if (image_end_s = '0') then
						-- TEST01_S32BE_33x01x02_C, TEST62_S32BE_05x14x03_C, TEST99_S32BE_08x06x10_C
						if (i_s < TEST01_S32BE_33x01x02_C'length) then
							data_s0_s <= std_logic_vector(TEST01_S32BE_33x01x02_C(i_s)(15 downto 0));
							i_s <= i_s + 1;
						end if;
					else	-- image_end_s = '1'
						i_s <= 0;
					end if;
				end if;
			end if;	
		end if;
	end process p_s0_update;

	-- Process to stop the simulation when finished (to use 'wait' statements, the process cannot be clocked)
	p_stop_sim : process is
	begin
		-- After one complete image, wait statement and request to finish the simulation
		wait until (image_end_s = '1');
		wait for 200 ns;
		flag_stop <= '1';
		
		-- Two possibilities to stop the simulation without VUnit framework (only the "assert" line really works, anyway)
		-- assert FALSE Report "Simulation Finished" severity FAILURE;
		-- std.env.finish;
	end process p_stop_sim;

	-- CCSDS123 Issue2 top entity	
	i_top_ccsds123_issue2 : entity vunit_lib.top_ccsds123_issue2
	generic map(
		SMPL_TYPE_G				=> tb_cfg.SMPL_TYPE_G,
		SMPL_ORDER_G			=> tb_cfg.SMPL_ORDER_G,
		HEADER_EN_G				=> tb_cfg.HEADER_EN_G,
		PREDICT_MODE_G			=> tb_cfg.PREDICT_MODE_G,
		LSUM_TYPE_G				=> tb_cfg.LSUM_TYPE_G,
		W_INIT_TYPE_G			=> tb_cfg.W_INIT_TYPE_G,
		PER_ERR_LIM_UPD_G		=> tb_cfg.PER_ERR_LIM_UPD_G,
		FIDEL_CTRL_TYPE_G		=> tb_cfg.FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G		=> tb_cfg.ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G		=> tb_cfg.REL_ERR_BAND_TYPE_G,
		ENCODER_TYPE_G			=> tb_cfg.ENCODER_TYPE_G,
		W_INIT_TABL_FLAG_G		=> '0',
		W_EXP_OFF_TABL_FLAG_G	=> '1',
		DAMP_TABLE_FLAG_G		=> '0',
		OFFSET_TABLE_FLAG_G		=> '1',
		ACCU_INIT_TABLE_FLAG_G	=> '0',
		RESTRICT_CODE_G			=> '1',
		UDEF_DATA_G				=> UDEF_DATA_C,
		SUPL_TABLE_TYPE_G		=> SUPL_TABLE_TYPE_C,
		SUPL_TABLE_PURPOSE_G	=> SUPL_TABLE_PURPOSE_C,
		SUPL_TABLE_STRUCT_G		=> SUPL_TABLE_STRUCT_C,
		SUPL_TABLE_UDATA_G		=> SUPL_TABLE_UDATA_C,
		ENDIANNESS_G			=> tb_cfg.ENDIANNESS_G
	)
	port map(
		clock_i					=> clock_s,
		reset_i					=> reset_s,
		enable_i				=> flag_start and init_fifo_s,

		init_fifo_o				=> init_fifo_s,
		image_start_o			=> image_start_s,
		image_end_o				=> image_end_s,
		
		data_s0_i				=> data_s0_s,
		valid_data_o			=> valid_data_s,
		output_data_o			=> output_data_s
	);

end architecture behavioural;