--------------------------------------------------------------------------------
-- University:	NTNU Trondheim
-- Project:		CCSDS123 Issue 2
-- Engineer:	Cristian Gil Morales
-- Date:		06/06/2021
--------------------------------------------------------------------------------
-- IP name:		top_ccsds123_issue2
--
-- Description: Top entity for the whole CCSDS123 Issue 2 algorithm on VHDL-2008
--
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.gen_labels.all;
use work.param_image.all;
use work.types_image.all;
use work.utils_image.all;
use work.comp_image.all;

use work.param_predictor.all;
use work.types_predictor.all;
use work.utils_predictor.all;
use work.comp_predictor.all;

use work.param_encoder.all;
use work.types_encoder.all;
use work.utils_encoder.all;
use work.comp_encoder.all;
	
entity top_ccsds123_issue2 is
	generic (
		-- 0: "signed" type samples, 1: "unsigned" type samples
		SMPL_TYPE_G				: std_logic := '0';
		-- 00: BSQ order, 01: BIP order, 10: BIL order
		SMPL_ORDER_G			: std_logic_vector(1 downto 0) := "01";
		-- 0: Encoder Header disabled, 1: Encoder Header enabled
		HEADER_EN_G				: std_logic := '0';
		-- 1: Full prediction mode, 0: Reduced prediction mode
		PREDICT_MODE_G			: std_logic := '1';
		-- 00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column
		LSUM_TYPE_G				: std_logic_vector(1 downto 0) := "01";
		-- 1: Custom weight init, 0: Default weight init
		W_INIT_TYPE_G			: std_logic := '1';
		-- 1: enabled, 0: disabled
		PER_ERR_LIM_UPD_G		: std_logic := '0';
		-- 00: lossless, 01: absolute error limit only, 10: relative error limit only, 11: both absolute and relative error limits
		FIDEL_CTRL_TYPE_G		: std_logic_vector(1 downto 0) := "01";
		-- 1: band-dependent, 0: band-independent (for both absolute and relative error limit assignments)
		ABS_ERR_BAND_TYPE_G		: std_logic := '1';
		REL_ERR_BAND_TYPE_G		: std_logic := '1';
		-- 00: Sample-Adaptive Entropy, 01: Hybrid Entropy, 10: Block-Adaptive Entropy
		ENCODER_TYPE_G			: std_logic_vector(1 downto 0) := "00";
		-- Flag to add the "Weight Initialization Table"
		W_INIT_TABL_FLAG_G		: std_logic := '1';
		-- Flag to add the "Weight Exponent Offset Table"
		W_EXP_OFF_TABL_FLAG_G	: std_logic := '0';
		-- Flag to add the "Damping Table"
		DAMP_TABLE_FLAG_G		: std_logic := '1';
		-- Flag to add the "Offset Table"
		OFFSET_TABLE_FLAG_G		: std_logic := '0';
		-- Flag to add the "Accumulator Initialization Table"
		ACCU_INIT_TABLE_FLAG_G	: std_logic := '0';
		-- 1: Restricted set of code options used, 0: Restricted set of code options NOT used
		RESTRICT_CODE_G			: std_logic := '0';
		-- User Defined Data
		UDEF_DATA_G				: std_logic_vector(7 downto 0) := x"AC";
		-- Array with values -> 00: unsigned integer, 01: signed integer, 10: float
		SUPL_TABLE_TYPE_G		: supl_table_type_t := (0 => "01");
		-- Array with values -> From 0 to 15 (check Table 3-1)
		SUPL_TABLE_PURPOSE_G	: supl_table_purpose_t := (0 => "0000");
		-- Array with values -> 00: zero-dimensional, 01: one-dimensional, 10: two-dimensional-zx, 11: two-dimensional-yx
		SUPL_TABLE_STRUCT_G		: supl_table_struct_t := (0 => "01");
		-- Array with values -> Supplementary User-Defined Data
		SUPL_TABLE_UDATA_G		: supl_table_udata_t := (0 => "0000");
		-- 0: Little Endian order, 1: Big Endian order
		ENDIANNESS_G			: std_logic := '0'
	);
	port (
		clock_i					: in  std_logic;
		reset_i					: in  std_logic;
		enable_i				: in  std_logic;
		
		init_fifo_o				: out std_logic;
		image_start_o			: out std_logic;							-- Flag to inform a new image is coming out
		image_end_o				: out std_logic;							-- Flag to inform current image has been finished
		
		data_s0_i				: in  std_logic_vector(D_C-1 downto 0);		-- "sz(t)" (original sample)
		valid_data_o			: out std_logic;							-- Flag to inform valid data is in the output line
		output_data_o			: out std_logic_vector(B_C*8-1 downto 0)
	);
end entity top_ccsds123_issue2;

architecture behavioural of top_ccsds123_issue2 is	
	------------------------------------------------------------------------------------------------------
	-- FUNCTIONS
	------------------------------------------------------------------------------------------------------	
	-- Local sum update, according limitations (00: Wide neighbour, 01: Narrow neighbour, 10: Wide column, 11: Narrow column)
	pure function set_lsum_type(lsum_type_in : std_logic_vector; predict_mode_in : std_logic) return std_logic_vector is
		variable lsum_type_v : std_logic_vector(1 downto 0);
	begin
		-- When Image has width NX=1, column-oriented local sums SHALL be used
		if (NX_C = 1) then
			if (lsum_type_in = WID_NEI_C) then
				lsum_type_v := WID_COL_C;
			elsif (lsum_type_in = NAR_NEI_C) then
				lsum_type_v := NAR_COL_C;
			else
				lsum_type_v	:= lsum_type_in;
			end if;
		-- Under "Full Prediction Mode", column-oriented are NOT SUGGESTED (lower priority than NX=1)
		elsif (predict_mode_in = '1') then
			if (lsum_type_in = WID_COL_C) then
				lsum_type_v := WID_NEI_C;
			elsif (lsum_type_in = NAR_COL_C) then
				lsum_type_v := NAR_NEI_C;
			else
				lsum_type_v	:= lsum_type_in;
			end if;
		else
			lsum_type_v	:= lsum_type_in;
		end if;
		
		return lsum_type_v;
	end function set_lsum_type;

	------------------------------------------------------------------------------------------------------
	-- DESIGN CONSTANTS
	------------------------------------------------------------------------------------------------------
	-- For the undefined case ("11"), automatically assumed to be BSQ order...
	constant SMPL_ORDER_C	: std_logic_vector(1 downto 0)	:= iif(SMPL_ORDER_G = "11", BSQ_C, SMPL_ORDER_G);
	-- User chooses the prediction mode, unless NX_C=1, when only 'reduced prediction mode' can be used
	constant PREDICT_MODE_C	: std_logic						:= iif(NX_C = 1, '0', PREDICT_MODE_G);
	-- Local sum update, according limitations
	constant LSUM_TYPE_C	: std_logic_vector(1 downto 0)	:= set_lsum_type(LSUM_TYPE_G, PREDICT_MODE_C);
	-- Periodic error limit updating is not permitted when BSQ encoding order is used
	constant PER_ERR_LIM_UPD_C : std_logic					:= iif(SMPL_ORDER_C /= BSQ_C, PER_ERR_LIM_UPD_G, '0');

	------------------------------------------------------------------------------------------------------
	-- SIGNALS
	------------------------------------------------------------------------------------------------------
	-- Enable and image coordinates to interconnect all sub-blocks
	signal enable_in_s		  : std_logic	:= '0';
	signal enable_mid1_s	  : std_logic	:= '0';
	signal enable_mid2_s	  : std_logic	:= '0';
	signal enable_out_s		  : std_logic	:= '0';
	signal img_coord_in_s	  : img_coord_t := reset_img_coord(SMPL_ORDER_C);
	signal img_coord_mid1_s	  : img_coord_t := reset_img_coord(SMPL_ORDER_C);
	signal img_coord_mid2_s	  : img_coord_t := reset_img_coord(SMPL_ORDER_C);
	signal img_coord_out_s	  : img_coord_t := reset_img_coord(SMPL_ORDER_C);

	signal init_fifo1_s		  : std_logic	:= '0';
	signal init_fifo2_s		  : std_logic	:= '0';
	signal image_start_s	  : std_logic	:= '0';
	signal image_end_s		  : std_logic	:= '0';
	signal valid_data_s		  : std_logic	:= '0';

	signal err_lim_upd_en_s	  : std_logic	:= '0';
	signal err_lim_val_s	  : std_logic_vector(D_C-1 downto 0)	:= (others => '0');
	signal data_mp_quan_in_s  : std_logic_vector(D_C-1 downto 0)	:= (others => '0');
	signal data_mp_quan_out_s : std_logic_vector(D_C-1 downto 0)	:= (others => '0');
	signal output_data_s	  : std_logic_vector(B_C*8-1 downto 0)	:= (others => '0');

begin
	-- Image coordinates and input errors management
	i_image_coord_control : image_coord_control
	generic map(
		SMPL_ORDER_G		=> SMPL_ORDER_C
	)
	port map(
		clock_i				=> clock_i,
		reset_i				=> reset_i,

		enable_i			=> enable_i,
		enable_o			=> enable_in_s,

		image_end_i			=> image_end_s,
		img_coord_o			=> img_coord_in_s
	);
	
	-- Predictor top entity
	i_top_predictor : top_predictor
	generic map(
		SMPL_TYPE_G			=> SMPL_TYPE_G,
		SMPL_ORDER_G		=> SMPL_ORDER_C,
		FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
		LSUM_TYPE_G			=> LSUM_TYPE_C,
		PREDICT_MODE_G		=> PREDICT_MODE_C,
		ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G,
		PER_ERR_LIM_UPD_G	=> PER_ERR_LIM_UPD_C,
		W_INIT_TYPE_G		=> W_INIT_TYPE_G
	)
	port map(
		clock_i				=> clock_i,
		reset_i				=> reset_i or image_end_s,

		enable_i			=> enable_in_s,
		enable_o			=> enable_mid1_s,
		img_coord_i			=> img_coord_in_s,
		img_coord_o			=> img_coord_mid1_s,

		data_s0_i			=> data_s0_i,
		data_mp_quan_o		=> data_mp_quan_in_s
	);
	
	-- Only if 'Periodic Error Limit Values Updating' is enabled (and so, used), an interconnection between Predictor and Encoder is required
	g_pred_enc_interconn : if (SMPL_ORDER_C /= BSQ_C and FIDEL_CTRL_TYPE_G /= LOSSLESS_C and PER_ERR_LIM_UPD_C = '1') generate
		i_pred_enc_interconn : pred_enc_interconn
		generic map(
			SMPL_ORDER_G		=> SMPL_ORDER_C,
			FIDEL_CTRL_TYPE_G	=> FIDEL_CTRL_TYPE_G,
			ABS_ERR_BAND_TYPE_G	=> ABS_ERR_BAND_TYPE_G,
			REL_ERR_BAND_TYPE_G	=> REL_ERR_BAND_TYPE_G
		)
		port map(
			clock_i          => clock_i,
			reset_i			 => reset_i or image_end_s,
			
			enable_i		 => enable_mid1_s,
			enable_o		 => enable_mid2_s,
			img_coord_i		 => img_coord_mid1_s,
			img_coord_o		 => img_coord_mid2_s,
			
			init_fifo_o		 => init_fifo1_s,

			err_lim_upd_en_o => err_lim_upd_en_s,
			err_lim_val_o	 => err_lim_val_s,
			data_mp_quan_i	 => data_mp_quan_in_s,
			data_mp_quan_o	 => data_mp_quan_out_s
		);
	else generate
		enable_mid2_s		<= enable_mid1_s;
		img_coord_mid2_s	<= img_coord_mid1_s;
		init_fifo1_s		<= '1';					-- If no FIFOs enabled, then this flag is set to '1' not to disturb the system
        err_lim_upd_en_s	<= '0';
		err_lim_val_s		<= (others => '0');
		data_mp_quan_out_s	<= data_mp_quan_in_s;
	end generate g_pred_enc_interconn;
	
	-- Encoder top entity
	i_top_encoder : top_encoder
	generic map(
		SMPL_TYPE_G				=> SMPL_TYPE_G,
		SMPL_ORDER_G			=> SMPL_ORDER_C,
		HEADER_EN_G				=> HEADER_EN_G,
		PREDICT_MODE_G			=> PREDICT_MODE_C,
		LSUM_TYPE_G				=> LSUM_TYPE_C,
		W_INIT_TYPE_G			=> W_INIT_TYPE_G,
		PER_ERR_LIM_UPD_G		=> PER_ERR_LIM_UPD_C,
		FIDEL_CTRL_TYPE_G		=> FIDEL_CTRL_TYPE_G,
		ABS_ERR_BAND_TYPE_G		=> ABS_ERR_BAND_TYPE_G,
		REL_ERR_BAND_TYPE_G		=> REL_ERR_BAND_TYPE_G,
		ENCODER_TYPE_G			=> ENCODER_TYPE_G,
		W_INIT_TABL_FLAG_G		=> W_INIT_TABL_FLAG_G,
		W_EXP_OFF_TABL_FLAG_G	=> W_EXP_OFF_TABL_FLAG_G,
		DAMP_TABLE_FLAG_G		=> DAMP_TABLE_FLAG_G,
		OFFSET_TABLE_FLAG_G		=> OFFSET_TABLE_FLAG_G,
		ACCU_INIT_TABLE_FLAG_G	=> ACCU_INIT_TABLE_FLAG_G,
		RESTRICT_CODE_G			=> RESTRICT_CODE_G,
		UDEF_DATA_G				=> UDEF_DATA_G,
		SUPL_TABLE_TYPE_G		=> SUPL_TABLE_TYPE_G,
		SUPL_TABLE_PURPOSE_G	=> SUPL_TABLE_PURPOSE_G,
		SUPL_TABLE_STRUCT_G		=> SUPL_TABLE_STRUCT_G,
		SUPL_TABLE_UDATA_G		=> SUPL_TABLE_UDATA_G,
		ENDIANNESS_G			=> ENDIANNESS_G
	)
	port map(
		clock_i				=> clock_i,
		reset_i				=> reset_i,

		enable_i			=> enable_mid2_s,
		enable_o			=> enable_out_s,
		img_coord_i			=> img_coord_mid2_s,
		img_coord_o			=> img_coord_out_s,
		
		init_fifo_o			=> init_fifo2_s,
		image_start_o		=> image_start_s,
		image_end_o			=> image_end_s,
		
		err_lim_upd_en_i    => err_lim_upd_en_s,
		err_lim_val_i		=> err_lim_val_s,
		data_mp_quan_i		=> data_mp_quan_out_s,
		valid_data_o		=> valid_data_s,
		output_data_o		=> output_data_s
	);
	
	-- Outputs
	init_fifo_o		<= init_fifo1_s and init_fifo2_s;
	image_start_o	<= image_start_s;
	image_end_o		<= image_end_s;
	valid_data_o	<= valid_data_s;
	output_data_o	<= output_data_s;
end architecture behavioural;