######################################################################################################
########################################## PROJECT VARIABLES #########################################
######################################################################################################

# Set the project name
set xil_proj_name "ccsds123issue2_project"

# Set the default library
set default_library "work"

# Set the SoC HW reference for the project
set fpga_board_ref xczu9eg-ffvb1156-2-e;					# Other references: xc7z020clg400-1, xczu9eg-ffvc900-1-e
set fpga_board_part xilinx.com:zcu102:part0:3.3
set ev_board_ref zcu102

# Project main HDL to use. Possible values: "VHDL", "Verilog" or "Mixed"
set project_language "VHDL"
# Specific language for both the source/simulation and constraint files
set sources_language "VHDL 2008"
set constr_language  "XDC"

# The number of CPU cores to use for the "Synthesis" and "Implementation" processes
set num_cores	 6
# Enable/Disable the Synthesis Out-Of-Context (disabled makes process faster and more efficient)
set synth_ooc	 false
# To automatically perform the "Synthesis" process
set synth_enable true
# To automatically perform the "Implementation" process (it also includes the bitstream generation)
set imple_enable true

######################################################################################################
################################# PROJECT CREATION AND CONFIGURATION #################################
######################################################################################################
puts " -----> INFO: Project is created and configured!"

# Set the reference directory for source file relative paths (by default the value is script directory path)
set origin_dir "."

# Set the directory path for the current user
set user_dir "~"

# Create project
create_project -force ${xil_proj_name} ./${xil_proj_name} -part ${fpga_board_ref}

# Set the directory path for the (above) new project
set proj_dir [get_property directory [current_project]]

# Set project properties
set obj [current_project]

set_property -name "default_lib" -value ${default_library} -objects ${obj}
# set_property -name "part" -value "${fpga_board_ref}" -objects ${obj}
set_property -name "board_part" -value "${fpga_board_part}" -objects ${obj}
set_property -name "dsa.board_id" -value "${ev_board_ref}" -objects ${obj}
set_property -name "dsa.accelerator_binary_content" -value "bitstream" -objects ${obj}
set_property -name "dsa.accelerator_binary_format" -value "xclbin2" -objects ${obj}
set_property -name "dsa.description" -value "Vivado generated DSA" -objects ${obj}
set_property -name "dsa.dr_bd_base_address" -value "0" -objects ${obj}
set_property -name "dsa.emu_dir" -value "emu" -objects ${obj}
set_property -name "dsa.flash_interface_type" -value "bpix16" -objects ${obj}
set_property -name "dsa.flash_offset_address" -value "0" -objects ${obj}
set_property -name "dsa.flash_size" -value "1024" -objects ${obj}
set_property -name "dsa.host_architecture" -value "x86_64" -objects ${obj}
set_property -name "dsa.host_interface" -value "pcie" -objects ${obj}
set_property -name "dsa.num_compute_units" -value "60" -objects ${obj}
set_property -name "dsa.platform_state" -value "pre_synth" -objects ${obj}
set_property -name "dsa.vendor" -value "xilinx" -objects ${obj}
set_property -name "dsa.version" -value "0.0" -objects ${obj}
set_property -name "enable_vhdl_2008" -value "1" -objects ${obj}
set_property -name "ip_cache_permissions" -value "read write" -objects ${obj}
set_property -name "ip_output_repo" -value "${proj_dir}/${xil_proj_name}.cache/ip" -objects ${obj}
set_property -name "mem.enable_memory_map_generation" -value "1" -objects ${obj}
set_property -name "sim.central_dir" -value "${proj_dir}/${xil_proj_name}.ip_user_files" -objects ${obj}
set_property -name "sim.ip.auto_export_scripts" -value "1" -objects ${obj}
set_property -name "simulator_language" -value ${project_language} -objects ${obj}
set_property -name "target_language" -value ${project_language} -objects ${obj}
set_property -name "xpm_libraries" -value "XPM_CDC" -objects ${obj}

######################################################################################################
####################################### SOURCING ALL TCL FILES #######################################
######################################################################################################
puts " -----> INFO: All TCL files are sourced!"

source scripts/files_source.tcl
source scripts/files_vivado_ip.tcl
source scripts/files_bd.tcl
source scripts/files_constraint.tcl
source scripts/artifacts_generation.tcl

######################################################################################################
####################################### SOURCE FILESET CREATION ######################################
######################################################################################################
puts " -----> INFO: Source files are introduced into the project!"

# Create 'sources_1' fileset (if not found)
if {[string equal [get_filesets -quiet sources_1] ""]} {
	create_fileset -srcset sources_1
}

# Set 'sources_1' fileset object (first element is set later as the top entity)
# NOTE: List "source_files_list" has previously been sourced from file "source_files.tcl"
add_files -norecurse -fileset [get_filesets sources_1] ${source_files_list}

# Name of the top entity file (the first element from "source_files_list") is extracted
set top_entity_name [file rootname [file tail [lindex ${source_files_list} 0]]]

foreach {source_file} ${source_files_list} {
    set file_obj [get_files -of_objects [get_filesets sources_1] [list "*$source_file"]]
	set_property -name "file_type" -value ${sources_language} -objects ${file_obj}
	set_property -name "library" -value ${default_library} -objects ${file_obj}
	set_property -name "used_in_synthesis" -value true -objects ${file_obj}
	set_property -name "used_in_simulation" -value true -objects ${file_obj}
}

# Set 'sources_1' fileset properties (defining the first element from previous list as the top entity)
set_property -name "top" -value ${top_entity_name} -objects [get_filesets sources_1]

######################################################################################################
###################################### VIVADO IP(s) INTEGRATION ######################################
######################################################################################################
puts " -----> INFO: Vivado IPs (if any) are introduced into the project!"

# Create a wrapper for every previous block design, and add them into the project (ONLY if defined!)
if {[info exists vivado_ip_list]} {
	foreach {vivado_ip_file} ${vivado_ip_list} {
		# Vivado IP file is called
		source ${vivado_ip_file}
	}
}

######################################################################################################
##################################### BLOCK DESIGN(S) INTEGRATION ####################################
######################################################################################################
puts " -----> INFO: Block designs (if any) are introduced into the project!"

# Instantiate block design(s)
# NOTE: List "bd_files_list" has previously been sourced from file "bd_files.tcl"

# Create a wrapper for every previous block design, and add them into the project (ONLY if defined!)
if {[info exists bd_files_list]} {
	foreach {bd_file} ${bd_files_list} {
		# Block design file is called
		source ${bd_file}
		# File name (and without extension)
		set bd_filename [file rootname [file tail ${bd_file}]]
		# A wrapper for the Block Design is created
		make_wrapper -files [get_files ${origin_dir}/${xil_proj_name}/${xil_proj_name}.srcs/sources_1/bd/${bd_filename}/${bd_filename}.bd] -top
		# The BD wrapper file is added to the project
		add_files -norecurse ${origin_dir}/${xil_proj_name}/${xil_proj_name}.srcs/sources_1/bd/${bd_filename}/hdl/${bd_filename}_wrapper.vhd
	}
}

######################################################################################################
#################################### CONSTRAINTS FILESET CREATION ####################################
######################################################################################################
puts " -----> INFO: Constraint files are introduced into the project!"

# Create 'constrs_1' fileset (if not found)
if {[string equal [get_filesets -quiet constrs_1] ""]} {
	create_fileset -constrset constrs_1
}

# Set 'constrs_1' fileset object
# NOTE: List "constr_files_list" has previously been sourced from file "constr_files.tcl"
add_files -norecurse -fileset [get_filesets constrs_1] ${constr_files_list}

foreach {constr_file} ${constr_files_list} {
	set file_obj [get_files -of_objects [get_filesets constrs_1] [list "*$constr_file"]]
	set_property -name "file_type" -value ${constr_language} -objects ${file_obj}
	set_property -name "library" -value ${default_library} -objects ${file_obj}
	set_property -name "used_in_synthesis" -value true -objects ${file_obj}
	set_property -name "used_in_implementation" -value true -objects ${file_obj}
}

# Set 'constrs_1' fileset properties
set_property -name "target_part" -value "${fpga_board_ref}" -objects [get_filesets constrs_1]

######################################################################################################
##################################### SIMULATION FILESET CREATION ####################################
######################################################################################################
puts " -----> INFO: Simulation files are introduced into the project!"

# Create 'sim_1' fileset (if not found)
if {[string equal [get_filesets -quiet sim_1] ""]} {
	create_fileset -simset sim_1
}

# Set 'sim_1' fileset properties
set obj [get_filesets sim_1]
set_property -name "top" -value ${top_entity_name} -objects ${obj}
set_property -name "top_lib" -value ${default_library} -objects ${obj}

######################################################################################################
###################################### PROJECT CREATION FINISHED #####################################
######################################################################################################
puts " -----> INFO: Project has been successfully created!"

update_compile_order -fileset sources_1
update_compile_order -fileset sim_1

######################################################################################################
##################################### SYNTHESIS FILESET FINISHED #####################################
######################################################################################################
puts " -----> INFO: Synthesis process is defined!"

# Create 'synth_1' run (if not found)
if {[string equal [get_runs -quiet synth_1] ""]} {
	create_run -name synth_1 -part ${fpga_board_ref} -flow {Vivado Synthesis 2019} -strategy "Vivado Synthesis Defaults" -report_strategy {No Reports} -constrset constrs_1
} else {
	set_property strategy "Vivado Synthesis Defaults" [get_runs synth_1]
	set_property flow "Vivado Synthesis 2019" [get_runs synth_1]
}

set obj [get_runs synth_1]
set_property set_report_strategy_name 1 ${obj}
set_property report_strategy {Vivado Synthesis Default Reports} ${obj}
set_property set_report_strategy_name 0 ${obj}

set_property -name "part" -value "${fpga_board_ref}" -objects ${obj}
set_property -name "strategy" -value "Vivado Synthesis Defaults" -objects ${obj}

# set the current synth run
current_run -synthesis [get_runs synth_1]

######################################################################################################
################################### IMPLEMENTATION FILESET FINISHED ##################################
######################################################################################################
puts " -----> INFO: Implementation process is defined!"

# Create 'impl_1' run (if not found)
if {[string equal [get_runs -quiet impl_1] ""]} {
	create_run -name impl_1 -part ${fpga_board_ref} -flow {Vivado Implementation 2019} -strategy "Vivado Implementation Defaults" -report_strategy {No Reports} -constrset constrs_1 -parent_run synth_1
} else {
	set_property strategy "Vivado Implementation Defaults" [get_runs impl_1]
	set_property flow "Vivado Implementation 2019" [get_runs impl_1]
}

set obj [get_runs impl_1]
set_property set_report_strategy_name 1 ${obj}
set_property report_strategy {Vivado Implementation Default Reports} ${obj}
set_property set_report_strategy_name 0 ${obj}

set_property -name "part" -value "${fpga_board_ref}" -objects ${obj}
set_property -name "strategy" -value "Vivado Implementation Defaults" -objects ${obj}
set_property -name "steps.write_bitstream.args.readback_file" -value "0" -objects ${obj}
set_property -name "steps.write_bitstream.args.verbose" -value "0" -objects ${obj}

# set the current impl run
current_run -implementation [get_runs impl_1]

####################################################################################################
################ SYNTHESIS/IMPLEMENTATION + BITSTREAM/HW DESCRIPTION FILE GENERATION ###############
####################################################################################################
puts " -----> INFO: Artifacts are generated!"

# Xilinx Vivado "Synthesis" process execution
if { ${synth_enable} == true } {
	run_synthesis ${synth_ooc} ${num_cores}
	
	# Xilinx Vivado "Implementation" process execution
	if { ${imple_enable} == true } {
		run_implementation ${num_cores} ${origin_dir} ${xil_proj_name} ${top_entity_name}
	}
}
